"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const tyr_1 = require("../tyr");
const type_1 = require("../core/type");
const TextType = new type_1.default({
    name: 'text',
    typescript: 'string',
    compile(compiler, field) {
        if (compiler.stage === 'link') {
            const mediaType = field.def.in;
            if (mediaType) {
                if (!tyr_1.default.byName.mediaType.byId(mediaType))
                    throw compiler.err(field.pathName, `"in" must be a valid media type but got "${mediaType}"`);
                field.mediaType = mediaType;
            }
        }
    },
    compare(field, a, b) {
        switch (field.mediaType) {
            case 'text/html':
                return tyr_1.default.unhtmlize(a).localeCompare(tyr_1.default.unhtmlize(b), undefined, {
                    sensitivity: 'base',
                });
            default:
                return a.localeCompare(b, undefined, { sensitivity: 'base' });
        }
    },
    query(path, where, query) {
        if (where) {
            query[path.name] = _.isArray(where) ? { $in: where } : where;
        }
    },
    matches(path, where, doc) {
        if (where) {
            let value = path.get(doc);
            if (value) {
                if (!_.isString(value)) {
                    value = '' + value;
                }
                if (_.isArray(where)) {
                    return where.indexOf(value) >= 0;
                }
                else if (where instanceof RegExp) {
                    return value.match(where);
                }
                else {
                    return value === where;
                }
            }
        }
        return true;
    },
    width: 200,
});
exports.default = TextType;
//# sourceMappingURL=text.js.map