"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tyranid_1 = require("tyranid");
const Province = new tyranid_1.Tyr.Collection({
    id: '_g2',
    name: 'province',
    dbName: 'provinces',
    internal: true,
    static: true,
    fields: {
        _id: { is: 'integer' },
        name: { is: 'string', labelField: true },
        code: {
            is: 'string',
            description: 'ISO 3166-2 province code.  Should be FIPS 5-2?',
        },
        fips10_4: { is: 'string', description: 'FIPS 10-4 province code.' },
        anyCode: {
            is: 'string',
            description: 'AnyChart code, maybe we can nuke this, but AnyChart is still active',
        },
        country: { link: 'country' },
        region: { is: 'integer' },
        aliases: { is: 'string' },
        notes: { is: 'string' },
    },
    values: [
        /* eslint-disable quotes */
        ['_id', 'name', 'fips10_4', 'anyCode', 'country'],
        [1, 'Alabama', '01', 'USA-ALB', 4306, { code: 'AL', region: 4337 }],
        [2, 'Alaska', '02', 'USA-ALK', 4306, { code: 'AK', region: 4341 }],
        [3, 'Arizona', '04', 'USA-ARZ', 4306, { code: 'AZ', region: 4339 }],
        [4, 'Arkansas', '05', 'USA-ARK', 4306, { code: 'AR', region: 4344 }],
        [5, 'California', '06', 'USA-CAL', 4306, { code: 'CA', region: 4341 }],
        [6, 'Colorado', '08', 'USA-COL', 4306, { code: 'CO', region: 4339 }],
        [7, 'Connecticut', '09', 'USA-CNN', 4306, { code: 'CT', region: 4340 }],
        [8, 'Delaware', '10', 'USA-DEL', 4306, { code: 'DE', region: 4342 }],
        [9, 'Florida', '12', 'USA-FLA', 4306, { code: 'FL', region: 4342 }],
        [10, 'Georgia', '13', 'USA-GEO', 4306, { code: 'GA', region: 4342 }],
        [11, 'Hawaii', '15', 'USA-HAW', 4306, { code: 'HI', region: 4341 }],
        [12, 'Idaho', '16', 'USA-IDA', 4306, { code: 'ID', region: 4339 }],
        [13, 'Illinois', '17', 'USA-ILL', 4306, { code: 'IL', region: 4336 }],
        [14, 'Indiana', '18', 'USA-IND', 4306, { code: 'IN', region: 4336 }],
        [15, 'Iowa', '19', 'USA-IOW', 4306, { code: 'IA', region: 4343 }],
        [16, 'Kansas', '20', 'USA-KAN', 4306, { code: 'KS', region: 4343 }],
        [17, 'Kentucky', '21', 'USA-KEN', 4306, { code: 'KY', region: 4337 }],
        [18, 'Louisiana', '22', 'USA-LOU', 4306, { code: 'LA', region: 4344 }],
        [19, 'Maine', '23', 'USA-MAI', 4306, { code: 'ME', region: 4340 }],
        [20, 'Maryland', '24', 'USA-MRY', 4306, { code: 'MD', region: 4342 }],
        [21, 'Massachusetts', '25', 'USA-MSS', 4306, { code: 'MA', region: 4340 }],
        [22, 'Michigan', '26', 'USA-MCH', 4306, { code: 'MI', region: 4336 }],
        [23, 'Minnesota', '27', 'USA-MNN', 4306, { code: 'MN', region: 4343 }],
        [24, 'Mississippi', '28', 'USA-MSP', 4306, { code: 'MS', region: 4337 }],
        [25, 'Missouri', '29', 'USA-MOS', 4306, { code: 'MO', region: 4343 }],
        [26, 'Montana', '30', 'USA-MNT', 4306, { code: 'MT', region: 4339 }],
        [27, 'Nebraska', '31', 'USA-NEB', 4306, { code: 'NE', region: 4343 }],
        [28, 'Nevada', '32', 'USA-NEV', 4306, { code: 'NV', region: 4339 }],
        [29, 'New Hampshire', '33', 'USA-NHM', 4306, { code: 'NH', region: 4340 }],
        [30, 'New Jersey', '34', 'USA-NJR', 4306, { code: 'NJ', region: 4338 }],
        [31, 'New Mexico', '35', 'USA-NME', 4306, { code: 'NM', region: 4339 }],
        [32, 'New York', '36', 'USA-NYO', 4306, { code: 'NY', region: 4338 }],
        [33, 'North Carolina', '37', 'USA-NCR', 4306, { code: 'NC', region: 4342 }],
        [34, 'North Dakota', '38', 'USA-NDA', 4306, { code: 'ND', region: 4343 }],
        [35, 'Ohio', '39', 'USA-OHI', 4306, { code: 'OH', region: 4336 }],
        [36, 'Oklahoma', '40', 'USA-OKL', 4306, { code: 'OK', region: 4344 }],
        [37, 'Oregon', '41', 'USA-ORE', 4306, { code: 'OR', region: 4341 }],
        [38, 'Pennsylvania', '42', 'USA-PEN', 4306, { code: 'PA', region: 4338 }],
        [39, 'Rhode Island', '44', 'USA-RHI', 4306, { code: 'RI', region: 4340 }],
        [40, 'South Carolina', '45', 'USA-SCR', 4306, { code: 'SC', region: 4342 }],
        [41, 'South Dakota', '46', 'USA-SDA', 4306, { code: 'SD', region: 4343 }],
        [42, 'Tennessee', '47', 'USA-TNN', 4306, { code: 'TN', region: 4337 }],
        [43, 'Texas', '48', 'USA-TEX', 4306, { code: 'TX', region: 4344 }],
        [44, 'Utah', '49', 'USA-UTA', 4306, { code: 'UT', region: 4339 }],
        [45, 'Vermont', '50', 'USA-VRM', 4306, { code: 'VT', region: 4340 }],
        [46, 'Virginia', '51', 'USA-VRG', 4306, { code: 'VA', region: 4342 }],
        [47, 'Washington', '53', 'USA-WSH', 4306, { code: 'WA', region: 4341 }],
        [48, 'West Virginia', '54', 'USA-WVR', 4306, { code: 'WV', region: 4342 }],
        [49, 'Wisconsin', '55', 'USA-WIS', 4306, { code: 'WI', region: 4336 }],
        [50, 'Wyoming', '56', 'USA-WYM', 4306, { code: 'WY', region: 4339 }],
        [51, 'Armed Forces Americas', '', , 4306, { code: 'AA' }],
        [
            52,
            'Armed Forces Europe, Middle East, & Canada',
            '',
            ,
            4306,
            { code: 'AE' },
        ],
        [53, 'Armed Forces Pacific', '', , 4306, { code: 'AP' }],
        [
            54,
            'American Samoa',
            '03',
            ,
            4306,
            { code: 'AS', notes: "Also FIPS 5-2 '60'" },
        ],
        [55, 'Federated States of Micronesia', '64', , 4306, { code: 'FM' }],
        [56, 'Guam', '14', , 4306, { code: 'GU' }],
        [57, 'Marshall Islands', '68', , 4306, { code: 'MH' }],
        [58, 'Northern Mariana Islands', '69', , 4306, { code: 'MP' }],
        [59, 'Puerto Rico', '43', , 4306, { code: 'PR' }],
        [60, 'Palau', '70', , 4306, { code: 'PW' }],
        [61, 'Virgin Islands', '52', , 4306, { code: 'VI' }],
        [62, 'Canillo', '02', , 4077],
        [63, 'Encamp', '03', , 4077],
        [64, 'La Massana', '04', , 4077],
        [65, 'Ordino', '05', , 4077],
        [66, 'Sant Julia de Loria', '06', , 4077],
        [67, 'Andorra la Vella', '07', , 4077],
        [68, 'Escaldes-Engordany', '08', , 4077],
        [69, 'Abu Dhabi', '01', 'ARE-AZA', 4078],
        [70, 'Ajman', '02', 'ARE-AJM', 4078],
        [71, 'Dubai', '03', 'ARE-DBY', 4078],
        [72, 'Fujairah', '04', 'ARE-AFJ', 4078],
        [73, 'Ras Al Khaimah', '05', 'ARE-RAK', 4078],
        [74, 'Sharjah', '06', 'ARE-ASH', 4078],
        [75, 'Umm Al Quwain', '07', 'ARE-UAQ', 4078],
        [76, 'Badakhshan', '01', 'AFG-BDK', 4079],
        [77, 'Badghis', '02', 'AFG-BDG', 4079],
        [78, 'Baghlan', '03', 'AFG-BGH', 4079],
        [79, 'Bamian', '05', 'AFG-BAM', 4079],
        [80, 'Farah', '06', 'AFG-FAR', 4079],
        [81, 'Faryab', '07', 'AFG-FRY', 4079],
        [82, 'Ghazni', '08', 'AFG-GHZ', 4079],
        [83, 'Ghowr', '09', 'AFG-GHW', 4079],
        [84, 'Helmand', '10', 'AFG-HLM', 4079],
        [85, 'Herat', '11', 'AFG-HER', 4079],
        [86, 'Kabol', '13', 'AFG-KAB', 4079],
        [87, 'Kapisa', '14', 'AFG-KAP', 4079],
        [88, 'Konar', '15', 'AFG-KNR', 4079],
        [89, 'Laghman', '16', 'AFG-LGH', 4079],
        [90, 'Lowgar', '17', 'AFG-LWG', 4079],
        [91, 'Nangarhar', '18', 'AFG-NNG', 4079],
        [92, 'Nimruz', '19', 'AFG-NMR', 4079],
        [93, 'Paktia', '21', 'AFG-PAK', 4079],
        [94, 'Parvan', '22', 'AFG-PRV', 4079],
        [95, 'Kandahar', '23', 'AFG-QND', 4079, { aliases: 'Quandahar' }],
        [96, 'Kondoz', '24', 'AFG-KNZ', 4079],
        [97, 'Takhar', '26', 'AFG-TKH', 4079],
        [98, 'Vardak', '27', 'AFG-VRD', 4079],
        [99, 'Zabol', '28', 'AFG-ZAB', 4079],
        [100, 'Paktika', '29', 'AFG-PKT', 4079],
        [101, 'Balkh', '30', 'AFG-BLK', 4079],
        [102, 'Jowzjan', '31', 'AFG-JWZ', 4079],
        [103, 'Samangan', '32', 'AFG-SMN', 4079],
        [104, 'Sar-e Pol', '33', 'AFG-JWZ', 4079],
        [105, 'Konar', '34', 'AFG-KNR', 4079],
        [106, 'Laghman', '35', 'AFG-LGH', 4079],
        [107, 'Paktia', '36', 'AFG-PAK', 4079],
        [108, 'Khowst', '37', 'AFG-PAK', 4079, { notes: 'City, Provice 93' }],
        [109, 'Nurestan', '38', 'AFG-KNR', 4079],
        [110, 'Oruzgan', '39', 'AFG-ORZ', 4079],
        [111, 'Parvan', '40', 'AFG-PRV', 4079],
        [112, 'Daykondi', '41', 'AFG-ORZ', 4079],
        [113, 'Panjshir', '42', 'AFG-KAP', 4079],
        [114, 'Barbuda', '01', , 4080],
        [115, 'Saint George', '04', , 4310],
        [116, 'Saint John', '04', , 4154],
        [117, 'Saint Mary', '13', 'JAM-SMA', 4188],
        [118, 'Saint Paul', '10', , 4135],
        [119, 'Saint Peter', '03', , 4227],
        [120, 'Saint Philip', '08', , 4080],
        [121, 'Berat', '40', 'ALB-BER', 4082],
        [122, 'Diber', '41', 'ALB-DIB', 4082],
        [123, 'Durres', '42', 'ALB-DRR', 4082],
        [124, 'Elbasan', '43', 'ALB-ELB', 4082],
        [125, 'Fier', '44', 'ALB-FIE', 4082],
        [126, 'Gjirokaster', '45', 'ALB-GJR', 4082],
        [127, 'Korce', '46', 'ALB-KOR', 4082],
        [128, 'Kukes', '47', 'ALB-KUK', 4082],
        [129, 'Lezhe', '48', 'ALB-LEZ', 4082],
        [130, 'Shkoder', '49', 'ALB-SHK', 4082],
        [131, 'Tirane', '50', 'ALB-TIR', 4082],
        [132, 'Vlore', '51', 'ALB-VLO', 4082],
        [133, 'Aragatsotn', '01', , 4083],
        [134, 'Ararat', '02', , 4083],
        [135, 'Armavir', '03', , 4083],
        [136, "Geghark'unik'", '04', , 4083],
        [137, "Kotayk'", '05', , 4083],
        [138, 'Lorri', '06', , 4083],
        [139, 'Shirak', '07', , 4083],
        [140, "Syunik'", '08', , 4083],
        [141, 'Tavush', '09', , 4083],
        [142, "Vayots' Dzor", '10', , 4083],
        [143, 'Yerevan', '11', , 4083],
        [144, 'Benguela', '01', 'AGO-BNG', 4085],
        [145, 'Bie', '02', 'AGO-BIE', 4085],
        [146, 'Cabinda', '03', 'AGO-CBN', 4085],
        [147, 'Cuando Cubango', '04', 'AGO-CCB', 4085],
        [148, 'Cuanza Norte', '05', 'AGO-CNO', 4085],
        [149, 'Cuanza Sul', '06', 'AGO-CSU', 4085],
        [150, 'Cunene', '07', 'AGO-CUN', 4085],
        [151, 'Huambo', '08', 'AGO-HUA', 4085],
        [152, 'Huila', '16', 'COL-HUI', 4125],
        [153, 'Luanda', '10', 'AGO-LUA', 4085],
        [154, 'Malanje', '12', 'AGO-MLN', 4085],
        [155, 'Moxico', '14', 'AGO-MOX', 4085],
        [156, 'Uige', '15', 'AGO-UIG', 4085],
        [157, 'Zaire', '16', 'AGO-ZAI', 4085],
        [158, 'Lunda Norte', '17', 'AGO-LNO', 4085],
        [159, 'Lunda Sul', '18', 'AGO-LSU', 4085],
        [160, 'Bengo', '19', 'AGO-BEN', 4085],
        [161, 'Luanda', '20', 'AGO-LUA', 4085],
        [162, 'Buenos Aires', '01', 'ARG-BAI', 4088],
        [163, 'Catamarca', '02', 'ARG-CTM', 4088],
        [164, 'Chaco', '20', 'PRY-CHA', 4261],
        [165, 'Chubut', '04', 'ARG-CHB', 4088],
        [166, 'Cordoba', '12', 'COL-CRD', 4125],
        [167, 'Corrientes', '06', 'ARG-CRR', 4088],
        [168, 'Distrito Federal', '25', 'VEN-DFD', 4311],
        [169, 'Entre Rios', '08', 'ARG-ERI', 4088],
        [170, 'Formosa', '09', 'ARG-FRM', 4088],
        [171, 'Jujuy', '10', 'ARG-JUJ', 4088],
        [172, 'La Pampa', '11', 'ARG-LPA', 4088],
        [173, 'La Rioja', '27', 'ESP-LJR', 4143],
        [174, 'Mendoza', '13', 'ARG-MND', 4088],
        [175, 'Misiones', '12', 'PRY-MSN', 4261],
        [176, 'Neuquen', '15', 'ARG-NEU', 4088],
        [177, 'Rio Negro', '12', 'URY-RNE', 4307],
        [178, 'Salta', '17', 'ARG-SAL', 4088],
        [179, 'San Juan', '23', , 4136],
        [180, 'San Luis', '19', 'ARG-SLU', 4088],
        [181, 'Santa Cruz', '16', , 4128],
        [182, 'Santa Fe', '21', 'ARG-SFE', 4088],
        [183, 'Santiago del Estero', '22', 'ARG-SDE', 4088],
        [184, 'Tierra del Fuego', '23', 'ARG-TDF', 4088],
        [185, 'Tucuman', '24', 'ARG-TCM', 4088],
        [186, 'Burgenland', '01', 'AUT-BRG', 4090],
        [187, 'Karnten', '02', 'AUT-KRN', 4090],
        [188, 'Niederosterreich', '03', 'AUT-NDR', 4090],
        [189, 'Oberosterreich', '04', 'AUT-OBR', 4090],
        [190, 'Salzburg', '05', 'AUT-SLZ', 4090],
        [191, 'Steiermark', '06', 'AUT-STR', 4090],
        [192, 'Tirol', '07', 'AUT-TIR', 4090],
        [193, 'Vorarlberg', '08', 'AUT-VRR', 4090],
        [194, 'Wien', '09', 'AUT-WIE', 4090],
        [195, 'Australian Capital Territory', '01', 'AS01', 4091],
        [196, 'New South Wales', '02', 'AS02', 4091],
        [197, 'Northern Territory', '03', 'AS03', 4091],
        [198, 'Queensland', '04', 'AS04', 4091],
        [199, 'South Australia', '05', 'AS05', 4091],
        [200, 'Tasmania', '06', 'AS06', 4091],
        [201, 'Victoria', '12', , 4299],
        [202, 'Western Australia', '08', 'AS08', 4091],
        [203, 'Abseron', '01', , 4094],
        [204, 'Agcabadi', '02', , 4094],
        [205, 'Agdam', '03', , 4094],
        [206, 'Agdas', '04', , 4094],
        [207, 'Agstafa', '05', , 4094],
        [208, 'Agsu', '06', , 4094],
        [209, 'Ali Bayramli', '07', , 4094],
        [210, 'Astara', '08', , 4094],
        [211, 'Baki', '09', , 4094],
        [212, 'Balakan', '10', , 4094],
        [213, 'Barda', '11', , 4094],
        [214, 'Beylaqan', '12', , 4094],
        [215, 'Bilasuvar', '13', , 4094],
        [216, 'Cabrayil', '14', , 4094],
        [217, 'Calilabad', '15', , 4094],
        [218, 'Daskasan', '16', , 4094],
        [219, 'Davaci', '17', , 4094],
        [220, 'Fuzuli', '18', , 4094],
        [221, 'Gadabay', '19', , 4094],
        [222, 'Ganca', '20', , 4094],
        [223, 'Goranboy', '21', , 4094],
        [224, 'Goycay', '22', , 4094],
        [225, 'Haciqabul', '23', , 4094],
        [226, 'Imisli', '24', , 4094],
        [227, 'Ismayilli', '25', , 4094],
        [228, 'Kalbacar', '26', , 4094],
        [229, 'Kurdamir', '27', , 4094],
        [230, 'Lacin', '28', , 4094],
        [231, 'Lankaran', '29', , 4094],
        [232, 'Lankaran', '30', , 4094],
        [233, 'Lerik', '31', , 4094],
        [234, 'Masalli', '32', , 4094],
        [235, 'Mingacevir', '33', , 4094],
        [236, 'Naftalan', '34', , 4094],
        [237, 'Naxcivan', '35', , 4094],
        [238, 'Neftcala', '36', , 4094],
        [239, 'Oguz', '37', , 4094],
        [240, 'Qabala', '38', , 4094],
        [241, 'Qax', '39', , 4094],
        [242, 'Qazax', '40', , 4094],
        [243, 'Qobustan', '41', , 4094],
        [244, 'Quba', '42', , 4094],
        [245, 'Qubadli', '43', , 4094],
        [246, 'Qusar', '44', , 4094],
        [247, 'Saatli', '45', , 4094],
        [248, 'Sabirabad', '46', , 4094],
        [249, 'Saki', '47', , 4094],
        [250, 'Saki', '48', , 4094],
        [251, 'Salyan', '49', , 4094],
        [252, 'Samaxi', '50', , 4094],
        [253, 'Samkir', '51', , 4094],
        [254, 'Samux', '52', , 4094],
        [255, 'Siyazan', '53', , 4094],
        [256, 'Sumqayit', '54', , 4094],
        [257, 'Susa', '55', , 4094],
        [258, 'Susa', '56', , 4094],
        [259, 'Tartar', '57', , 4094],
        [260, 'Tovuz', '58', , 4094],
        [261, 'Ucar', '59', , 4094],
        [262, 'Xacmaz', '60', , 4094],
        [263, 'Xankandi', '61', , 4094],
        [264, 'Xanlar', '62', , 4094],
        [265, 'Xizi', '63', , 4094],
        [266, 'Xocali', '64', , 4094],
        [267, 'Xocavand', '65', , 4094],
        [268, 'Yardimli', '66', , 4094],
        [269, 'Yevlax', '67', , 4094],
        [270, 'Yevlax', '68', , 4094],
        [271, 'Zangilan', '69', , 4094],
        [272, 'Zaqatala', '70', , 4094],
        [273, 'Zardab', '71', , 4094],
        [274, 'Federation of Bosnia and Herzegovina', '01', , 4095],
        [275, 'Republika Srpska', '02', , 4095],
        [276, 'Christ Church', '01', , 4096],
        [277, 'Saint Andrew', '02', , 4310],
        [278, 'Saint George', '03', , 4096],
        [279, 'Saint James', '12', 'JAM-SJA', 4188],
        [280, 'Saint John', '05', , 4096],
        [281, 'Saint Joseph', '06', , 4135],
        [282, 'Saint Lucy', '07', , 4096],
        [283, 'Saint Michael', '08', , 4096],
        [284, 'Saint Peter', '09', , 4096],
        [285, 'Saint Philip', '10', , 4096],
        [286, 'Saint Thomas', '14', 'JAM-STH', 4188],
        [
            287,
            'Barisal',
            '85',
            'BGD-BRS',
            4097,
            { notes: 'County, Province |Barisal|' },
        ],
        [
            288,
            'Bandarban',
            '04',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            289,
            'Comilla',
            '05',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            290,
            'Mymensingh',
            '12',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            291,
            'Noakhali',
            '13',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            292,
            'Patuakhali',
            '15',
            'BGD-BRS',
            4097,
            { notes: 'County, Province |Barisal|' },
        ],
        [
            293,
            'Bagerhat',
            '22',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            294,
            'Bhola',
            '23',
            'BGD-BRS',
            4097,
            { notes: 'County, Province |Barisal|' },
        ],
        [
            295,
            'Bogra',
            '24',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            296,
            'Barguna',
            '25',
            'BGD-BRS',
            4097,
            { notes: 'County, Province |Barisal|' },
        ],
        [
            297,
            'Brahmanbaria',
            '26',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            298,
            'Chandpur',
            '27',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            299,
            'Chapai Nawabganj',
            '28',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            300,
            'Chattagram',
            '29',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            301,
            'Chuadanga',
            '30',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            302,
            "Cox's Bazar",
            '31',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            303,
            'Dhaka',
            '81',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            304,
            'Dinajpur',
            '33',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            305,
            'Faridpur',
            '34',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            306,
            'Feni',
            '35',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            307,
            'Gaibandha',
            '36',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            308,
            'Gazipur',
            '37',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            309,
            'Gopalganj',
            '38',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            310,
            'Habiganj',
            '39',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Sylhet|' },
        ],
        [
            311,
            'Jaipurhat',
            '40',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            312,
            'Jamalpur',
            '41',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            313,
            'Jessore',
            '42',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            314,
            'Jhalakati',
            '43',
            'BGD-BRS',
            4097,
            { notes: 'County, Province |Barisal|' },
        ],
        [
            315,
            'Jhenaidah',
            '44',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            316,
            'Khagrachari',
            '45',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            317,
            'Khulna',
            '46',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            318,
            'Kishorganj',
            '47',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            319,
            'Kurigram',
            '48',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            320,
            'Kushtia',
            '49',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            321,
            'Laksmipur',
            '50',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            322,
            'Lalmonirhat',
            '51',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            323,
            'Madaripur',
            '52',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            324,
            'Magura',
            '53',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            325,
            'Manikganj',
            '54',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            326,
            'Meherpur',
            '55',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            327,
            'Moulavibazar',
            '56',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Sylhet|' },
        ],
        [
            328,
            'Munshiganj',
            '57',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            329,
            'Naogaon',
            '58',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            330,
            'Narail',
            '59',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            331,
            'Narayanganj',
            '60',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            332,
            'Narsingdi',
            '61',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            333,
            'Nator',
            '62',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            334,
            'Netrakona',
            '63',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            335,
            'Nilphamari',
            '64',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            336,
            'Pabna',
            '65',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            337,
            'Panchagar',
            '66',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            338,
            'Parbattya Chattagram',
            '67',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            339,
            'Pirojpur',
            '68',
            'BGD-BRS',
            4097,
            { notes: 'County, Province |Barisal|' },
        ],
        [
            340,
            'Rajbari',
            '69',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            341,
            'Rajshahi',
            '70',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            342,
            'Rangpur',
            '71',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            343,
            'Satkhira',
            '72',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            344,
            'Shariyatpur',
            '73',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            345,
            'Sherpur',
            '74',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            346,
            'Sirajganj',
            '75',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            347,
            'Sunamganj',
            '76',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Sylhet|' },
        ],
        [
            348,
            'Sylhet',
            '77',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Sylhet|' },
        ],
        [
            349,
            'Tangail',
            '78',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [
            350,
            'Thakurgaon',
            '79',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            351,
            'Khulna',
            '82',
            'BGD-KHL',
            4097,
            { notes: 'County, Province |Khulna|' },
        ],
        [
            352,
            'Rajshahi',
            '83',
            'BGD-RJS',
            4097,
            { notes: 'County, Province |Rajshahi|' },
        ],
        [
            353,
            'Chittagong',
            '84',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Chittagong|' },
        ],
        [
            354,
            'Sylhet',
            '86',
            'BGD-CHT',
            4097,
            { notes: 'County, Province |Sylhet|' },
        ],
        [355, 'Antwerpen', '01', 'BEL-ANT', 4098],
        [356, 'Brabant', '02', 'BEL-BRB', 4098],
        [357, 'Hainaut', '03', 'BEL-HAI', 4098],
        [358, 'Liege', '04', 'BEL-LIE', 4098],
        [359, 'Limburg', '05', 'NLD-LMB', 4241],
        [360, 'Luxembourg', '03', , 4210],
        [361, 'Namur', '07', 'BEL-NAM', 4098],
        [362, 'Oost-Vlaanderen', '08', 'BEL-OVL', 4098],
        [363, 'West-Vlaanderen', '09', 'BEL-WVL', 4098],
        [364, 'Brabant Wallon', '10', 'BEL-NVL', 4098],
        [365, 'Brussels Hoofdstedelijk Gewest', '11', 'BEL-BRS', 4098],
        [366, 'Vlaams-Brabant', '12', 'BEL-BRB', 4098],
        [367, 'Bam', '15', 'BFA-BAM', 4099],
        [368, 'Boulkiemde', '19', 'BFA-BLK', 4099],
        [369, 'Ganzourgou', '20', 'BFA-GNZ', 4099],
        [370, 'Gnagna', '21', 'BFA-GNG', 4099],
        [371, 'Kouritenga', '28', 'BFA-KRT', 4099],
        [372, 'Oudalan', '33', 'BFA-ODL', 4099],
        [373, 'Passore', '34', 'BFA-PSS', 4099],
        [374, 'Sanguie', '36', 'BFA-SNG', 4099],
        [375, 'Soum', '40', 'BFA-SOM', 4099],
        [376, 'Tapoa', '42', 'BFA-TAP', 4099],
        [377, 'Zoundweogo', '44', 'BFA-ZND', 4099],
        [378, 'Bale', '45', 'BFA-MHN', 4099],
        [379, 'Banwa', '46', 'BFA-KSS', 4099],
        [380, 'Bazega', '47', 'BFA-BZG', 4099],
        [381, 'Bougouriba', '48', 'BFA-BGR', 4099],
        [382, 'Boulgou', '49', 'BFA-BLG', 4099],
        [383, 'Gourma', '50', 'BFA-GRM', 4099],
        [384, 'Houet', '51', 'BFA-HUT', 4099],
        [385, 'Ioba', '52', 'BFA-BGR', 4099],
        [386, 'Kadiogo', '53', 'BFA-KDG', 4099],
        [387, 'Kenedougou', '54', 'BFA-KND', 4099],
        [388, 'Komoe', '55', 'BFA-COM', 4099, { aliases: 'Comoe' }],
        [389, 'Komondjari', '56', 'BFA-GRM', 4099],
        [390, 'Kompienga', '57', 'BFA-GRM', 4099],
        [391, 'Kossi', '58', 'BFA-KSS', 4099],
        [392, 'Koulpelogo', '59', 'BFA-BLG', 4099],
        [393, 'Kourweogo', '60', 'BFA-PSS', 4099],
        [394, 'Leraba', '61', 'BFA-COM', 4099],
        [395, 'Loroum', '62', 'BFA-YTN', 4099],
        [396, 'Mouhoun', '63', 'BFA-MHN', 4099],
        [397, 'Namentenga', '64', 'BFA-NMT', 4099],
        [398, 'Naouri', '65', 'BFA-NHR', 4099],
        [399, 'Nayala', '66', 'BFA-SOR', 4099],
        [400, 'Noumbiel', '67', 'BFA-PON', 4099],
        [401, 'Oubritenga', '68', 'BFA-OBR', 4099],
        [402, 'Poni', '69', 'BFA-PON', 4099],
        [403, 'Sanmatenga', '70', 'BFA-SNM', 4099],
        [404, 'Seno', '71', 'BFA-SEN', 4099],
        [405, 'Sissili', '72', 'BFA-SSS', 4099],
        [406, 'Sourou', '73', 'BFA-SOR', 4099],
        [407, 'Tuy', '74', 'BFA-HUT', 4099],
        [408, 'Yagha', '75', 'BFA-SEN', 4099],
        [409, 'Yatenga', '76', 'BFA-YTN', 4099],
        [410, 'Ziro', '77', 'BFA-SSS', 4099],
        [411, 'Zondoma', '78', 'BFA-YTN', 4099],
        [412, 'Mikhaylovgrad', '33', 'BGR-MKH', 4100],
        [413, 'Blagoevgrad', '38', 'BGR-SOF', 4100],
        [414, 'Burgas', '39', 'BGR-BRG', 4100],
        [415, 'Dobrich', '40', 'BGR-VAR', 4100],
        [416, 'Gabrovo', '41', 'BGR-LVC', 4100],
        [417, 'Grad Sofiya', '42', 'BGR-GSO', 4100],
        [418, 'Khaskovo', '43', 'BGR-KHS', 4100],
        [419, 'Kurdzhali', '44', 'BGR-KHS', 4100],
        [420, 'Kyustendil', '45', 'BGR-SOF', 4100],
        [421, 'Lovech', '46', 'BGR-LVC', 4100],
        [422, 'Pazardzhik', '48', 'BGR-PLD', 4100],
        [423, 'Pernik', '49', 'BGR-SOF', 4100],
        [424, 'Pleven', '50', 'BGR-LVC', 4100],
        [425, 'Plovdiv', '51', 'BGR-PLD', 4100],
        [426, 'Razgrad', '52', 'BGR-RZG', 4100],
        [427, 'Ruse', 'L3', , 4275],
        [428, 'Shumen', '54', 'BGR-VAR', 4100],
        [429, 'Silistra', '55', 'BGR-RZG', 4100],
        [430, 'Sliven', '56', 'BGR-BRG', 4100],
        [431, 'Smolyan', '57', 'BGR-PLD', 4100],
        [432, 'Sofiya', '58', 'BGR-SOF', 4100],
        [433, 'Stara Zagora', '59', 'BGR-KHS', 4100],
        [434, 'Turgovishte', '60', 'BGR-RZG', 4100],
        [435, 'Varna', '61', 'BGR-VAR', 4100],
        [436, 'Veliko Turnovo', '62', 'BGR-LVC', 4100],
        [437, 'Vidin', '63', 'BGR-MKH', 4100],
        [438, 'Vratsa', '64', 'BGR-MKH', 4100],
        [439, 'Yambol', '65', 'BGR-BRG', 4100],
        [440, 'Al Hadd', '01', 'BHR-AMH', 4101],
        [441, 'Al Manamah', '02', 'BHR-AMN', 4101],
        [442, 'Al Muharraq', '03', 'BHR-AMH', 4101],
        [443, 'Jidd Hafs', '05', 'BHR-JHF', 4101],
        [444, 'Sitrah', '06', 'BHR-STR', 4101],
        [445, 'Al Mintaqah al Gharbiyah', '08', 'BHR-AMG', 4101],
        [
            446,
            'Mintaqat Juzur Hawar',
            '09',
            'BHR-JHW',
            4101,
            { aliases: 'Juzar Hawar' },
        ],
        [447, 'Al Mintaqah ash Shamaliyah', '10', 'BHR-AMS', 4101],
        [448, 'Al Mintaqah al Wusta', '11', 'BHR-AMW', 4101],
        [449, 'Madinat', '12', 'BHR-MIS', 4101, { aliases: 'Madinat Isa' }],
        [450, 'Ar Rifa', '13', 'BHR-ARM', 4101],
        [451, 'Madinat Hamad', '14', 'BHR-MHM', 4101],
        [452, 'Al Muharraq', '15', 'BHR-AMH', 4101],
        [453, 'Al Asimah', '16', 'BHR-JHF', 4101],
        [454, 'Al Janubiyah', '17', 'BHR-ARM', 4101],
        [455, 'Ash Shamaliyah', '30', 'SDN-ASH', 4271],
        [456, 'Al Wusta', '27', 'SDN-AWU', 4271],
        [457, 'Bujumbura', '02', 'BDI-BJM', 4102],
        [458, 'Bubanza', '09', 'BDI-BBN', 4102],
        [459, 'Bururi', '10', 'BDI-BUR', 4102],
        [460, 'Cankuzo', '11', 'BDI-CNK', 4102],
        [461, 'Cibitoke', '12', 'BDI-CBT', 4102],
        [462, 'Gitega', '13', 'BDI-GIT', 4102],
        [463, 'Karuzi', '14', 'BDI-KRZ', 4102],
        [464, 'Kayanza', '15', 'BDI-KYN', 4102],
        [465, 'Kirundo', '16', 'BDI-KRN', 4102],
        [466, 'Makamba', '17', 'BDI-MKM', 4102],
        [467, 'Muyinga', '18', 'BDI-MYN', 4102],
        [468, 'Ngozi', '19', 'BDI-NGO', 4102],
        [469, 'Rutana', '20', 'BDI-RTN', 4102],
        [470, 'Ruyigi', '21', 'BDI-RUY', 4102],
        [471, 'Muramvya', '22', 'BDI-MRM', 4102],
        [472, 'Mwaro', '23', 'BDI-MRM', 4102],
        [473, 'Atakora', '01', 'BEN-ATK', 4103],
        [474, 'Atlantique', '02', 'BEN-ATL', 4103],
        [475, 'Borgou', '03', 'BEN-BOR', 4103],
        [476, 'Mono', '04', 'BEN-MON', 4103],
        [477, 'Oueme', '05', 'BEN-OUE', 4103],
        [478, 'Zou', '06', 'BEN-ZOU', 4103],
        [479, 'Littoral', '05', 'CMR-LTT', 4123],
        [480, 'Devonshire', '01', , 4104],
        [481, 'Hamilton', '02', , 4104],
        [482, 'Hamilton', '03', , 4104],
        [483, 'Paget', '04', , 4104],
        [484, 'Pembroke', '05', , 4104],
        [485, "Saint George's", '07', , 4104],
        [486, 'Sandys', '08', , 4104],
        [487, 'Smiths', '09', , 4104],
        [
            488,
            'Southampton',
            'M4',
            'GBR-ENG-HMP',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [489, 'Warwick', '11', , 4104],
        [490, 'Alibori', '07', , 4105],
        [491, 'Belait', '08', , 4105],
        [492, 'Brunei and Muara', '09', , 4105],
        [493, 'Temburong', '10', , 4105],
        [494, 'Collines', '11', , 4105],
        [495, 'Kouffo', '12', , 4105],
        [496, 'Donga', '13', , 4105],
        [497, 'Littoral', '14', , 4105],
        [498, 'Tutong', '15', , 4105],
        [499, 'Oueme', '16', , 4105],
        [500, 'Plateau', '49', 'NI19', 4239],
        [501, 'Zou', '18', , 4105],
        [502, 'Chuquisaca', '01', 'BOL-CHQ', 4106],
        [503, 'Cochabamba', '02', 'BOL-CCH', 4106],
        [504, 'El Beni', '03', 'BOL-EBE', 4106],
        [505, 'La Paz', '06', 'SLV-LPA', 4284],
        [506, 'Oruro', '05', 'BOL-ORU', 4106],
        [507, 'Pando', '06', 'BOL-PAN', 4106],
        [508, 'Potosi', '07', 'BOL-POT', 4106],
        [509, 'Tarija', '09', 'BOL-TAR', 4106],
        [510, 'Acre', '01', 'BRA-ACR', 4107],
        [511, 'Alagoas', '02', 'BRA-ALG', 4107],
        [512, 'Amapa', '03', 'BRA-AMA', 4107],
        [513, 'Amazonas', '01', 'VEN-AMZ', 4311],
        [514, 'Bahia', '05', 'BRA-BAH', 4107],
        [515, 'Ceara', '06', 'BRA-CEA', 4107],
        [516, 'Espirito Santo', '08', 'BRA-ESA', 4107],
        [517, 'Mato Grosso do Sul', '11', 'BRA-MGD', 4107],
        [518, 'Maranhao', '13', 'BRA-MRN', 4107],
        [519, 'Mato Grosso', '14', 'BRA-MGR', 4107],
        [520, 'Minas Gerais', '15', 'BRA-MGE', 4107],
        [521, 'Para', '15', 'SUR-PAR', 4282],
        [522, 'Paraiba', '17', 'BRA-PAA', 4107],
        [523, 'Parana', '18', 'BRA-PAN', 4107],
        [524, 'Piaui', '20', 'BRA-PIA', 4107],
        [525, 'Rio de Janeiro', '21', 'BRA-RDJ', 4107],
        [526, 'Rio Grande do Norte', '22', 'BRA-RGN', 4107],
        [527, 'Rio Grande do Sul', '23', 'BRA-RGS', 4107],
        [528, 'Rondonia', '24', 'BRA-RND', 4107],
        [529, 'Roraima', '25', 'BRA-ROR', 4107],
        [530, 'Santa Catarina', '15', , 4128],
        [531, 'Sao Paulo', '27', 'BRA-SPA', 4107],
        [532, 'Sergipe', '28', 'BRA-SRG', 4107],
        [533, 'Goias', '29', 'BRA-GOI', 4107],
        [534, 'Pernambuco', '30', 'BRA-PRN', 4107],
        [535, 'Tocantins', '31', 'BRA-TCN', 4107],
        [536, 'Bimini', '05', , 4108],
        [537, 'Cat Island', '06', , 4108],
        [538, 'Exuma', '10', , 4108],
        [539, 'Inagua', '13', , 4108],
        [540, 'Long Island', '15', , 4108],
        [541, 'Mayaguana', '16', , 4108],
        [542, 'Ragged Island', '18', , 4108],
        [543, 'Harbour Island', '22', , 4108],
        [544, 'New Providence', '23', , 4108],
        [545, 'Acklins and Crooked Islands', '24', , 4108],
        [546, 'Freeport', '25', , 4108],
        [547, 'Fresh Creek', '26', , 4108],
        [548, "Governor's Harbour", '27', , 4108],
        [549, 'Green Turtle Cay', '28', , 4108],
        [550, 'High Rock', '29', , 4108],
        [551, 'Kemps Bay', '30', , 4108],
        [552, 'Marsh Harbour', '31', , 4108],
        [553, 'Nichollstown and Berry Islands', '32', , 4108],
        [554, 'Rock Sound', '33', , 4108],
        [555, 'Sandy Point', '34', , 4108],
        [556, 'San Salvador and Rum Cay', '35', , 4108],
        [557, 'Bumthang', '05', 'BTN-BMT', 4109],
        [558, 'Chhukha', '06', 'BTN-CHH', 4109],
        [559, 'Chirang', '07', 'BTN-CHR', 4109],
        [560, 'Daga', '08', 'BTN-DAG', 4109],
        [561, 'Geylegphug', '09', 'BTN-GYL', 4109],
        [562, 'Ha', '01', , 4297],
        [563, 'Lhuntshi', '11', 'BTN-LHN', 4109],
        [564, 'Mongar', '12', 'BTN-MNG', 4109],
        [565, 'Paro', '13', 'BTN-PAR', 4109],
        [566, 'Pemagatsel', '14', 'BTN-PMG', 4109],
        [567, 'Punakha', '15', 'BTN-PNK', 4109],
        [568, 'Samchi', '16', 'BTN-SMC', 4109],
        [569, 'Samdrup', '17', 'BTN-SDJ', 4109],
        [570, 'Shemgang', '18', 'BTN-SHM', 4109],
        [571, 'Tashigang', '19', 'BTN-TSH', 4109],
        [572, 'Thimphu', '20', 'BTN-THM', 4109],
        [573, 'Tongsa', '21', 'BTN-TNG', 4109],
        [574, 'Wangdi Phodrang', '22', 'BTN-WNP', 4109],
        [575, 'Central', '02', 'ZMB-CNT', 4321],
        [576, 'Ghanzi', '03', 'BWA-GHN', 4111],
        [577, 'Kgalagadi', '04', 'BWA-KGL', 4111],
        [578, 'Kgatleng', '05', 'BWA-KGT', 4111],
        [579, 'Kweneng', '06', 'BWA-KWN', 4111],
        [580, 'North-East', '08', 'BWA-NEA', 4111],
        [581, 'South-East', '09', 'BWA-SEA', 4111],
        [582, 'Southern', '07', 'ZMB-STH', 4321],
        [583, 'North-West', '10', 'ZAF-NWS', 4320],
        [584, "Brestskaya Voblasts'", '01', 'BLR-BRS', 4112],
        [585, "Homyel'skaya Voblasts'", '02', 'BLR-GML', 4112],
        [586, "Hrodzyenskaya Voblasts'", '03', 'BLR-GRD', 4112],
        [587, 'Minsk', '04', 'BLR-MNS', 4112],
        [588, "Minskaya Voblasts'", '05', 'BLR-MNS', 4112],
        [589, "Mahilyowskaya Voblasts'", '06', 'BLR-MGL', 4112],
        [590, "Vitsyebskaya Voblasts'", '07', 'BLR-VTB', 4112],
        [591, 'Belize District', '01', 'BLZ-BLZ', 4113, { aliases: 'Belize' }],
        [592, 'Cayo', '02', 'BLZ-CAY', 4113],
        [593, 'Corozal', '03', 'BLZ-CRZ', 4113],
        [594, 'Orange Walk', '04', 'BLZ-OWA', 4113],
        [595, 'Stann Creek', '05', 'BLZ-SCR', 4113],
        [596, 'Toledo', 'G5', 'PHL-CVS', 4252],
        [597, 'Alberta', '01', 'CAN-ALB', 4114, { code: 'AB' }],
        [598, 'British Columbia', '02', 'CAN-BCL', 4114, { code: 'BC' }],
        [599, 'Manitoba', '03', 'CAN-MNT', 4114, { code: 'MB' }],
        [600, 'New Brunswick', '04', 'CAN-NBR', 4114, { code: 'NB' }],
        [601, 'Newfoundland and Labrador', '05', 'CAN-NWF', 4114, { code: 'NL' }],
        [602, 'Nova Scotia', '07', 'CAN-NSC', 4114, { code: 'NS' }],
        [603, 'Ontario', '08', 'CAN-ONT', 4114, { code: 'ON' }],
        [604, 'Prince Edward Island', '09', 'CAN-PEI', 4114, { code: 'PE' }],
        [605, 'Quebec', '10', 'CAN-QUE', 4114, { code: 'QC' }],
        [606, 'Saskatchewan', '11', 'CAN-SSK', 4114, { code: 'SK' }],
        [607, 'Yukon Territory', '12', 'CAN-YTR', 4114, { code: 'YT' }],
        [608, 'Northwest Territories', '13', 'CAN-NTR', 4114, { code: 'NT' }],
        [609, 'Nunavut', '14', 'CAN-NTR', 4114, { code: 'NU' }],
        [610, 'Bandundu', '01', 'ZAR-BAN', 4116],
        [611, 'Equateur', '02', 'ZAR-EQT', 4116],
        [612, 'Kasai-Oriental', '04', 'ZAR-KOR', 4116],
        [613, 'Katanga', '05', 'ZAR-SHA', 4116, { aliases: 'Shaba' }],
        [614, 'Kinshasa', '06', 'ZAR-KNS', 4116],
        [615, 'Kivu', '07', 'ZAR-KIV', 4116],
        [616, 'Bas-Congo', '08', 'ZAR-BZA', 4116, { aliases: 'Bas-Zaire' }],
        [617, 'Orientale', '09', 'ZAR-HZA', 4116, { aliases: 'Haut-Zaire' }],
        [618, 'Maniema', '10', 'ZAR-KIV', 4116],
        [619, 'Nord-Kivu', '11', 'ZAR-KIV', 4116],
        [620, 'Sud-Kivu', '12', 'ZAR-KIV', 4116],
        [621, 'Cuvette', '13', 'ZAR-EQT', 4116],
        [622, 'Bamingui-Bangoran', '01', 'CAF-BBN', 4117],
        [623, 'Basse-Kotto', '02', 'CAF-BKO', 4117],
        [624, 'Haute-Kotto', '03', 'CAF-HKO', 4117],
        [625, 'Mambere-Kadei', '04', 'CAF-HSN', 4117, { aliases: 'Haute-Sangha' }],
        [626, 'Haut-Mbomou', '05', 'CAF-HMB', 4117],
        [627, 'Kemo', '06', 'CAF-KGR', 4117],
        [628, 'Lobaye', '07', 'CAF-LOB', 4117],
        [629, 'Mbomou', '08', 'CAF-MBO', 4117],
        [630, 'Nana-Mambere', '09', 'CAF-NMM', 4117],
        [631, 'Ouaka', '11', 'CAF-OUA', 4117],
        [632, 'Ouham', '12', 'CAF-OUH', 4117],
        [633, 'Ouham-Pende', '13', 'CAF-OPE', 4117],
        [634, 'Vakaga', '14', 'CAF-VAK', 4117, { aliases: 'Cuvette-Ouest' }],
        [635, 'Nana-Grebizi', '15', 'CAF-GRB', 4117, { aliases: 'Gribingui' }],
        [636, 'Sangha-Mbaere', '16', 'CAF-SNG', 4117],
        [637, 'Ombella-Mpoko', '17', 'CAF-OMP', 4117],
        [638, 'Bangui', '18', 'CAF-OMP', 4117],
        [639, 'Bouenza', '01', 'COG-BOU', 4118],
        [640, 'Cuvette', '03', 'COG-CVT', 4118],
        [641, 'Kouilou', '04', 'COG-KOU', 4118],
        [642, 'Lekoumou', '05', 'COG-LEK', 4118],
        [643, 'Likouala', '06', 'COG-LIK', 4118],
        [644, 'Niari', '07', 'COG-NIA', 4118],
        [645, 'Plateaux', '25', 'TOG-OGO', 4290],
        [646, 'Sangha', '10', 'COG-SNG', 4118],
        [647, 'Pool', '11', 'COG-POO', 4118],
        [648, 'Brazzaville', '12', 'COG-POO', 4118],
        [649, 'Aargau', '01', 'CHE-AAR', 4119],
        [650, 'Ausser-Rhoden', '02', 'CHE-ASR', 4119],
        [651, 'Basel-Landschaft', '03', 'CHE-BLN', 4119],
        [652, 'Basel-Stadt', '04', 'CHE-BST', 4119],
        [653, 'Bern', '05', 'CHE-BER', 4119],
        [654, 'Fribourg', '06', 'CHE-FRB', 4119],
        [655, 'Geneve', '07', 'CHE-GEN', 4119],
        [656, 'Glarus', '08', 'CHE-GLR', 4119],
        [657, 'Graubunden', '09', 'CHE-GRB', 4119],
        [658, 'Inner-Rhoden', '10', 'CHE-AIR', 4119],
        [659, 'Luzern', '11', 'CHE-LZR', 4119],
        [660, 'Neuchatel', '12', 'CHE-NCH', 4119],
        [661, 'Nidwalden', '13', 'CHE-NDW', 4119],
        [662, 'Obwalden', '14', 'CHE-OBW', 4119],
        [663, 'Sankt Gallen', '15', 'CHE-SGL', 4119],
        [664, 'Schaffhausen', '16', 'CHE-SCH', 4119],
        [665, 'Schwyz', '17', 'CHE-SCW', 4119],
        [666, 'Solothurn', '18', 'CHE-SLT', 4119],
        [667, 'Thurgau', '19', 'CHE-THR', 4119],
        [668, 'Ticino', '20', 'CHE-TIC', 4119],
        [669, 'Uri', '21', 'CHE-URI', 4119],
        [670, 'Valais', '22', 'CHE-VAL', 4119],
        [671, 'Vaud', '23', 'CHE-VAU', 4119],
        [672, 'Zug', '24', 'CHE-ZUG', 4119],
        [673, 'Zurich', '25', 'CHE-ZRC', 4119],
        [674, 'Jura', '26', 'CHE-JUR', 4119],
        [675, 'Sassandra', '51', 'CIV-SNP', 4120],
        [676, 'Abidjan', '61', 'CIV-ABD', 4120],
        [677, 'Agneby', '74', 'CIV-ADZ', 4120],
        [678, 'Bafing', '75', 'CIV-TOU', 4120],
        [679, 'Bas-Sassandra', '76', 'CIV-SNP', 4120],
        [680, 'Denguele', '77', 'CIV-ODN', 4120],
        [681, 'Dix-Huit Montagnes', '78', 'CIV-MAN', 4120],
        [682, 'Fromager', '79', 'CIV-SOB', 4120],
        [683, 'Haut-Sassandra', '80', 'CIV-ISS', 4120],
        [684, 'Lacs', '81', 'CIV-YMS', 4120],
        [685, 'Lagunes', '82', 'CIV-ABD', 4120],
        [686, 'Marahoue', '83', 'CIV-DAL', 4120],
        [687, 'Moyen-Cavally', '84', 'CIV-GUI', 4120],
        [688, 'Moyen-Comoe', '85', 'CIV-ABN', 4120],
        [689, "N'zi-Comoe", '86', 'CIV-DKR', 4120],
        [690, 'Savanes', '26', 'TOG-TON', 4290],
        [691, 'Sud-Bandama', '88', 'CIV-SSS', 4120],
        [692, 'Sud-Comoe', '89', 'CIV-ABS', 4120],
        [693, 'Vallee du Bandama', '90', 'CIV-DBK', 4120],
        [694, 'Worodougou', '91', 'CIV-MNK', 4120],
        [695, 'Zanzan', '92', 'CIV-BON', 4120],
        [696, 'Valparaiso', '01', 'CHL-VLP', 4122],
        [697, 'Aisen del General Carlos Ibanez del Campo', '02', 'CHL-ADG', 4122],
        [698, 'Antofagasta', '03', 'CHL-ANT', 4122],
        [699, 'Araucania', '04', 'CHL-ARC', 4122],
        [700, 'Atacama', '05', 'CHL-ATC', 4122],
        [701, 'Bio-Bio', '06', 'CHL-BIO', 4122],
        [702, 'Coquimbo', '07', 'CHL-CQM', 4122],
        [703, "Libertador General Bernardo O'Higgins", '08', 'CHL-LGB', 4122],
        [704, 'Los Lagos', '09', 'CHL-LLA', 4122],
        [705, 'Magallanes y de la Antartica Chilena', '10', 'CHL-MYA', 4122],
        [706, 'Maule', '11', 'CHL-MAU', 4122],
        [707, 'Region Metropolitana', '12', 'CHL-RMT', 4122],
        [708, 'Tarapaca', '13', 'CHL-TRP', 4122],
        [709, 'Est', '11', 'RWA-KBN', 4267],
        [710, 'Nord-Ouest', '03', 'HTI-NOU', 4175],
        [711, 'Ouest', '14', 'RWA-GSN', 4267],
        [712, 'Sud-Ouest', '09', 'CMR-SOU', 4123],
        [713, 'Adamaoua', '10', 'CMR-ADM', 4123],
        [714, 'Centre', '07', 'HTI-CNT', 4175],
        [715, 'Extreme-Nord', '12', 'CMR-ENO', 4123],
        [716, 'Nord', '13', 'RWA-RHN', 4267],
        [717, 'Sud', '15', 'RWA-GKN', 4267],
        [718, 'Anhui', '01', 'CHN-ANH', 4124],
        [719, 'Zhejiang', '02', 'CHN-ZHJ', 4124],
        [720, 'Jiangxi', '03', 'CHN-JNG', 4124],
        [721, 'Jiangsu', '04', 'CHN-JNS', 4124],
        [722, 'Jilin', '05', 'CHN-JIL', 4124],
        [723, 'Qinghai', '06', 'CHN-QNG', 4124],
        [724, 'Fujian', '07', 'CHN-FUJ', 4124],
        [725, 'Heilongjiang', '08', 'CHN-HLN', 4124],
        [726, 'Henan', '09', 'CHN-HEN', 4124],
        [727, 'Hebei', '10', 'CHN-HEB', 4124],
        [728, 'Hunan', '11', 'CHN-HUN', 4124],
        [729, 'Hubei', '12', 'CHN-HUB', 4124],
        [730, 'Xinjiang', '13', 'CHN-XNJ', 4124],
        [731, 'Xizang', '14', 'CHN-XZN', 4124],
        [732, 'Gansu', '15', 'CHN-GAN', 4124],
        [733, 'Guangxi', '16', 'CHN-GNG', 4124],
        [734, 'Guizhou', '18', 'CHN-GUI', 4124],
        [735, 'Liaoning', '19', 'CHN-LNN', 4124],
        [736, 'Nei Mongol', '20', 'CHN-NMN', 4124],
        [737, 'Ningxia', '21', 'CHN-NNG', 4124],
        [738, 'Beijing', '22', 'CHN-BJN', 4124],
        [739, 'Shanghai', '23', 'CHN-SHN', 4124],
        [740, 'Shanxi', '24', 'CHN-SHX', 4124],
        [741, 'Shandong', '25', 'CHN-SHD', 4124],
        [742, 'Shaanxi', '26', 'CHN-SHA', 4124],
        [743, 'Tianjin', '28', 'CHN-TNJ', 4124],
        [744, 'Yunnan', '29', 'CHN-YNN', 4124],
        [745, 'Guangdong', '30', 'CHN-GND', 4124],
        [746, 'Hainan', '31', 'CHN-HAI', 4124],
        [747, 'Sichuan', '32', 'CHN-SCH', 4124],
        [748, 'Chongqing', '33', 'CHN-SCH', 4124],
        [749, 'Antioquia', '02', 'COL-ANT', 4125],
        [750, 'Arauca', '03', 'COL-ARA', 4125],
        [751, 'Atlantico', '04', 'COL-ATL', 4125],
        [752, 'Bolivar Department', '05', 'COL-BLV', 4125],
        [753, 'Boyaca Department', '06', 'COL-BOY', 4125],
        [754, 'Caldas Department', '07', 'COL-CLD', 4125],
        [755, 'Caqueta', '08', 'COL-CAQ', 4125],
        [756, 'Cauca', '09', 'COL-CAU', 4125],
        [757, 'Cesar', '10', 'COL-CES', 4125],
        [758, 'Choco', '11', 'COL-CHO', 4125],
        [759, 'Guaviare', '14', 'COL-GUA', 4125],
        [760, 'Guainia', '15', 'COL-GUI', 4125],
        [761, 'La Guajira', '17', 'COL-LGU', 4125],
        [762, 'Magdalena Department', '18', 'COL-MGD', 4125],
        [763, 'Meta', '19', 'COL-MET', 4125],
        [764, 'Narino', '20', 'COL-NAR', 4125],
        [765, 'Norte de Santander', '21', 'COL-NDS', 4125],
        [766, 'Putumayo', '22', 'COL-PTM', 4125],
        [767, 'Quindio', '23', 'COL-QUI', 4125],
        [768, 'Risaralda', '24', 'COL-RSR', 4125],
        [769, 'San Andres y Providencia', '25', 'COL-SAY', 4125],
        [770, 'Santander', '26', 'COL-SNT', 4125],
        [771, 'Sucre', '19', 'VEN-SUC', 4311],
        [772, 'Tolima', '28', 'COL-TOL', 4125],
        [773, 'Valle del Cauca', '29', 'COL-VDC', 4125],
        [774, 'Vaupes', '30', 'COL-VAU', 4125],
        [775, 'Vichada', '31', 'COL-VCH', 4125],
        [776, 'Casanare', '32', 'COL-CSN', 4125],
        [777, 'Cundinamarca', '33', 'COL-CND', 4125],
        [778, 'Distrito Especial', '34', 'COL-DES', 4125],
        [779, 'Bolivar', '06', 'VEN-BLV', 4311],
        [780, 'Boyaca', '36', 'COL-BOY', 4125],
        [781, 'Caldas', '37', 'COL-CLD', 4125],
        [782, 'Magdalena', '38', 'COL-MGD', 4125],
        [783, 'Alajuela', '01', 'CRI-ALJ', 4126],
        [784, 'Cartago', '02', 'CRI-CRT', 4126],
        [785, 'Guanacaste', '03', 'CRI-GNC', 4126],
        [786, 'Heredia', '04', 'CRI-HER', 4126],
        [787, 'Limon', '06', 'CRI-LIM', 4126],
        [788, 'Puntarenas', '07', 'CRI-PNT', 4126],
        [789, 'San Jose', '16', 'URY-SJO', 4307],
        [790, 'Pinar del Rio', '01', 'CUB-PDR', 4127],
        [791, 'Ciudad de la Habana', '02', 'CUB-CHB', 4127],
        [792, 'Matanzas', '03', 'CUB-MTN', 4127],
        [793, 'Isla de la Juventud', '04', 'CUB-IDL', 4127],
        [794, 'Camaguey', '05', 'CUB-CMG', 4127],
        [795, 'Ciego de Avila', '07', 'CUB-CDA', 4127],
        [796, 'Cienfuegos', '08', 'CUB-CNF', 4127],
        [797, 'Granma', '09', 'CUB-GRN', 4127],
        [798, 'Guantanamo', '10', 'CUB-GNT', 4127],
        [799, 'La Habana', '11', 'CUB-LHB', 4127],
        [800, 'Holguin', '12', 'CUB-HLG', 4127],
        [801, 'Las Tunas', '13', 'CUB-LTU', 4127],
        [802, 'Sancti Spiritus', '14', 'CUB-SSP', 4127],
        [803, 'Santiago de Cuba', '15', 'CUB-SDC', 4127],
        [804, 'Villa Clara', '16', 'CUB-VCL', 4127],
        [805, 'Boa Vista', '01', , 4128],
        [806, 'Brava', '02', , 4128],
        [807, 'Maio', '04', , 4128],
        [808, 'Paul', '05', , 4128],
        [809, 'Ribeira Grande', '07', , 4128],
        [810, 'Sal', '08', , 4128],
        [811, 'Sao Nicolau', '10', , 4128],
        [812, 'Sao Vicente', '11', , 4128],
        [813, 'Mosteiros', '13', , 4128],
        [814, 'Praia', '14', , 4128],
        [815, 'Sao Domingos', '17', , 4128],
        [816, 'Sao Filipe', '18', , 4128],
        [817, 'Sao Miguel', '19', , 4128],
        [818, 'Tarrafal', '20', , 4128],
        [819, 'Famagusta', '01', 'CYP-FMG', 4130],
        [820, 'Kyrenia', '02', 'CYP-KYR', 4130],
        [821, 'Larnaca', '03', 'CYP-LRN', 4130],
        [822, 'Nicosia', '04', 'CYP-NIC', 4130],
        [823, 'Limassol', '05', 'CYP-LMS', 4130],
        [824, 'Paphos', '06', 'CYP-PPH', 4130],
        [825, 'Blansko', '03', 'CZE-JHM', 4131],
        [826, 'Breclav', '04', 'CZE-JHM', 4131],
        [827, 'Hradec Kralove', '20', 'CZE-VYC', 4131],
        [828, 'Jablonec nad Nisou', '21', 'CZE-SVC', 4131],
        [829, 'Jicin', '23', 'CZE-VYC', 4131],
        [830, 'Jihlava', '24', 'CZE-JHM', 4131],
        [831, 'Kolin', '30', 'CZE-STR', 4131],
        [832, 'Liberec', '33', 'CZE-SVC', 4131],
        [833, 'Melnik', '36', 'CZE-STR', 4131],
        [834, 'Mlada Boleslav', '37', 'CZE-STR', 4131],
        [835, 'Nachod', '39', 'CZE-VYC', 4131],
        [836, 'Nymburk', '41', 'CZE-STR', 4131],
        [837, 'Pardubice', '45', 'CZE-VYC', 4131],
        [838, 'Hlavni mesto Praha', '52', 'CZE-STR', 4131, { aliases: 'Prague' }],
        [839, 'Semily', '61', 'CZE-VYC', 4131],
        [840, 'Trutnov', '70', 'CZE-VYC', 4131],
        [841, 'Jihomoravsky kraj', '78', 'CZE-JHM', 4131],
        [842, 'Jihocesky kraj', '79', 'CZE-JHC', 4131],
        [843, 'Vysocina', '80', 'CZE-JHM', 4131],
        [844, 'Karlovarsky kraj', '81', 'CZE-ZPD', 4131],
        [845, 'Kralovehradecky kraj', '82', 'CZE-VYC', 4131],
        [846, 'Liberecky kraj', '83', 'CZE-SVC', 4131],
        [847, 'Olomoucky kraj', '84', 'CZE-SVM', 4131],
        [848, 'Moravskoslezsky kraj', '85', 'CZE-SVM', 4131],
        [849, 'Pardubicky kraj', '86', 'CZE-VYC', 4131],
        [850, 'Plzensky kraj', '87', 'CZE-ZPD', 4131],
        [851, 'Stredocesky kraj', '88', 'CZE-STR', 4131],
        [852, 'Ustecky kraj', '89', 'CZE-SVC', 4131],
        [853, 'Zlinsky kraj', '90', 'CZE-JHM', 4131],
        [854, 'Baden-Wurttemberg', '01', 'DEU-BWR', 4132],
        [855, 'Bayern', '02', 'DEU-BYR', 4132],
        [856, 'Bremen', '03', 'DEU-BRM', 4132],
        [857, 'Hamburg', '04', 'DEU-HMB', 4132],
        [858, 'Hessen', '05', 'DEU-HSS', 4132],
        [859, 'Niedersachsen', '06', 'DEU-NDR', 4132],
        [860, 'Nordrhein-Westfalen', '07', 'DEU-NWS', 4132],
        [861, 'Rheinland-Pfalz', '08', 'DEU-RPF', 4132],
        [862, 'Saarland', '09', 'DEU-SRL', 4132],
        [863, 'Schleswig-Holstein', '10', 'DEU-SHL', 4132],
        [864, 'Brandenburg', '11', 'DEU-BRN', 4132],
        [865, 'Mecklenburg-Vorpommern', '12', 'DEU-MKV', 4132],
        [866, 'Sachsen', '13', 'DEU-SCH', 4132],
        [867, 'Sachsen-Anhalt', '14', 'DEU-SCA', 4132],
        [868, 'Thuringen', '15', 'DEU-THR', 4132],
        [869, 'Berlin', '16', 'DEU-BER', 4132],
        [870, 'Ali Sabieh', '01', 'DJI-ASA', 4133],
        [871, 'Obock', '04', 'DJI-OBC', 4133],
        [872, 'Tadjoura', '05', 'DJI-TDJ', 4133],
        [873, 'Dikhil', '06', 'DJI-DKH', 4133],
        [874, 'Djibouti', '07', 'DJI-DJI', 4133],
        [875, 'Arta', '20', 'GRC-IPR', 4165, { region: 4394 }],
        [876, 'Arhus', '01', 'DNK-ARH', 4134],
        [877, 'Bornholm', '02', 'DNK-BRN', 4134],
        [878, 'Frederiksborg', '03', 'DNK-FRD', 4134],
        [879, 'Fyn', '04', 'DNK-FYN', 4134],
        [880, 'Kobenhavn', '05', 'DNK-KBN', 4134],
        [881, 'Staden Kobenhavn', '06', 'DNK-SKB', 4134],
        [882, 'Nordjylland', '07', 'DNK-NRD', 4134],
        [883, 'Ribe', '08', 'DNK-RIB', 4134],
        [884, 'Ringkobing', '09', 'DNK-RNG', 4134],
        [885, 'Roskilde', '10', 'DNK-RSK', 4134],
        [886, 'Sonderjylland', '11', 'DNK-SND', 4134],
        [887, 'Storstrom', '12', 'DNK-STR', 4134],
        [888, 'Vejle', '13', 'DNK-VEJ', 4134],
        [889, 'Vestsjalland', '14', 'DNK-VST', 4134],
        [890, 'Viborg', '15', 'DNK-VBR', 4134],
        [
            891,
            'Hovedstaden',
            '17',
            'DNK-FRD',
            4134,
            { notes: 'Region of 02, 03, 05, 06' },
        ],
        [
            892,
            'Midtjyllen',
            '18',
            'DNK-ARH',
            4134,
            { notes: 'Region of 01, 09, southern part of 15' },
        ],
        [893, 'Nordjylland', '19', 'DNK-NRD', 4134],
        [894, 'Sjelland', '20', 'DNK-VST', 4134, { notes: 'Region of 10, 12, 14' }],
        [
            895,
            'Syddanmark',
            '21',
            'DNK-FYN',
            4134,
            { notes: 'Region of 04, 08, 11, 13' },
        ],
        [896, 'Saint David', '03', , 4310],
        [897, 'Saint Luke', '07', , 4135],
        [898, 'Saint Mark', '05', , 4154],
        [899, 'Saint Patrick', '05', , 4310],
        [900, 'Azua', '01', , 4136],
        [901, 'Baoruco', '02', , 4136],
        [902, 'Barahona', '03', , 4136],
        [903, 'Dajabon', '04', , 4136],
        [904, 'Distrito Nacional', '05', , 4136],
        [905, 'Duarte', '06', , 4136],
        [906, 'Espaillat', '08', , 4136],
        [907, 'Independencia', '09', , 4136],
        [908, 'La Altagracia', '10', , 4136],
        [909, 'Elias Pina', '11', , 4136],
        [910, 'La Romana', '12', , 4136],
        [911, 'Maria Trinidad Sanchez', '14', , 4136],
        [912, 'Monte Cristi', '15', , 4136],
        [913, 'Pedernales', '16', , 4136],
        [914, 'Peravia', '17', , 4136],
        [915, 'Puerto Plata', '18', , 4136],
        [916, 'Salcedo', '19', , 4136],
        [917, 'Samana', '20', , 4136],
        [918, 'Sanchez Ramirez', '21', , 4136],
        [919, 'San Pedro De Macoris', '24', , 4136],
        [920, 'Santiago', '25', , 4136],
        [921, 'Santiago Rodriguez', '26', , 4136],
        [922, 'Valverde', '27', , 4136],
        [923, 'El Seibo', '28', , 4136],
        [924, 'Hato Mayor', '29', , 4136],
        [925, 'La Vega', '30', , 4136],
        [926, 'Monsenor Nouel', '31', , 4136],
        [927, 'Monte Plata', '32', , 4136],
        [928, 'San Cristobal', '33', , 4136],
        [929, 'Distrito Nacional', '34', , 4136],
        [930, 'Peravia', '35', , 4136],
        [931, 'San Jose de Ocoa', '36', , 4136],
        [932, 'Santo Domingo', '37', , 4136],
        [933, 'Alger', '01', 'DZA-ALG', 4137],
        [934, 'Batna', '03', 'DZA-BAT', 4137],
        [935, 'Constantine', '04', 'DZA-CNS', 4137],
        [936, 'Medea', '06', 'DZA-MED', 4137],
        [937, 'Mostaganem', '07', 'DZA-MST', 4137],
        [938, 'Oran', '09', 'DZA-ORA', 4137],
        [939, 'Saida', '10', 'DZA-SAI', 4137],
        [940, 'Setif', '12', 'DZA-SET', 4137],
        [941, 'Tiaret', '13', 'DZA-TIA', 4137],
        [942, 'Tizi Ouzou', '14', 'DZA-TOU', 4137],
        [943, 'Tlemcen', '15', 'DZA-TLM', 4137],
        [944, 'Bejaia', '18', 'DZA-BEJ', 4137],
        [945, 'Biskra', '19', 'DZA-BSK', 4137],
        [946, 'Blida', '20', 'DZA-BLI', 4137],
        [947, 'Bouira', '21', 'DZA-BOU', 4137],
        [948, 'Djelfa', '22', 'DZA-DJL', 4137],
        [949, 'Guelma', '23', 'DZA-GUE', 4137],
        [950, 'Jijel', '24', 'DZA-JIJ', 4137],
        [951, 'Laghouat', '25', 'DZA-LGH', 4137],
        [952, 'Mascara', '26', 'DZA-MSC', 4137],
        [953, "M'sila", '27', 'DZA-MSL', 4137],
        [954, 'Oum el Bouaghi', '29', 'DZA-OEB', 4137],
        [955, 'Sidi Bel Abbes', '30', 'DZA-SBA', 4137],
        [956, 'Skikda', '31', 'DZA-SKK', 4137],
        [957, 'Tebessa', '33', 'DZA-TBS', 4137],
        [958, 'Adrar', '07', 'MRT-ADR', 4226],
        [959, 'Ain Defla', '35', 'DZA-ADF', 4137],
        [960, 'Ain Temouchent', '36', 'DZA-ATM', 4137],
        [961, 'Annaba', '37', 'DZA-ANN', 4137],
        [962, 'Bechar', '38', 'DZA-BCH', 4137],
        [963, 'Bordj Bou Arreridj', '39', 'DZA-BBA', 4137],
        [964, 'Boumerdes', '40', 'DZA-BMR', 4137],
        [965, 'Chlef', '41', 'DZA-CHL', 4137],
        [966, 'El Bayadh', '42', 'DZA-EBY', 4137],
        [967, 'El Oued', '43', 'DZA-EOD', 4137],
        [968, 'El Tarf', '44', 'DZA-ETR', 4137],
        [969, 'Ghardaia', '45', 'DZA-GHD', 4137],
        [970, 'Illizi', '46', 'DZA-ILL', 4137],
        [971, 'Khenchela', '47', 'DZA-KHN', 4137],
        [972, 'Mila', '48', 'DZA-MIL', 4137],
        [973, 'Naama', '49', 'DZA-NAM', 4137],
        [974, 'Ouargla', '50', 'DZA-ORG', 4137],
        [975, 'Relizane', '51', 'DZA-RLZ', 4137],
        [976, 'Souk Ahras', '52', 'DZA-SAH', 4137],
        [977, 'Tamanghasset', '53', 'DZA-TMN', 4137],
        [978, 'Tindouf', '54', 'DZA-TND', 4137],
        [979, 'Tipaza', '55', 'DZA-TPZ', 4137],
        [980, 'Tissemsilt', '56', 'DZA-TSS', 4137],
        [981, 'Galapagos', '01', 'ECU-GLP', 4138],
        [982, 'Azuay', '02', 'ECU-AZU', 4138],
        [983, 'Canar', '04', 'ECU-CAN', 4138],
        [984, 'Carchi', '05', 'ECU-CRC', 4138],
        [985, 'Chimborazo', '06', 'ECU-CHM', 4138],
        [986, 'Cotopaxi', '07', 'ECU-CTP', 4138],
        [987, 'El Oro', '08', 'ECU-EOR', 4138],
        [988, 'Esmeraldas', '09', 'ECU-ESM', 4138],
        [989, 'Guayas', '10', 'ECU-GUA', 4138],
        [990, 'Imbabura', '11', 'ECU-IMB', 4138],
        [991, 'Loja', '12', 'ECU-LOJ', 4138],
        [992, 'Los Rios', '13', 'ECU-LRI', 4138],
        [993, 'Manabi', '14', 'ECU-MAN', 4138],
        [994, 'Morona-Santiago', '15', 'ECU-MSN', 4138],
        [995, 'Pastaza', '17', 'ECU-PST', 4138],
        [996, 'Pichincha', '18', 'ECU-PCH', 4138],
        [997, 'Tungurahua', '19', 'ECU-TNG', 4138],
        [998, 'Zamora-Chinchipe', '20', 'ECU-ZCH', 4138],
        [999, 'Sucumbios', '22', 'ECU-SCM', 4138],
        [1000, 'Napo', '23', 'ECU-NAP', 4138],
        [1001, 'Orellana', '24', 'ECU-NAP', 4138],
        [1002, 'Harjumaa', '01', , 4139],
        [1003, 'Hiiumaa', '02', , 4139],
        [1004, 'Ida-Virumaa', '03', , 4139],
        [1005, 'Jarvamaa', '04', , 4139],
        [1006, 'Jogevamaa', '05', , 4139],
        [1007, 'Kohtla-Jarve', '06', , 4139],
        [1008, 'Laanemaa', '07', , 4139],
        [1009, 'Laane-Virumaa', '08', , 4139],
        [1010, 'Narva', '09', , 4139],
        [1011, 'Parnu', '10', , 4139],
        [1012, 'Parnumaa', '11', , 4139],
        [1013, 'Polvamaa', '12', , 4139],
        [1014, 'Raplamaa', '13', , 4139],
        [1015, 'Saaremaa', '14', , 4139],
        [1016, 'Sillamae', '15', , 4139],
        [1017, 'Tallinn', '16', , 4139],
        [1018, 'Tartu', '17', , 4139],
        [1019, 'Tartumaa', '18', , 4139],
        [1020, 'Valgamaa', '19', , 4139],
        [1021, 'Viljandimaa', '20', , 4139],
        [1022, 'Vorumaa', '21', , 4139],
        [1023, 'Ad Daqahliyah', '01', 'EGY-ADQ', 4140],
        [1024, 'Al Bahr al Ahmar', '02', 'EGY-ABA', 4140],
        [1025, 'Al Buhayrah', '03', 'EGY-ABH', 4140],
        [1026, 'Al Fayyum', '04', 'EGY-AFY', 4140],
        [1027, 'Al Gharbiyah', '05', 'EGY-AGH', 4140],
        [1028, 'Al Iskandariyah', '06', 'EGY-AIS', 4140],
        [1029, "Al Isma'iliyah", '07', 'EGY-AIM', 4140],
        [1030, 'Al Jizah', '08', 'EGY-AJI', 4140],
        [1031, 'Al Minufiyah', '09', 'EGY-AMN', 4140],
        [1032, 'Al Minya', '10', 'EGY-AMI', 4140],
        [1033, 'Al Qahirah', '11', 'EGY-AQH', 4140],
        [1034, 'Al Qalyubiyah', '12', 'EGY-AQL', 4140],
        [1035, 'Al Wadi al Jadid', '13', 'EGY-AWA', 4140],
        [1036, 'Ash Sharqiyah', '31', 'SDN-ASR', 4271],
        [1037, 'As Suways', '15', 'EGY-ASW', 4140],
        [1038, 'Aswan', '16', 'EGY-ASN', 4140],
        [1039, 'Asyut', '17', 'EGY-ASY', 4140],
        [1040, 'Bani Suwayf', '18', 'EGY-BSW', 4140],
        [1041, "Bur Sa'id", '19', 'EGY-BSA', 4140],
        [1042, 'Dumyat', '20', 'EGY-DMY', 4140],
        [1043, 'Kafr ash Shaykh', '21', 'EGY-KAS', 4140],
        [1044, 'Matruh', '22', 'EGY-MNT', 4140],
        [1045, 'Qina', '23', 'EGY-QIN', 4140],
        [1046, 'Suhaj', '24', 'EGY-SUH', 4140],
        [1047, "Janub Sina'", '26', 'EGY-JSI', 4140],
        [1048, "Shamal Sina'", '27', 'EGY-SSI', 4140],
        [1049, 'Anseba', '01', 'ERI-BRK', 4142],
        [1050, 'Debub', '02', 'ERI-SRY', 4142],
        [1051, "Debubawi K'eyih Bahri", '03', 'ERI-DNK', 4142],
        [1052, 'Gash Barka', '04', 'ERI-BRK', 4142],
        [1053, "Ma'akel", '05', 'ERI-HMS', 4142],
        [1054, "Semenawi K'eyih Bahri", '06', 'ERI-HMS', 4142],
        [1055, 'Islas Baleares', '07', 'ESP-BLR', 4143],
        [1056, 'Madrid', '29', 'ESP-MDR', 4143],
        [1057, 'Murcia', '31', 'ESP-MUR', 4143],
        [1058, 'Navarra', '32', 'ESP-NVR', 4143],
        [1059, 'Asturias', '34', 'ESP-AST', 4143],
        [1060, 'Cantabria', '39', 'ESP-CNT', 4143],
        [1061, 'Andalucia', '51', 'ESP-AND', 4143],
        [1062, 'Aragon', '52', 'ESP-ARG', 4143],
        [
            1063,
            'Canarias',
            '53',
            'ESP-AND',
            4143,
            {
                aliases: 'Canary Islands',
                notes: 'This is not really in Andalucia, it is off-the-map, so is included in the nearest land area.',
            },
        ],
        [1064, 'Castilla-La Mancha', '54', 'ESP-CLM', 4143],
        [1065, 'Castilla y Leon', '55', 'ESP-CYL', 4143],
        [1066, 'Catalonia', '56', 'ESP-CTL', 4143],
        [1067, 'Extremadura', '57', 'ESP-EXT', 4143],
        [1068, 'Galicia', '58', 'ESP-GLC', 4143],
        [1069, 'Pais Vasco', '59', 'ESP-PSV', 4143],
        [1070, 'Comunidad Valenciana', '60', 'ESP-VLN', 4143],
        [1071, 'Amhara', '02', 'ETH-GND', 4144],
        [1072, 'Somali', '07', 'ETH-HRR', 4144],
        [1073, 'Gambella', '08', 'ETH-ILB', 4144],
        [1074, 'Addis Abeba', '10', 'ETH-SHE', 4144],
        [1075, 'Tigray', '12', 'ETH-TGR', 4144],
        [1076, 'Benishangul', '13', 'ETH-WEL', 4144],
        [1077, 'Afar', '14', 'ETH-WEO', 4144],
        [1078, 'Adis Abeba', '44', 'ETH-SHE', 4144],
        [1079, 'Afar', '45', 'ETH-WEO', 4144],
        [1080, 'Amara', '46', 'ETH-SHE', 4144],
        [1081, 'Binshangul Gumuz', '47', 'ETH-WEL', 4144],
        [1082, 'Dire Dawa', '48', 'ETH-HRR', 4144],
        [1083, 'Gambela Hizboch', '49', 'ETH-ILB', 4144],
        [1084, 'Hareri Hizb', '50', 'ETH-HRR', 4144],
        [1085, 'Oromiya', '51', 'ETH-SID', 4144],
        [1086, 'Sumale', '52', 'ETH-HRR', 4144],
        [1087, 'Tigray', '53', 'ETH-TGR', 4144],
        [
            1088,
            'YeDebub Biheroch Bihereseboch na Hizboch',
            '54',
            'ETH-KEF',
            4144,
            { aliases: "Southern Nations, Nationalities, and People's Region" },
        ],
        [1089, 'Aland', '01', 'FIN-AHR', 4146],
        [1090, 'Lapland', '06', 'FIN-LAP', 4146],
        [1091, 'Oulu', '08', 'FIN-OUL', 4146],
        [
            1092,
            'Southern Finland',
            '13',
            'FIN-UUS',
            4146,
            { notes: 'Region of Kymi, Uusimaa' },
        ],
        [
            1093,
            'Eastern Finland',
            '14',
            'FIN-PKR',
            4146,
            { notes: 'Region of Kuopio, Mikkeli, and Pohjois-Karjala' },
        ],
        [
            1094,
            'Western Finland',
            '15',
            'FIN-VAA',
            4146,
            { notes: 'Region of Hame, Keski-Suomi, Turku Ja Pori, and Vaasa' },
        ],
        [1095, 'Eastern', '03', 'ZMB-EST', 4321],
        [1096, 'Northern', '05', 'ZMB-NRT', 4321],
        [1097, 'Rotuma', '04', , 4147],
        [1098, 'Western', '01', 'ZMB-WST', 4321],
        [1099, 'Kosrae', '01', , 4149],
        [1100, 'Pohnpei', '02', , 4149],
        [1101, 'Chuuk', '03', , 4149],
        [1102, 'Yap', '04', , 4149],
        [1103, 'Aquitaine', '97', 'FRA-AQT', 4151],
        [1104, 'Auvergne', '98', 'FRA-AVR', 4151],
        [1105, 'Basse-Normandie', '99', 'FRA-BNR', 4151],
        [1106, 'Bourgogne', 'A1', 'FRA-BOR', 4151],
        [1107, 'Bretagne', 'A2', 'FRA-BRT', 4151],
        [1108, 'Champagne-Ardenne', 'A4', 'FRA-CHA', 4151],
        [1109, 'Corse', 'A5', 'FRA-COR', 4151],
        [1110, 'Franche-Comte', 'A6', 'FRA-FRC', 4151],
        [1111, 'Haute-Normandie', 'A7', 'FRA-HNR', 4151],
        [1112, 'Ile-de-France', 'A8', 'FRA-IDF', 4151],
        [1113, 'Languedoc-Roussillon', 'A9', 'FRA-LNR', 4151],
        [1114, 'Limousin', 'B1', 'FRA-LIM', 4151],
        [1115, 'Lorraine', 'B2', 'FRA-LOR', 4151],
        [1116, 'Midi-Pyrenees', 'B3', 'FRA-MPY', 4151],
        [1117, 'Nord-Pas-de-Calais', 'B4', 'FRA-NPC', 4151],
        [1118, 'Pays de la Loire', 'B5', 'FRA-PDL', 4151],
        [1119, 'Picardie', 'B6', 'FRA-PIC', 4151],
        [1120, 'Poitou-Charentes', 'B7', 'FRA-PCH', 4151],
        [1121, "Provence-Alpes-Cote d'Azur", 'B8', 'FRA-PAC', 4151],
        [1122, 'Rhone-Alpes', 'B9', 'FRA-RHA', 4151],
        [1123, 'Alsace', 'C1', 'FRA-ALS', 4151],
        [1124, 'Estuaire', '01', 'GAB-EST', 4152],
        [1125, 'Haut-Ogooue', '02', 'GAB-HOG', 4152],
        [1126, 'Moyen-Ogooue', '03', 'GAB-MOG', 4152],
        [1127, 'Ngounie', '04', 'GAB-NGO', 4152],
        [1128, 'Nyanga', '05', 'GAB-NYN', 4152],
        [1129, 'Ogooue-Ivindo', '06', 'GAB-OIV', 4152],
        [1130, 'Ogooue-Lolo', '07', 'GAB-OLO', 4152],
        [1131, 'Ogooue-Maritime', '08', 'GAB-OMR', 4152],
        [1132, 'Woleu-Ntem', '09', 'GAB-WNT', 4152],
        [
            1133,
            'Avon',
            '01',
            'GBR-ENG-AVO',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1134,
            'Berkshire',
            '03',
            'GBR-ENG-BRK',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1135,
            'Cleveland',
            '07',
            'GBR-ENG-CLV',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1136,
            'Greater London',
            '17',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1137,
            'Greater Manchester',
            '18',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1138,
            'Hereford and Worcester',
            '20',
            'GBR-ENG-HAW',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1139,
            'Humberside',
            '22',
            'GBR-ENG-HMB',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1140,
            'Merseyside',
            '28',
            'GBR-ENG-MRS',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1141,
            'South Yorkshire',
            '37',
            'GBR-ENG-SYR',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1142,
            'Tyne and Wear',
            '41',
            'GBR-ENG-TAW',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1143,
            'West Midlands',
            '43',
            'GBR-ENG-WMD',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1144,
            'West Yorkshire',
            '45',
            'GBR-ENG-WYR',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [1145, 'Grampian', '82', 'GBR-SCT-GRM', 4153, { region: 4347 }],
        [1146, 'Lothian', '84', 'GBR-SCT-LTH', 4153, { region: 4347 }],
        [1147, 'Strathclyde', '87', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1148, 'Tayside', '88', 'GBR-SCT-TYS', 4153, { region: 4347 }],
        [1149, 'Clwyd', '90', 'GBR-WLS-CLW', 4153, { region: 4348 }],
        [1150, 'Dyfed', '91', 'GBR-WLS-DYF', 4153, { region: 4348 }],
        [1151, 'Gwent', '92', 'GBR-WLS-GWN', 4153, { region: 4348 }],
        [1152, 'Mid Glamorgan', '94', 'GBR-WLS-MGL', 4153, { region: 4348 }],
        [1153, 'South Glamorgan', '96', 'GBR-WLS-SGL', 4153, { region: 4348 }],
        [1154, 'West Glamorgan', '97', 'GBR-WLS-WGL', 4153, { region: 4348 }],
        [
            1155,
            'Barking and Dagenham',
            'A1',
            'GBR-ENG-GLN',
            4153,
            { region: 4348, notes: 'Unitary District' },
        ],
        [
            1156,
            'Barnet',
            'A2',
            'GBR-ENG-GLN',
            4153,
            { region: 4348, notes: 'Unitary District' },
        ],
        [
            1157,
            'Barnsley',
            'A3',
            'GBR-ENG-SYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1158,
            'Bath and North East Somerset',
            'A4',
            'GBR-ENG-SMR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1159,
            'Bedfordshire',
            'A5',
            'GBR-ENG-BDF',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1160,
            'Bexley',
            'A6',
            'GBR-ENG-GLN',
            4153,
            { region: 4348, notes: 'Unitary District' },
        ],
        [
            1161,
            'Birmingham',
            'A7',
            'GBR-ENG-WMD',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1162,
            'Blackburn with Darwen',
            'A8',
            'GBR-ENG-LNC',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1163,
            'Blackpool',
            'A9',
            'GBR-ENG-LNC',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1164,
            'Bolton',
            'B1',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1165,
            'Bournemouth',
            'B2',
            'GBR-ENG-DRS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1166,
            'Bracknell Forest',
            'B3',
            'GBR-ENG-BRK',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1167,
            'Bradford',
            'B4',
            'GBR-ENG-WYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1168,
            'Brent',
            'B5',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1169,
            'Brighton and Hove',
            'B6',
            'GBR-ENG-ESS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1170,
            'Bristol, City of',
            'B7',
            'GBR-ENG-AVO',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1171,
            'Bromley',
            'B8',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1172,
            'Buckinghamshire',
            'B9',
            'GBR-ENG-BCK',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1173,
            'Bury',
            'C1',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1174,
            'Calderdale',
            'C2',
            'GBR-ENG-WYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1175,
            'Cambridgeshire',
            'C3',
            'GBR-ENG-CMB',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1176,
            'Camden',
            'C4',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1177,
            'Cheshire',
            'C5',
            'GBR-ENG-CHS',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1178,
            'Cornwall',
            'C6',
            'GBR-ENG-CRN',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1179,
            'Coventry',
            'C7',
            'GBR-ENG-WMD',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1180,
            'Croydon',
            'C8',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1181,
            'Cumbria',
            'C9',
            'GBR-ENG-CMR',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1182,
            'Darlington',
            'D1',
            'GBR-ENG-DRH',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1183,
            'Derby',
            'D2',
            'GBR-ENG-DRB',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1184,
            'Derbyshire',
            'D3',
            'GBR-ENG-DRB',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1185,
            'Devon',
            'D4',
            'GBR-ENG-DEV',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1186,
            'Doncaster',
            'D5',
            'GBR-ENG-SYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1187,
            'Dorset',
            'D6',
            'GBR-ENG-DRS',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1188,
            'Dudley',
            'D7',
            'GBR-ENG-WMD',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1189,
            'Durham',
            'D8',
            'GBR-ENG-DRH',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1190,
            'Ealing',
            'D9',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1191,
            'East Riding of Yorkshire',
            'E1',
            'GBR-ENG-HMB',
            4153,
            {
                region: 4345,
                aliases: 'North Wolds } ],East Yorkshire',
                notes: 'Unitary District',
            },
        ],
        [
            1192,
            'East Sussex',
            'E2',
            'GBR-ENG-ESS',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1193,
            'Enfield',
            'E3',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1194,
            'Essex',
            'E4',
            'GBR-ENG-ESX',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1195,
            'Gateshead',
            'E5',
            'GBR-ENG-TAW',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1196,
            'Gloucestershire',
            'E6',
            'GBR-ENG-GLC',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1197,
            'Greenwich',
            'E7',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1198,
            'Hackney',
            'E8',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1199,
            'Halton',
            'E9',
            'GBR-ENG-CHS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1200,
            'Hammersmith and Fulham',
            'F1',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1201,
            'Hampshire',
            'F2',
            'GBR-ENG-HMP',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1202,
            'Haringey',
            'F3',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1203,
            'Harrow',
            'F4',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1204,
            'Hartlepool',
            'F5',
            'GBR-ENG-DRH',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1205,
            'Havering',
            'F6',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1206,
            'Hertfordshire',
            'F7',
            'GBR-ENG-HRT',
            4153,
            { region: 4345, aliases: 'Herefordshire', notes: 'County' },
        ],
        [
            1207,
            'Hertford',
            'F8',
            'GBR-ENG-HRT',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1208,
            'Hillingdon',
            'F9',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1209,
            'Hounslow',
            'G1',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1210,
            'Isle of Wight',
            'G2',
            'GBR-ENG-IOW',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1211,
            'Islington',
            'G3',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1212,
            'Kensington and Chelsea',
            'G4',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1213,
            'Kent',
            'G5',
            'GBR-ENG-KEN',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1214,
            'Kingston upon Hull, City of',
            'G6',
            'GBR-ENG-HMB',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1215,
            'Kingston upon Thames',
            'G7',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1216,
            'Kirklees',
            'G8',
            'GBR-ENG-WYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1217,
            'Knowsley',
            'G9',
            'GBR-ENG-MRS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1218,
            'Lambeth',
            'H1',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1219,
            'Lancashire',
            'H2',
            'GBR-ENG-LNC',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1220,
            'Leeds',
            'H3',
            'GBR-ENG-WYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1221,
            'Leicester',
            'H4',
            'GBR-ENG-LCS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1222,
            'Leicestershire',
            'H5',
            'GBR-ENG-LCS',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1223,
            'Lewisham',
            'H6',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1224,
            'Lincolnshire',
            'H7',
            'GBR-ENG-LNL',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1225,
            'Liverpool',
            'H8',
            'GBR-ENG-MRS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1226,
            'London, City of',
            'H9',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1227,
            'Luton',
            'I1',
            'GBR-ENG-BDF',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [1228, 'Manchester', '04', 'JAM-MNC', 4188],
        [
            1229,
            'Medway',
            'I3',
            'GBR-ENG-KEN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1230,
            'Merton',
            'I4',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1231,
            'Middlesbrough',
            'I5',
            'GBR-ENG-NYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1232,
            'Milton Keynes',
            'I6',
            'GBR-ENG-BCK',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1233,
            'Newcastle upon Tyne',
            'I7',
            'GBR-ENG-TAW',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1234,
            'Newham',
            'I8',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1235,
            'Norfolk',
            'I9',
            'GBR-ENG-NRF',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1236,
            'Northamptonshire',
            'J1',
            'GBR-ENG-NRT',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1237,
            'North East Lincolnshire',
            'J2',
            'GBR-ENG-HMB',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1238,
            'North Lincolnshire',
            'J3',
            'GBR-ENG-HMB',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1239,
            'North Somerset',
            'J4',
            'GBR-ENG-SMR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1240,
            'North Tyneside',
            'J5',
            'GBR-ENG-TAW',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1241,
            'Northumberland',
            'J6',
            'GBR-ENG-NRH',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1242,
            'North Yorkshire',
            'J7',
            'GBR-ENG-NYR',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1243,
            'Nottingham',
            'J8',
            'GBR-ENG-NTT',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1244,
            'Nottinghamshire',
            'J9',
            'GBR-ENG-NTT',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1245,
            'Oldham',
            'K1',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1246,
            'Oxfordshire',
            'K2',
            'GBR-ENG-OXF',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1247,
            'Peterborough',
            'K3',
            'GBR-ENG-CMB',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1248,
            'Plymouth',
            'K4',
            'GBR-ENG-DEV',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1249,
            'Poole',
            'K5',
            'GBR-ENG-DRS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1250,
            'Portsmouth',
            'K6',
            'GBR-ENG-HMP',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1251,
            'Reading',
            'K7',
            'GBR-ENG-BRK',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1252,
            'Redbridge',
            'K8',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1253,
            'Redcar and Cleveland',
            'K9',
            'GBR-ENG-NYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1254,
            'Richmond upon Thames',
            'L1',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1255,
            'Rochdale',
            'L2',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1256,
            'Rotherham',
            'L3',
            'GBR-ENG-SYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1257,
            'Rutland',
            'L4',
            'GBR-ENG-LCS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1258,
            'Salford',
            'L5',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1259,
            'Shropshire',
            'L6',
            'GBR-ENG-SAL',
            4153,
            { region: 4345, aliases: 'Salop', notes: 'County' },
        ],
        [
            1260,
            'Sandwell',
            'L7',
            'GBR-ENG-WMD',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1261,
            'Sefton',
            'L8',
            'GBR-ENG-MRS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1262,
            'Sheffield',
            'L9',
            'GBR-ENG-SYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1263,
            'Slough',
            'M1',
            'GBR-ENG-BRK',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1264,
            'Solihull',
            'M2',
            'GBR-ENG-WMD',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1265,
            'Somerset',
            'M3',
            'GBR-ENG-SMR',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1266,
            'Southend-on-Sea',
            'M5',
            'GBR-ENG-ESX',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1267,
            'South Gloucestershire',
            'M6',
            'GBR-ENG-GLC',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1268,
            'South Tyneside',
            'M7',
            'GBR-ENG-TAW',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1269,
            'Southwark',
            'M8',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1270,
            'Staffordshire',
            'M9',
            'GBR-ENG-STF',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1271,
            'St. Helens',
            'N1',
            'GBR-ENG-MRS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1272,
            'Stockport',
            'N2',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1273,
            'Stockton-on-Tees',
            'N3',
            'GBR-ENG-DRH',
            4153,
            {
                region: 4345,
                notes: 'Unitary District, Parts of this are also in North Yorkshire, the part south of the Tees.',
            },
        ],
        [
            1274,
            'Stoke-on-Trent',
            'N4',
            'GBR-ENG-STF',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1275,
            'Suffolk',
            'N5',
            'GBR-ENG-SFF',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1276,
            'Sunderland',
            'N6',
            'GBR-ENG-TAW',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1277,
            'Surrey',
            'N7',
            'GBR-ENG-SRR',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1278,
            'Sutton',
            'N8',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1279,
            'Swindon',
            'N9',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, aliases: 'Swinton', notes: 'Unitary District' },
        ],
        [
            1280,
            'Tameside',
            'O1',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1281,
            'Telford and Wrekin',
            'O2',
            'GBR-ENG-SAL',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1282,
            'Thurrock',
            'O3',
            'GBR-ENG-ESX',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1283,
            'Torbay',
            'O4',
            'GBR-ENG-DEV',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1284,
            'Tower Hamlets',
            'O5',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1285,
            'Trafford',
            'O6',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1286,
            'Wakefield',
            'O7',
            'GBR-ENG-WYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1287,
            'Walsall',
            'O8',
            'GBR-ENG-WMD',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1288,
            'Waltham Forest',
            'O9',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1289,
            'Wandsworth',
            'P1',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1290,
            'Warrington',
            'P2',
            'GBR-ENG-CHS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1291,
            'Warwickshire',
            'P3',
            'GBR-ENG-WRW',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1292,
            'West Berkshire',
            'P4',
            'GBR-ENG-BRK',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1293,
            'Westminster',
            'P5',
            'GBR-ENG-GLN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1294,
            'West Sussex',
            'P6',
            'GBR-ENG-WSS',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1295,
            'Wigan',
            'P7',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1296,
            'Wiltshire',
            'P8',
            'GBR-ENG-WLT',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1297,
            'Windsor and Maidenhead',
            'P9',
            'GBR-ENG-BRK',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1298,
            'Wirral',
            'Q1',
            'GBR-ENG-MRS',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1299,
            'Wokingham',
            'Q2',
            'GBR-ENG-BRK',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1300,
            'Wolverhampton',
            'Q3',
            'GBR-ENG-WMD',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [
            1301,
            'Worcestershire',
            'Q4',
            'GBR-ENG-HAW',
            4153,
            { region: 4345, notes: 'County' },
        ],
        [
            1302,
            'York',
            'Q5',
            'GBR-ENG-NYR',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [1303, 'Antrim', 'Q6', 'GBR-NIR-ANT', 4153, { region: 4346 }],
        [1304, 'Ards', 'Q7', 'GBR-NIR-ARD', 4153, { region: 4346 }],
        [1305, 'Armagh', 'Q8', 'GBR-NIR-ARM', 4153, { region: 4346 }],
        [1306, 'Ballymena', 'Q9', 'GBR-NIR-BLL', 4153, { region: 4346 }],
        [1307, 'Ballymoney', 'R1', 'GBR-NIR-BLY', 4153, { region: 4346 }],
        [1308, 'Banbridge', 'R2', 'GBR-NIR-BNB', 4153, { region: 4346 }],
        [1309, 'Belfast', 'R3', 'GBR-NIR-BLF', 4153, { region: 4346 }],
        [1310, 'Carrickfergus', 'R4', 'GBR-NIR-CRR', 4153, { region: 4346 }],
        [1311, 'Castlereagh', 'R5', 'GBR-NIR-CST', 4153, { region: 4346 }],
        [1312, 'Coleraine', 'R6', 'GBR-NIR-CLR', 4153, { region: 4346 }],
        [1313, 'Cookstown', 'R7', 'GBR-NIR-CKS', 4153, { region: 4346 }],
        [1314, 'Craigavon', 'R8', 'GBR-NIR-CRG', 4153, { region: 4346 }],
        [1315, 'Down', 'R9', 'GBR-NIR-DOW', 4153, { region: 4346 }],
        [1316, 'Dungannon', 'S1', 'GBR-NIR-DNG', 4153, { region: 4346 }],
        [1317, 'Fermanagh', 'S2', 'GBR-NIR-FRM', 4153, { region: 4346 }],
        [1318, 'Larne', 'S3', 'GBR-NIR-LAR', 4153, { region: 4346 }],
        [1319, 'Limavady', 'S4', 'GBR-NIR-LMV', 4153, { region: 4346 }],
        [1320, 'Lisburn', 'S5', 'GBR-NIR-LSB', 4153, { region: 4346 }],
        [
            1321,
            'Derry',
            'S6',
            'GBR-NIR-LND',
            4153,
            { region: 4346, aliases: 'Londonderry' },
        ],
        [1322, 'Magherafelt', 'S7', 'GBR-NIR-MGH', 4153, { region: 4346 }],
        [1323, 'Moyle', 'S8', 'GBR-NIR-MOY', 4153, { region: 4346 }],
        [1324, 'Newry and Mourne', 'S9', 'GBR-NIR-NAM', 4153, { region: 4346 }],
        [1325, 'Newtownabbey', 'T1', 'GBR-NIR-NWT', 4153, { region: 4346 }],
        [1326, 'North Down', 'T2', 'GBR-NIR-NDO', 4153, { region: 4346 }],
        [1327, 'Omagh', 'T3', 'GBR-NIR-OMG', 4153, { region: 4346 }],
        [1328, 'Strabane', 'T4', 'GBR-NIR-STR', 4153, { region: 4346 }],
        [1329, 'Aberdeen City', 'T5', 'GBR-SCT-GRM', 4153, { region: 4347 }],
        [1330, 'Aberdeenshire', 'T6', 'GBR-SCT-GRM', 4153, { region: 4347 }],
        [1331, 'Angus', 'T7', 'GBR-SCT-TYS', 4153, { region: 4347 }],
        [1332, 'Argyll and Bute', 'T8', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [
            1333,
            'Scottish Borders, The',
            'T9',
            'GBR-SCT-BRD',
            4153,
            { region: 4347 },
        ],
        [1334, 'Clackmannanshire', 'U1', 'GBR-SCT-CNT', 4153, { region: 4347 }],
        [
            1335,
            'Dumfries and Galloway',
            'U2',
            'GBR-SCT-DAG',
            4153,
            { region: 4347 },
        ],
        [1336, 'Dundee City', 'U3', 'GBR-SCT-TYS', 4153, { region: 4347 }],
        [1337, 'East Ayrshire', 'U4', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1338, 'East Dunbartonshire', 'U5', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1339, 'East Lothian', 'U6', 'GBR-SCT-LTH', 4153, { region: 4347 }],
        [1340, 'East Renfrewshire', 'U7', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1341, 'Edinburgh, City of', 'U8', 'GBR-SCT-LTH', 4153, { region: 4347 }],
        [1342, 'Falkirk', 'U9', 'GBR-SCT-CNT', 4153, { region: 4347 }],
        [
            1343,
            'Fife',
            'V1',
            'GBR-SCT-FYF',
            4153,
            { region: 4347, aliases: 'Fyfe' },
        ],
        [1344, 'Glasgow City', 'V2', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1345, 'Highland', 'V3', 'GBR-SCT-HGH', 4153, { region: 4347 }],
        [1346, 'Inverclyde', 'V4', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1347, 'Midlothian', 'V5', 'GBR-SCT-LTH', 4153, { region: 4347 }],
        [1348, 'Moray', 'V6', 'GBR-SCT-GRM', 4153, { region: 4347 }],
        [1349, 'North Ayrshire', 'V7', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1350, 'North Lanarkshire', 'V8', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1351, 'Orkney', 'V9', 'GBR-SCT-ORK', 4153, { region: 4347 }],
        [1352, 'Perth and Kinross', 'W1', 'GBR-SCT-TYS', 4153, { region: 4347 }],
        [1353, 'Renfrewshire', 'W2', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1354, 'Shetland Islands', 'W3', 'GBR-SCT-SHT', 4153, { region: 4347 }],
        [1355, 'South Ayrshire', 'W4', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1356, 'South Lanarkshire', 'W5', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1357, 'Stirling', 'W6', 'GBR-SCT-CNT', 4153, { region: 4347 }],
        [1358, 'West Dunbartonshire', 'W7', 'GBR-SCT-STT', 4153, { region: 4347 }],
        [1359, 'Eilean Siar', 'W8', 'GBR-SCT-WIS', 4153, { region: 4347 }],
        [1360, 'West Lothian', 'W9', 'GBR-SCT-LTH', 4153, { region: 4347 }],
        [1361, 'Isle of Anglesey', 'X1', 'GBR-WLS-GWY', 4153, { region: 4348 }],
        [1362, 'Blaenau Gwent', 'X2', 'GBR-WLS-GWN', 4153, { region: 4348 }],
        [1363, 'Bridgend', 'X3', 'GBR-WLS-MGL', 4153, { region: 4348 }],
        [1364, 'Caerphilly', 'X4', 'GBR-WLS-GWN', 4153, { region: 4348 }],
        [1365, 'Cardiff', 'X5', 'GBR-WLS-SGL', 4153, { region: 4348 }],
        [1366, 'Ceredigion', 'X6', 'GBR-WLS-DYF', 4153, { region: 4348 }],
        [1367, 'Carmarthenshire', 'X7', 'GBR-WLS-DYF', 4153, { region: 4348 }],
        [1368, 'Conwy', 'X8', 'GBR-WLS-GWY', 4153, { region: 4348 }],
        [1369, 'Denbighshire', 'X9', 'GBR-WLS-CLW', 4153, { region: 4348 }],
        [1370, 'Flintshire', 'Y1', 'GBR-WLS-CLW', 4153, { region: 4348 }],
        [1371, 'Gwynedd', 'Y2', 'GBR-WLS-GWY', 4153, { region: 4348 }],
        [1372, 'Merthyr Tydfil', 'Y3', 'GBR-WLS-MGL', 4153, { region: 4348 }],
        [1373, 'Monmouthshire', 'Y4', 'GBR-WLS-GWN', 4153, { region: 4348 }],
        [1374, 'Neath Port Talbot', 'Y5', 'GBR-WLS-WGL', 4153, { region: 4348 }],
        [1375, 'Newport', 'Y6', 'GBR-WLS-GWN', 4153, { region: 4348 }],
        [1376, 'Pembrokeshire', 'Y7', 'GBR-WLS-DYF', 4153, { region: 4348 }],
        [1377, 'Powys', 'Y8', 'GBR-WLS-PWY', 4153, { region: 4348 }],
        [1378, 'Rhondda Cynon Taff', 'Y9', 'GBR-WLS-MGL', 4153, { region: 4348 }],
        [1379, 'Swansea', 'Z1', 'GBR-WLS-WGL', 4153, { region: 4348 }],
        [1380, 'Torfaen', 'Z2', 'GBR-WLS-GWN', 4153, { region: 4348 }],
        [
            1381,
            'Vale of Glamorgan, The',
            'Z3',
            'GBR-WLS-SGL',
            4153,
            { region: 4348 },
        ],
        [1382, 'Wrexham', 'Z4', 'GBR-WLS-CLW', 4153, { region: 4348 }],
        [1383, 'Abashis Raioni', '01', , 4155],
        [1384, 'Abkhazia', '02', , 4155],
        [1385, 'Adigenis Raioni', '03', , 4155],
        [1386, 'Ajaria', '04', , 4155],
        [1387, 'Akhalgoris Raioni', '05', , 4155],
        [1388, "Akhalk'alak'is Raioni", '06', , 4155],
        [1389, "Akhalts'ikhis Raioni", '07', , 4155],
        [1390, 'Akhmetis Raioni', '08', , 4155],
        [1391, 'Ambrolauris Raioni', '09', , 4155],
        [1392, 'Aspindzis Raioni', '10', , 4155],
        [1393, "Baghdat'is Raioni", '11', , 4155],
        [1394, 'Bolnisis Raioni', '12', , 4155],
        [1395, 'Borjomis Raioni', '13', , 4155],
        [1396, "Chiat'ura", '14', , 4155],
        [1397, "Ch'khorotsqus Raioni", '15', , 4155],
        [1398, "Ch'okhatauris Raioni", '16', , 4155],
        [1399, "Dedop'listsqaros Raioni", '17', , 4155],
        [1400, 'Dmanisis Raioni', '18', , 4155],
        [1401, "Dushet'is Raioni", '19', , 4155],
        [1402, 'Gardabanis Raioni', '20', , 4155],
        [1403, 'Gori', '21', , 4155],
        [1404, 'Goris Raioni', '22', , 4155],
        [1405, 'Gurjaanis Raioni', '23', , 4155],
        [1406, 'Javis Raioni', '24', , 4155],
        [1407, "K'arelis Raioni", '25', , 4155],
        [1408, 'Kaspis Raioni', '26', , 4155],
        [1409, 'Kharagaulis Raioni', '27', , 4155],
        [1410, 'Khashuris Raioni', '28', , 4155],
        [1411, 'Khobis Raioni', '29', , 4155],
        [1412, 'Khonis Raioni', '30', , 4155],
        [1413, "K'ut'aisi", '31', , 4155],
        [1414, 'Lagodekhis Raioni', '32', , 4155],
        [1415, "Lanch'khut'is Raioni", '33', , 4155],
        [1416, 'Lentekhis Raioni', '34', , 4155],
        [1417, 'Marneulis Raioni', '35', , 4155],
        [1418, 'Martvilis Raioni', '36', , 4155],
        [1419, 'Mestiis Raioni', '37', , 4155],
        [1420, "Mts'khet'is Raioni", '38', , 4155],
        [1421, 'Ninotsmindis Raioni', '39', , 4155],
        [1422, 'Onis Raioni', '40', , 4155],
        [1423, "Ozurget'is Raioni", '41', , 4155],
        [1424, "P'ot'i", '42', , 4155],
        [1425, 'Qazbegis Raioni', '43', , 4155],
        [1426, 'Qvarlis Raioni', '44', , 4155],
        [1427, "Rust'avi", '45', , 4155],
        [1428, "Sach'kheris Raioni", '46', , 4155],
        [1429, 'Sagarejos Raioni', '47', , 4155],
        [1430, 'Samtrediis Raioni', '48', , 4155],
        [1431, 'Senakis Raioni', '49', , 4155],
        [1432, 'Sighnaghis Raioni', '50', , 4155],
        [1433, "T'bilisi", '51', , 4155],
        [1434, "T'elavis Raioni", '52', , 4155],
        [1435, "T'erjolis Raioni", '53', , 4155],
        [1436, "T'et'ritsqaros Raioni", '54', , 4155],
        [1437, "T'ianet'is Raioni", '55', , 4155],
        [1438, 'Tqibuli', '56', , 4155],
        [1439, "Ts'ageris Raioni", '57', , 4155],
        [1440, 'Tsalenjikhis Raioni', '58', , 4155],
        [1441, 'Tsalkis Raioni', '59', , 4155],
        [1442, 'Tsqaltubo', '60', , 4155],
        [1443, 'Vanis Raioni', '61', , 4155],
        [1444, "Zestap'onis Raioni", '62', , 4155],
        [1445, 'Zugdidi', '63', , 4155],
        [1446, 'Zugdidis Raioni', '64', , 4155],
        [1447, 'Greater Accra', '01', 'GHA-GAC', 4158],
        [1448, 'Ashanti', '02', 'GHA-ASH', 4158],
        [1449, 'Brong-Ahafo', '03', 'GHA-BAH', 4158],
        [1450, 'Volta', '08', 'GHA-VOL', 4158],
        [1451, 'Upper East', '10', 'GHA-UPE', 4158],
        [1452, 'Upper West', '11', 'GHA-UPW', 4158],
        [1453, 'Nordgronland', '01', 'GRL-NRD', 4160],
        [1454, 'Ostgronland', '02', 'GRL-OST', 4160],
        [1455, 'Vestgronland', '03', 'GRL-VST', 4160],
        [1456, 'Banjul', '01', 'GMB-BNJ', 4161],
        [1457, 'Lower River', '02', 'GMB-LRI', 4161],
        [1458, 'Central River', '03', 'GMB-MIS', 4161],
        [1459, 'Upper River', '04', 'GMB-URI', 4161],
        [1460, 'North Bank', '07', 'GMB-NBA', 4161],
        [1461, 'Beyla', '01', 'GIN-BEY', 4162],
        [1462, 'Boffa', '02', 'GIN-BOF', 4162],
        [1463, 'Boke', '03', 'GIN-BOK', 4162],
        [1464, 'Conakry', '04', 'GIN-CNK', 4162],
        [1465, 'Dabola', '05', 'GIN-DAB', 4162],
        [1466, 'Dalaba', '06', 'GIN-DAL', 4162],
        [1467, 'Dinguiraye', '07', 'GIN-DNG', 4162],
        [1468, 'Faranah', '09', 'GIN-FRN', 4162],
        [1469, 'Forecariah', '10', 'GIN-FRC', 4162],
        [1470, 'Fria', '11', 'GIN-FRI', 4162],
        [1471, 'Gaoual', '12', 'GIN-GAO', 4162],
        [1472, 'Gueckedou', '13', 'GIN-GCK', 4162],
        [1473, 'Kerouane', '15', 'GIN-KER', 4162],
        [1474, 'Kindia', '16', 'GIN-KIN', 4162],
        [1475, 'Kissidougou', '17', 'GIN-KSS', 4162],
        [1476, 'Koundara', '18', 'GIN-KND', 4162],
        [1477, 'Kouroussa', '19', 'GIN-KRS', 4162],
        [1478, 'Macenta', '21', 'GIN-MCN', 4162],
        [1479, 'Mali', '22', 'GIN-MLI', 4162],
        [1480, 'Mamou', '23', 'GIN-MAM', 4162],
        [1481, 'Pita', '25', 'GIN-PIT', 4162],
        [1482, 'Telimele', '27', 'GIN-TLM', 4162],
        [1483, 'Tougue', '28', 'GIN-TOU', 4162],
        [1484, 'Yomou', '29', 'GIN-YOM', 4162],
        [1485, 'Coyah', '30', 'GIN-COY', 4162],
        [1486, 'Dubreka', '31', 'GIN-COY', 4162],
        [1487, 'Kankan', '32', 'GIN-KNK', 4162],
        [1488, 'Koubia', '33', 'GIN-KOB', 4162],
        [1489, 'Labe', '34', 'GIN-LAB', 4162],
        [1490, 'Lelouma', '35', 'GIN-LLM', 4162],
        [1491, 'Lola', '36', 'GIN-LOL', 4162],
        [1492, 'Mandiana', '37', 'GIN-MND', 4162],
        [1493, 'Nzerekore', '38', 'GIN-NZR', 4162],
        [1494, 'Siguiri', '39', 'GIN-SIG', 4162],
        [1495, 'Annobon', '03', 'GNQ-LTR', 4164],
        [1496, 'Bioko Norte', '04', 'GNQ-BKN', 4164],
        [1497, 'Bioko Sur', '05', 'GNQ-BKS', 4164],
        [1498, 'Centro Sur', '06', 'GNQ-CNS', 4164],
        [1499, 'Kie-Ntem', '07', 'GNQ-KNT', 4164],
        [1500, 'Litoral', '08', 'GNQ-LTR', 4164],
        [1501, 'Wele-Nzas', '09', 'GNQ-WLN', 4164],
        [1502, 'Evros', '01', 'GRC-THR', 4165, { region: 4397 }],
        [1503, 'Rodhopi', '02', 'GRC-THR', 4165, { region: 4397 }],
        [1504, 'Xanthi', '03', 'GRC-THR', 4165, { region: 4397 }],
        [1505, 'Drama', '04', 'GRC-MCD', 4165, { region: 4395 }],
        [1506, 'Serrai', '05', 'GRC-MCD', 4165, { region: 4395 }],
        [1507, 'Kilkis', '06', 'GRC-MCD', 4165, { region: 4395 }],
        [1508, 'Pella', '07', 'GRC-MCD', 4165, { region: 4395 }],
        [1509, 'Florina', '08', 'GRC-MCD', 4165, { region: 4395 }],
        [1510, 'Kastoria', '09', 'GRC-MCD', 4165, { region: 4395 }],
        [1511, 'Grevena', '10', 'GRC-MCD', 4165, { region: 4395 }],
        [1512, 'Kozani', '11', 'GRC-MCD', 4165, { region: 4395 }],
        [1513, 'Imathia', '12', 'GRC-MCD', 4165, { region: 4395 }],
        [1514, 'Thessaloniki', '13', 'GRC-MCD', 4165, { region: 4395 }],
        [1515, 'Kavala', '14', 'GRC-MCD', 4165, { region: 4395 }],
        [1516, 'Khalkidhiki', '15', 'GRC-MCD', 4165, { region: 4395 }],
        [1517, 'Pieria', '16', 'GRC-MCD', 4165, { region: 4395 }],
        [1518, 'Ioannina', '17', 'GRC-IPR', 4165, { region: 4394 }],
        [1519, 'Thesprotia', '18', 'GRC-IPR', 4165, { region: 4394 }],
        [1520, 'Preveza', '19', 'GRC-IPR', 4165, { region: 4394 }],
        [1521, 'Larisa', '21', 'GRC-THS', 4165, { region: 4398 }],
        [1522, 'Trikala', '22', 'GRC-THS', 4165, { region: 4398 }],
        [1523, 'Kardhitsa', '23', 'GRC-THS', 4165, { region: 4398 }],
        [1524, 'Magnisia', '24', 'GRC-THS', 4165, { region: 4398 }],
        [
            1525,
            'Kerkira',
            '25',
            'GRC-IIS',
            4165,
            { region: 4393, aliases: 'Corfu' },
        ],
        [1526, 'Levkas', '26', 'GRC-IIS', 4165, { region: 4393 }],
        [
            1527,
            'Kefallinia',
            '27',
            'GRC-IIS',
            4165,
            { region: 4393, aliases: 'Kefalonia' },
        ],
        [1528, 'Zakinthos', '28', 'GRC-IIS', 4165, { region: 4393 }],
        [1529, 'Fthiotis', '29', 'GRC-CGE', 4165, { region: 4391 }],
        [1530, 'Evritania', '30', 'GRC-CGE', 4165, { region: 4391 }],
        [1531, 'Aitolia kai Akarnania', '31', 'GRC-CGE', 4165, { region: 4391 }],
        [1532, 'Fokis', '32', 'GRC-CGE', 4165, { region: 4391 }],
        [1533, 'Voiotia', '33', 'GRC-CGE', 4165, { region: 4391 }],
        [1534, 'Evvoia', '34', 'GRC-CGE', 4165, { region: 4391 }],
        [1535, 'Attiki', '35', 'GRC-CGE', 4165, { region: 4391 }],
        [1536, 'Argolis', '36', 'GRC-PLP', 4165, { region: 4396 }],
        [1537, 'Korinthia', '37', 'GRC-PLP', 4165, { region: 4396 }],
        [1538, 'Akhaia', '38', 'GRC-PLP', 4165, { region: 4396 }],
        [1539, 'Ilia', '39', 'GRC-PLP', 4165, { region: 4396 }],
        [1540, 'Messinia', '40', 'GRC-PLP', 4165, { region: 4396 }],
        [1541, 'Arkadhia', '41', 'GRC-PLP', 4165, { region: 4396 }],
        [1542, 'Lakonia', '42', 'GRC-PLP', 4165, { region: 4396 }],
        [1543, 'Khania', '43', 'GRC-CRT', 4165, { region: 4392 }],
        [1544, 'Rethimni', '44', 'GRC-CRT', 4165, { region: 4392 }],
        [1545, 'Iraklion', '45', 'GRC-CRT', 4165, { region: 4392 }],
        [1546, 'Lasithi', '46', 'GRC-CRT', 4165, { region: 4392 }],
        [1547, 'Dhodhekanisos', '47', 'GRC-AIS', 4165, { region: 4390 }],
        [1548, 'Samos', '48', 'GRC-AIS', 4165, { region: 4390 }],
        [1549, 'Kikladhes', '49', 'GRC-AIS', 4165, { region: 4390 }],
        [1550, 'Khios', '50', 'GRC-AIS', 4165, { region: 4390 }],
        [1551, 'Lesvos', '51', 'GRC-AIS', 4165, { region: 4390 }],
        [1552, 'Alta Verapaz', '01', 'GTM-AVR', 4167],
        [1553, 'Baja Verapaz', '02', 'GTM-BVR', 4167],
        [1554, 'Chimaltenango', '03', 'GTM-CHM', 4167],
        [1555, 'Chiquimula', '04', 'GTM-CHQ', 4167],
        [1556, 'El Progreso', '05', 'GTM-EPR', 4167],
        [1557, 'Escuintla', '06', 'GTM-ESC', 4167],
        [
            1558,
            'Guatemala Department',
            '07',
            'GTM-GTM',
            4167,
            { aliases: 'Guatemala' },
        ],
        [1559, 'Huehuetenango', '08', 'GTM-HHT', 4167],
        [1560, 'Izabal', '09', 'GTM-IZB', 4167],
        [1561, 'Jalapa', '10', 'GTM-JAL', 4167],
        [1562, 'Jutiapa', '11', 'GTM-JUT', 4167],
        [1563, 'Peten', '12', 'GTM-PET', 4167],
        [1564, 'Quetzaltenango', '13', 'GTM-QZL', 4167],
        [1565, 'Quiche', '14', 'GTM-QUI', 4167],
        [1566, 'Retalhuleu', '15', 'GTM-RTL', 4167],
        [1567, 'Sacatepequez', '16', 'GTM-SCT', 4167],
        [1568, 'San Marcos', '17', 'GTM-SMR', 4167],
        [1569, 'Santa Rosa', '18', 'GTM-SRO', 4167],
        [1570, 'Solola', '19', 'GTM-SOL', 4167],
        [1571, 'Suchitepequez', '20', 'GTM-SCH', 4167],
        [1572, 'Totonicapan', '21', 'GTM-TTN', 4167],
        [1573, 'Zacapa', '22', 'GTM-ZAC', 4167],
        [1574, 'Bafata', '01', 'GNB-BAF', 4169],
        [1575, 'Quinara', '02', 'GNB-QNR', 4169],
        [1576, 'Oio', '04', 'GNB-OIO', 4169],
        [1577, 'Bolama', '05', 'GNB-BOL', 4169],
        [1578, 'Cacheu', '06', 'GNB-CAC', 4169],
        [1579, 'Tombali', '07', 'GNB-TMB', 4169],
        [1580, 'Gabu', '10', 'GNB-GAB', 4169],
        [1581, 'Bissau', '11', 'GNB-BIS', 4169],
        [1582, 'Biombo', '12', 'GNB-BMB', 4169],
        [1583, 'Barima-Waini', '10', , 4170],
        [1584, 'Cuyuni-Mazaruni', '11', , 4170],
        [1585, 'Demerara-Mahaica', '12', , 4170],
        [1586, 'East Berbice-Corentyne', '13', , 4170],
        [1587, 'Essequibo Islands-West Demerara', '14', , 4170],
        [1588, 'Mahaica-Berbice', '15', , 4170],
        [1589, 'Pomeroon-Supenaam', '16', , 4170],
        [1590, 'Potaro-Siparuni', '17', , 4170],
        [1591, 'Upper Demerara-Berbice', '18', , 4170],
        [1592, 'Upper Takutu-Upper Essequibo', '19', , 4170],
        [1593, 'Atlantida', '01', , 4173],
        [1594, 'Choluteca', '02', , 4173],
        [1595, 'Colon', '04', 'PAN-COL', 4248],
        [1596, 'Comayagua', '04', , 4173],
        [1597, 'Copan', '05', , 4173],
        [1598, 'Cortes', '06', , 4173],
        [1599, 'El Paraiso', '07', , 4173],
        [1600, 'Francisco Morazan', '08', , 4173],
        [1601, 'Gracias a Dios', '09', , 4173],
        [1602, 'Intibuca', '10', , 4173],
        [1603, 'Islas de la Bahia', '11', , 4173],
        [1604, 'Lempira', '13', , 4173],
        [1605, 'Ocotepeque', '14', , 4173],
        [1606, 'Olancho', '15', , 4173],
        [1607, 'Santa Barbara', '16', , 4173],
        [1608, 'Valle', '17', , 4173],
        [1609, 'Yoro', '18', , 4173],
        [1610, 'Bjelovarsko-Bilogorska', '01', , 4174],
        [1611, 'Brodsko-Posavska', '02', , 4174],
        [1612, 'Dubrovacko-Neretvanska', '03', , 4174],
        [1613, 'Istarska', '04', , 4174],
        [1614, 'Karlovacka', '05', , 4174],
        [1615, 'Koprivnicko-Krizevacka', '06', , 4174],
        [1616, 'Krapinsko-Zagorska', '07', , 4174],
        [1617, 'Licko-Senjska', '08', , 4174],
        [1618, 'Medimurska', '09', , 4174],
        [1619, 'Osjecko-Baranjska', '10', , 4174],
        [1620, 'Pozesko-Slavonska', '11', , 4174],
        [1621, 'Primorsko-Goranska', '12', , 4174],
        [1622, 'Sibensko-Kninska', '13', , 4174],
        [1623, 'Sisacko-Moslavacka', '14', , 4174],
        [1624, 'Splitsko-Dalmatinska', '15', , 4174],
        [1625, 'Varazdinska', '16', , 4174],
        [1626, 'Viroviticko-Podravska', '17', , 4174],
        [1627, 'Vukovarsko-Srijemska', '18', , 4174],
        [1628, 'Zadarska', '19', , 4174],
        [1629, 'Zagrebacka', '20', , 4174],
        [1630, 'Grad Zagreb', '21', , 4174],
        [1631, 'Artibonite', '06', 'HTI-ART', 4175],
        [1632, 'Nord-Est', '10', 'HTI-NES', 4175],
        [1633, 'Sud-Est', '13', 'HTI-SES', 4175],
        [1634, "Grand' Anse", '13', , 4270],
        [1635, 'Nippes', '15', 'HTI-GAN', 4175],
        [1636, 'Bacs-Kiskun', '01', 'HUN-BKS', 4176],
        [1637, 'Baranya', '02', 'HUN-BRN', 4176],
        [1638, 'Bekes', '03', 'HUN-BEK', 4176],
        [1639, 'Borsod-Abauj-Zemplen', '04', 'HUN-BAZ', 4176],
        [1640, 'Budapest', '05', 'HUN-BDP', 4176],
        [1641, 'Csongrad', '06', 'HUN-CSN', 4176],
        [1642, 'Debrecen', '07', 'HUN-HBI', 4176, { notes: 'City 1645' }],
        [1643, 'Fejer', '08', 'HUN-FEJ', 4176],
        [1644, 'Gyor-Moson-Sopron', '09', 'HUN-GSP', 4176],
        [1645, 'Hajdu-Bihar', '10', 'HUN-HBI', 4176],
        [1646, 'Heves', '11', 'HUN-HEV', 4176],
        [1647, 'Komarom-Esztergom', '12', 'HUN-KMR', 4176],
        [1648, 'Miskolc', '13', 'HUN-BAZ', 4176, { notes: 'City 1639' }],
        [1649, 'Nograd', '14', 'HUN-NGR', 4176],
        [1650, 'Pecs', '15', 'HUN-BRN', 4176, { notes: 'City 1637' }],
        [1651, 'Pest', '16', 'HUN-PES', 4176],
        [1652, 'Somogy', '17', 'HUN-SMG', 4176],
        [1653, 'Szabolcs-Szatmar-Bereg', '18', 'HUN-SSZ', 4176],
        [1654, 'Szeged', '19', 'HUN-CSN', 4176, { notes: 'City 1641' }],
        [1655, 'Jasz-Nagykun-Szolnok', '20', 'HUN-SZL', 4176],
        [1656, 'Tolna', '21', 'HUN-TOL', 4176],
        [1657, 'Vas', '22', 'HUN-VAS', 4176],
        [1658, 'Veszprem', '23', 'HUN-VSZ', 4176],
        [1659, 'Zala', '24', 'HUN-ZAL', 4176],
        [1660, 'Gyor', '25', 'HUN-GSP', 4176, { notes: 'City 1644' }],
        [1661, 'Bekescsaba', '26', 'HUN-BEK', 4176, { notes: 'City 1638' }],
        [1662, 'Dunaujvaros', '27', 'HUN-FEJ', 4176, { notes: 'City 1643' }],
        [1663, 'Eger', '28', 'HUN-HEV', 4176, { notes: 'City 1646' }],
        [1664, 'Hodmezovasarhely', '29', 'HUN-CSN', 4176, { notes: 'City 1641' }],
        [1665, 'Kaposvar', '30', 'HUN-SMG', 4176, { notes: 'City 1652' }],
        [1666, 'Kecskemet', '31', 'HUN-BKS', 4176, { notes: 'City 1636' }],
        [1667, 'Nagykanizsa', '32', 'HUN-ZAL', 4176, { notes: 'City 1659' }],
        [1668, 'Nyiregyhaza', '33', 'HUN-SSZ', 4176, { notes: 'City 1653' }],
        [1669, 'Sopron', '34', 'HUN-GSP', 4176, { notes: 'City 1644' }],
        [1670, 'Szekesfehervar', '35', 'HUN-FEJ', 4176, { notes: 'City 1643' }],
        [1671, 'Szolnok', '36', 'HUN-SZL', 4176, { notes: 'City 1655' }],
        [1672, 'Szombathely', '37', 'HUN-VAS', 4176, { notes: 'City 1657' }],
        [1673, 'Tatabanya', '38', 'HUN-KMR', 4176, { notes: 'City 1647' }],
        [1674, 'Veszprem', '39', 'HUN-VSZ', 4176, { notes: 'City 1658' }],
        [1675, 'Zalaegerszeg', '40', 'HUN-ZAL', 4176, { notes: 'City 1659' }],
        [1676, 'Salgotarjan', '41', 'HUN-NGR', 4176, { notes: 'City 1649' }],
        [1677, 'Szekszard', '42', 'HUN-TOL', 4176, { notes: 'City 1656' }],
        [1678, 'Aceh', '01', 'IDN-ACE', 4177],
        [1679, 'Bali', '02', 'IDN-BAL', 4177],
        [1680, 'Bengkulu', '03', 'IDN-BNG', 4177],
        [1681, 'Jakarta Raya', '04', 'IDN-JRA', 4177],
        [1682, 'Jambi', '05', 'IDN-JAM', 4177],
        [1683, 'Jawa Barat', '06', 'IDN-JBA', 4177],
        [1684, 'Jawa Tengah', '07', 'IDN-JTN', 4177],
        [1685, 'Jawa Timur', '08', 'IDN-JTI', 4177],
        [1686, 'Papua', '09', 'IDN-IJA', 4177],
        [1687, 'Yogyakarta', '10', 'IDN-YGY', 4177],
        [1688, 'Kalimantan Barat', '11', 'IDN-KBA', 4177],
        [1689, 'Kalimantan Selatan', '12', 'IDN-KSL', 4177],
        [1690, 'Kalimantan Tengah', '13', 'IDN-KTN', 4177],
        [1691, 'Kalimantan Timur', '14', 'IDN-KTI', 4177],
        [1692, 'Lampung', '15', 'IDN-LMP', 4177],
        [1693, 'Maluku', '16', 'IDN-MAL', 4177],
        [1694, 'Nusa Tenggara Barat', '17', 'IDN-NTB', 4177],
        [1695, 'Nusa Tenggara Timur', '18', 'IDN-NTT', 4177],
        [1696, 'Riau', '19', 'IDN-RIA', 4177],
        [1697, 'Sulawesi Selatan', '20', 'IDN-SSL', 4177],
        [1698, 'Sulawesi Tengah', '21', 'IDN-STN', 4177],
        [1699, 'Sulawesi Tenggara', '22', 'IDN-STG', 4177],
        [1700, 'Sulawesi Utara', '23', 'IDN-SUT', 4177],
        [1701, 'Sumatera Barat', '24', 'IDN-SBA', 4177],
        [1702, 'Sumatera Selatan', '25', 'IDN-SST', 4177],
        [1703, 'Sumatera Utara', '26', 'IDN-SUA', 4177],
        [1704, 'Maluku', '28', 'IDN-KTI', 4177],
        [1705, 'Maluku Utara', '29', 'IDN-MAL', 4177],
        [1706, 'Jawa Barat', '30', 'IDN-JBA', 4177],
        [1707, 'Sulawesi Utara', '31', 'IDN-SUT', 4177],
        [1708, 'Sumatera Selatan', '32', 'IDN-SST', 4177],
        [1709, 'Banten', '33', 'IDN-JBA', 4177],
        [1710, 'Gorontalo', '34', 'IDN-SUT', 4177],
        [1711, 'Kepulauan Bangka Belitung', '35', 'IDN-SST', 4177],
        [1712, 'Papua', '36', 'IDN-IJA', 4177],
        [1713, 'Riau', '37', 'IDN-RIA', 4177],
        [1714, 'Sulawesi Selatan', '38', 'IDN-SSL', 4177],
        [1715, 'Irian Jaya Barat', '39', 'IDN-IJA', 4177],
        [1716, 'Kepulauan Riau', '40', 'IDN-RIA', 4177],
        [1717, 'Sulawesi Barat', '41', 'IDN-SSL', 4177],
        [
            1718,
            'Carlow',
            '01',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1719,
            'Cavan',
            '02',
            'IRE-ULT',
            4178,
            { notes: 'County, Province Ulster' },
        ],
        [
            1720,
            'Clare',
            '03',
            'IRE-MUN',
            4178,
            { notes: 'County, Province Munster' },
        ],
        [
            1721,
            'Cork',
            '04',
            'IRE-MUN',
            4178,
            { notes: 'County, Province Munster' },
        ],
        [
            1722,
            'Donegal',
            '06',
            'IRE-ULT',
            4178,
            { notes: 'County, Province Ulster' },
        ],
        [
            1723,
            'Dublin',
            '07',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1724,
            'Galway',
            '10',
            'IRE-CNN',
            4178,
            { notes: 'County, Province Connacht' },
        ],
        [
            1725,
            'Kerry',
            '11',
            'IRE-MUN',
            4178,
            { notes: 'County, Province Munster' },
        ],
        [
            1726,
            'Kildare',
            '12',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1727,
            'Kilkenny',
            '13',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1728,
            'Leitrim',
            '14',
            'IRE-CNN',
            4178,
            { notes: 'County, Province Connacht' },
        ],
        [
            1729,
            'Laois',
            '15',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1730,
            'Limerick',
            '16',
            'IRE-MUN',
            4178,
            { notes: 'County, Province Munster' },
        ],
        [
            1731,
            'Longford',
            '18',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1732,
            'Louth',
            '19',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1733,
            'Mayo',
            '20',
            'IRE-CNN',
            4178,
            { notes: 'County, Province Connacht' },
        ],
        [
            1734,
            'Meath',
            '21',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1735,
            'Monaghan',
            '22',
            'IRE-ULT',
            4178,
            { notes: 'County, Province Ulster' },
        ],
        [
            1736,
            'Offaly',
            '23',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1737,
            'Roscommon',
            '24',
            'IRE-CNN',
            4178,
            { notes: 'County, Province Connacht' },
        ],
        [
            1738,
            'Sligo',
            '25',
            'IRE-CNN',
            4178,
            { notes: 'County, Province Connacht' },
        ],
        [
            1739,
            'Tipperary',
            '26',
            'IRE-MUN',
            4178,
            { notes: 'County, Province Munster' },
        ],
        [
            1740,
            'Waterford',
            '27',
            'IRE-MUN',
            4178,
            { notes: 'County, Province Munster' },
        ],
        [
            1741,
            'Westmeath',
            '29',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1742,
            'Wexford',
            '30',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [
            1743,
            'Wicklow',
            '31',
            'IRE-LEN',
            4178,
            { notes: 'County, Province Leinster' },
        ],
        [1744, 'HaDarom', '01', 'ISR-HDR', 4179],
        [1745, 'HaMerkaz', '02', 'ISR-HMR', 4179],
        [1746, 'HaZafon', '03', 'ISR-HZF', 4179],
        [1747, 'Hefa', '04', 'ISR-HEF', 4179],
        [1748, 'Tel Aviv', '05', 'ISR-TAV', 4179],
        [1749, 'Yerushalayim', '06', 'ISR-YRS', 4179],
        [1750, 'Andaman and Nicobar Islands', '01', 'IND-AAN', 4181],
        [1751, 'Andhra Pradesh', '02', 'IND-APR', 4181],
        [1752, 'Assam', '03', 'IND-ASS', 4181],
        [1753, 'Chandigarh', '05', 'IND-CHN', 4181],
        [1754, 'Dadra and Nagar Haveli', '06', 'IND-DAN', 4181],
        [1755, 'Delhi', '07', 'IND-DEL', 4181],
        [1756, 'Gujarat', '09', 'IND-GJR', 4181],
        [1757, 'Haryana', '10', 'IND-HRY', 4181],
        [1758, 'Himachal Pradesh', '11', 'IND-HPR', 4181],
        [1759, 'Jammu and Kashmir', '12', 'IND-JAK', 4181],
        [1760, 'Kerala', '13', 'IND-KER', 4181],
        [1761, 'Lakshadweep', '14', 'IND-LKS', 4181],
        [1762, 'Maharashtra', '16', 'IND-MHR', 4181],
        [1763, 'Manipur', '17', 'IND-MNP', 4181],
        [1764, 'Meghalaya', '18', 'IND-MGH', 4181],
        [1765, 'Karnataka', '19', 'IND-KRN', 4181],
        [1766, 'Nagaland', '20', 'IND-NGL', 4181],
        [1767, 'Orissa', '21', 'IND-ORS', 4181],
        [1768, 'Puducherry', '22', 'IND-PND', 4181],
        [1769, 'Punjab', '04', 'PAK-PNJ', 4253],
        [1770, 'Rajasthan', '24', 'IND-RJS', 4181],
        [1771, 'Tamil Nadu', '25', 'IND-TNA', 4181],
        [1772, 'Tripura', '26', 'IND-TRP', 4181],
        [1773, 'West Bengal', '28', 'IND-WBN', 4181],
        [1774, 'Sikkim', '29', 'IND-SKK', 4181],
        [1775, 'Arunachal Pradesh', '30', 'IND-APD', 4181],
        [1776, 'Mizoram', '31', 'IND-MZR', 4181],
        [1777, 'Daman and Diu', '32', 'IND-DAD', 4181],
        [1778, 'Goa', '33', 'IND-GOA', 4181],
        [1779, 'Bihar', '34', 'IND-BIH', 4181],
        [1780, 'Madhya Pradesh', '35', 'IND-MPR', 4181],
        [1781, 'Uttar Pradesh', '36', 'IND-UPR', 4181],
        [1782, 'Chhattisgarh', '37', 'IND-MPR', 4181],
        [1783, 'Jharkhand', '38', 'IND-BIH', 4181],
        [1784, 'Uttarakhand', '39', 'IND-UPR', 4181],
        [1785, 'Al Anbar', '01', 'IRQ-AAN', 4183],
        [1786, 'Al Basrah', '02', 'IRQ-ABS', 4183],
        [1787, 'Al Muthanna', '03', 'IRQ-AMT', 4183],
        [1788, 'Al Qadisiyah', '04', 'IRQ-AQD', 4183],
        [1789, 'As Sulaymaniyah', '05', 'IRQ-ASL', 4183],
        [1790, 'Babil', '06', 'IRQ-BAB', 4183],
        [1791, 'Baghdad', '07', 'IRQ-BGH', 4183],
        [1792, 'Dahuk', '08', 'IRQ-DAH', 4183],
        [1793, 'Dhi Qar', '09', 'IRQ-DQA', 4183],
        [1794, 'Diyala', '10', 'IRQ-DIY', 4183],
        [1795, 'Arbil', '11', 'IRQ-IRB', 4183],
        [1796, "Karbala'", '12', 'IRQ-KRB', 4183],
        [1797, "At Ta'mim", '13', 'IRQ-ATA', 4183],
        [1798, 'Maysan', '14', 'IRQ-MYS', 4183],
        [1799, 'Ninawa', '15', 'IRQ-NIN', 4183],
        [1800, 'Wasit', '16', 'IRQ-WAS', 4183],
        [1801, 'An Najaf', '17', 'IRQ-ANA', 4183],
        [1802, 'Salah ad Din', '18', 'IRQ-SAD', 4183],
        [1803, 'Azarbayjan-e Bakhtari', '01', 'IRN-AEB', 4184],
        [1804, 'Azarbayjan-e Khavari', '02', 'IRN-AEK', 4184],
        [1805, 'Chahar Mahall va Bakhtiari', '03', 'IRN-CMV', 4184],
        [1806, 'Sistan va Baluchestan', '04', 'IRN-BVS', 4184],
        [1807, 'Kohkiluyeh va Buyer Ahmadi', '05', 'IRN-BAV', 4184],
        [1808, 'Fars', '07', 'IRN-FAR', 4184],
        [1809, 'Gilan', '08', 'IRN-GIL', 4184],
        [1810, 'Hamadan', '09', 'IRN-HMD', 4184],
        [1811, 'Ilam', '10', 'IRN-ILA', 4184],
        [1812, 'Hormozgan', '11', 'IRN-HRM', 4184],
        [1813, 'Kerman', '12', 'IRN-KRM', 4184],
        [1814, 'Bakhtaran', '13', 'IRN-BKH', 4184],
        [1815, 'Khuzestan', '15', 'IRN-KHZ', 4184],
        [1816, 'Kordestan', '16', 'IRN-KRD', 4184],
        [1817, 'Mazandaran', '17', 'IRN-MZN', 4184],
        [
            1818,
            'Semnan Province',
            '18',
            'IRN-SMN',
            4184,
            { aliases: 'Semn?n Province' },
        ],
        [1819, 'Markazi', '19', 'IRN-MRK', 4184],
        [1820, 'Zanjan', '21', 'IRN-ZNJ', 4184],
        [1821, 'Bushehr', '22', 'IRN-BUS', 4184],
        [1822, 'Lorestan', '23', 'IRN-LRS', 4184],
        [1823, 'Markazi', '24', 'IRN-MRK', 4184],
        [1824, 'Semnan', '25', 'IRN-SMN', 4184],
        [1825, 'Tehran', '26', 'IRN-TER', 4184],
        [1826, 'Zanjan', '27', 'IRN-ZNJ', 4184],
        [1827, 'Esfahan', '28', 'IRN-ESF', 4184],
        [1828, 'Kerman', '29', 'IRN-KRM', 4184],
        [1829, 'Khorasan', '30', 'IRN-KHR', 4184],
        [1830, 'Yazd', '31', 'IRN-YAZ', 4184],
        [1831, 'Ardabil', '32', 'IRN-AEK', 4184],
        [1832, 'East Azarbaijan', '33', 'IRN-AEK', 4184],
        [1833, 'Markazi', '34', 'IRN-MRK', 4184],
        [1834, 'Mazandaran', '35', 'IRN-MZN', 4184],
        [1835, 'Zanjan', '36', 'IRN-ZNJ', 4184],
        [1836, 'Golestan', '37', 'IRN-MZN', 4184],
        [1837, 'Qazvin', '38', 'IRN-ZNJ', 4184],
        [1838, 'Qom', '39', 'IRN-MRK', 4184],
        [1839, 'Yazd', '40', 'IRN-YAZ', 4184],
        [1840, 'Khorasan-e Janubi', '41', 'IRN-KHR', 4184],
        [1841, 'Khorasan-e Razavi', '42', 'IRN-KHR', 4184],
        [1842, 'Khorasan-e Shemali', '43', 'IRN-KHR', 4184],
        [1843, 'Arnessysla', '03', 'ISL-SDH', 4185],
        [1844, 'Austur-Hunavatnssysla', '05', 'ISL-NVS', 4185],
        [1845, 'Austur-Skaftafellssysla', '06', 'ISL-AST', 4185],
        [1846, 'Borgarfjardarsysla', '07', 'ISL-VSL', 4185],
        [1847, 'Eyjafjardarsysla', '09', 'ISL-NYS', 4185],
        [1848, 'Gullbringusysla', '10', 'ISL-RYK', 4185],
        [1849, 'Kjosarsysla', '15', 'ISL-RYK', 4185],
        [1850, 'Myrasysla', '17', 'ISL-VSL', 4185],
        [1851, 'Nordur-Mulasysla', '20', 'ISL-AST', 4185],
        [1852, 'Nordur-Tingeyjarsysla', '21', 'ISL-NYS', 4185],
        [1853, 'Rangarvallasysla', '23', 'ISL-SDH', 4185],
        [1854, 'Skagafjardarsysla', '28', 'ISL-NVS', 4185],
        [1855, 'Snafellsnes- og Hnappadalssysla', '29', 'ISL-VSL', 4185],
        [1856, 'Sudur-Mulasysla', '31', 'ISL-AST', 4185],
        [1857, 'Sudur-Tingeyjarsysla', '32', 'ISL-NYS', 4185],
        [1858, 'Vestur-Bardastrandarsysla', '34', 'ISL-VST', 4185],
        [1859, 'Vestur-Hunavatnssysla', '35', 'ISL-NVS', 4185],
        [1860, 'Vestur-Isafjardarsysla', '36', 'ISL-VST', 4185],
        [1861, 'Vestur-Skaftafellssysla', '37', 'ISL-SDH', 4185],
        [1862, 'Norourland Eystra', '40', 'ISL-NYS', 4185, { notes: 'Region' }],
        [1863, 'Norourland Vestra', '41', 'ISL-NVS', 4185, { notes: 'Region' }],
        [1864, 'Suourland', '42', 'ISL-SDH', 4185, { notes: 'Region' }],
        [1865, 'Suournes', '43', 'ISL-RYK', 4185],
        [1866, 'Vestfiroir', '44', 'ISL-VST', 4185, { notes: 'Region' }],
        [1867, 'Vesturland', '45', 'ISL-VSL', 4185, { notes: 'Region' }],
        [1868, 'Abruzzi', '01', 'ITA-ABR', 4186],
        [1869, 'Basilicata', '02', 'ITA-BSL', 4186],
        [1870, 'Calabria', '03', 'ITA-CLB', 4186],
        [1871, 'Campania', '04', 'ITA-CMP', 4186],
        [1872, 'Emilia-Romagna', '05', 'ITA-ERM', 4186],
        [1873, 'Friuli-Venezia Giulia', '06', 'ITA-FVG', 4186],
        [1874, 'Lazio', '07', 'ITA-LAZ', 4186],
        [1875, 'Liguria', '08', 'ITA-LIG', 4186],
        [1876, 'Lombardia', '09', 'ITA-LMB', 4186],
        [1877, 'Marche', '10', 'ITA-MRC', 4186],
        [1878, 'Molise', '11', 'ITA-MOL', 4186],
        [1879, 'Piemonte', '12', 'ITA-PMN', 4186],
        [1880, 'Puglia', '13', 'ITA-PUG', 4186],
        [1881, 'Sardegna', '14', 'ITA-SRD', 4186],
        [1882, 'Sicilia', '15', 'ITA-SIC', 4186],
        [1883, 'Toscana', '16', 'ITA-TSC', 4186],
        [1884, 'Trentino-Alto Adige', '17', 'ITA-TAA', 4186],
        [1885, 'Umbria', '18', 'ITA-UMB', 4186],
        [1886, "Valle d'Aosta", '19', 'ITA-VDA', 4186],
        [1887, 'Veneto', '20', 'ITA-VEN', 4186],
        [1888, 'Clarendon', '01', 'JAM-CLR', 4188],
        [1889, 'Hanover', '02', 'JAM-HNV', 4188],
        [1890, 'Portland', '07', 'JAM-PRT', 4188],
        [1891, 'Saint Ann', '09', 'JAM-SAN', 4188],
        [1892, 'Saint Catherine', '10', 'JAM-SCT', 4188],
        [1893, 'Saint Elizabeth', '11', 'JAM-SEL', 4188],
        [1894, 'Trelawny', '15', 'JAM-TRL', 4188],
        [1895, 'Westmoreland', '16', 'JAM-WST', 4188],
        [1896, 'Kingston', '17', 'JAM-KNG', 4188],
        [1897, "Al Balqa'", '02', 'JOR-ABA', 4189],
        [1898, 'Ma', '07', 'JOR-MAA', 4189, { aliases: "Ma'an" }],
        [1899, 'Al Karak', '09', 'JOR-AKA', 4189],
        [1900, 'Al Mafraq', '10', 'JOR-AMF', 4189],
        [1901, 'Amman Governorate', '11', 'JOR-AMM', 4189],
        [1902, 'At Tafilah', '12', 'JOR-ATF', 4189],
        [1903, 'Az Zarqa', '13', 'JOR-AZR', 4189],
        [1904, 'Irbid', '14', 'JOR-IRB', 4189],
        [1905, 'Amman', '16', 'JOR-AMM', 4189],
        [1906, 'Aichi', '01', 'JPN-AIC', 4190, { region: 4350 }],
        [1907, 'Akita', '02', 'JPN-AKI', 4190, { region: 4357 }],
        [1908, 'Aomori', '03', 'JPN-AOM', 4190, { region: 4357 }],
        [1909, 'Chiba', '04', 'JPN-CHI', 4190, { region: 4353 }],
        [1910, 'Ehime', '05', 'JPN-EHI', 4190, { region: 4356 }],
        [1911, 'Fukui', '06', 'JPN-FUK', 4190, { region: 4350 }],
        [1912, 'Fukuoka', '07', 'JPN-FUU', 4190, { region: 4354 }],
        [1913, 'Fukushima', '08', 'JPN-FKS', 4190, { region: 4357 }],
        [1914, 'Gifu', '09', 'JPN-GIF', 4190, { region: 4350 }],
        [1915, 'Gumma', '10', 'JPN-GUM', 4190, { region: 4353 }],
        [1916, 'Hiroshima', '11', 'JPN-HRS', 4190, { region: 4349 }],
        [1917, 'Hokkaido', '12', 'JPN-HKK', 4190, { region: 4351 }],
        [1918, 'Hyogo', '13', 'JPN-HYO', 4190, { region: 4352 }],
        [1919, 'Ibaraki', '14', 'JPN-IBR', 4190, { region: 4353 }],
        [1920, 'Ishikawa', '15', 'JPN-ISH', 4190, { region: 4350 }],
        [1921, 'Iwate', '16', 'JPN-IWA', 4190, { region: 4357 }],
        [1922, 'Kagawa', '17', 'JPN-KAG', 4190, { region: 4356 }],
        [1923, 'Kagoshima', '18', 'JPN-KGS', 4190, { region: 4354 }],
        [1924, 'Kanagawa', '19', 'JPN-KNG', 4190, { region: 4353 }],
        [1925, 'Kochi', '20', 'JPN-KOC', 4190, { region: 4356 }],
        [1926, 'Kumamoto', '21', 'JPN-KMM', 4190, { region: 4354 }],
        [1927, 'Kyoto', '22', 'JPN-KYO', 4190, { region: 4352 }],
        [1928, 'Mie', '23', 'JPN-MIE', 4190, { region: 4350 }],
        [1929, 'Miyagi', '24', 'JPN-MIY', 4190, { region: 4357 }],
        [1930, 'Miyazaki', '25', 'JPN-MYZ', 4190, { region: 4354 }],
        [1931, 'Nagano', '26', 'JPN-NAG', 4190, { region: 4350 }],
        [1932, 'Nagasaki', '27', 'JPN-NGS', 4190, { region: 4354 }],
        [1933, 'Nara', '28', 'JPN-NAR', 4190, { region: 4352 }],
        [1934, 'Niigata', '29', 'JPN-NII', 4190, { region: 4350 }],
        [1935, 'Oita', '30', 'JPN-OIT', 4190, { region: 4354 }],
        [1936, 'Okayama', '31', 'JPN-OKY', 4190, { region: 4349 }],
        [1937, 'Osaka', '32', 'JPN-OSA', 4190, { region: 4352 }],
        [1938, 'Saga', '33', 'JPN-SAG', 4190, { region: 4354 }],
        [1939, 'Saitama', '34', 'JPN-SAI', 4190, { region: 4353 }],
        [1940, 'Shiga', '35', 'JPN-SHI', 4190, { region: 4352 }],
        [1941, 'Shimane', '36', 'JPN-SHM', 4190, { region: 4349 }],
        [1942, 'Shizuoka', '37', 'JPN-SHZ', 4190, { region: 4350 }],
        [1943, 'Tochigi', '38', 'JPN-TCH', 4190, { region: 4353 }],
        [1944, 'Tokushima', '39', 'JPN-TKS', 4190, { region: 4356 }],
        [1945, 'Tokyo', '40', 'JPN-TOK', 4190, { region: 4353 }],
        [1946, 'Tottori', '41', 'JPN-TTT', 4190, { region: 4349 }],
        [1947, 'Toyama', '42', 'JPN-TOY', 4190, { region: 4350 }],
        [1948, 'Wakayama', '43', 'JPN-WKY', 4190, { region: 4352 }],
        [1949, 'Yamagata', '44', 'JPN-YMG', 4190, { region: 4357 }],
        [1950, 'Yamaguchi', '45', 'JPN-YMC', 4190, { region: 4349 }],
        [1951, 'Yamanashi', '46', 'JPN-YMN', 4190, { region: 4350 }],
        [1952, 'Okinawa', '47', 'JPN-OKN', 4190, { region: 4355 }],
        [1954, 'Nairobi Area', '05', 'KEN-NAR', 4191],
        [1955, 'North-Eastern', '06', 'KEN-NES', 4191],
        [1956, 'Nyanza', '07', 'KEN-NYN', 4191],
        [1957, 'Rift Valley', '08', 'KEN-RVL', 4191],
        [1958, 'Bishkek', '01', , 4192],
        [1959, 'Chuy', '02', , 4192],
        [1960, 'Jalal-Abad', '03', , 4192],
        [1961, 'Naryn', '04', , 4192],
        [1962, 'Osh', '05', , 4192],
        [1963, 'Talas', '06', , 4192],
        [1964, 'Ysyk-Kol', '07', , 4192],
        [1965, 'Osh', '08', , 4192],
        [1966, 'Batken', '09', , 4192],
        [1967, 'Kampong Cham', '02', 'KHM-KCH', 4193],
        [1968, 'Kampong Chhnang', '03', 'KHM-KCN', 4193],
        [1969, 'Kampong Spoe', '04', 'KHM-KSP', 4193, { aliases: 'Kampong Speu' }],
        [1970, 'Kampong Thum', '05', 'KHM-KTH', 4193],
        [1971, 'Kampot', '06', 'KHM-KMP', 4193],
        [1972, 'Kandal', '07', 'KHM-KND', 4193],
        [1973, 'Kaoh Kong', '08', 'KHM-KKO', 4193, { aliases: 'Koh Kong' }],
        [1974, 'Kracheh', '09', 'KHM-KRC', 4193],
        [1975, 'Mondol Kiri', '10', 'KHM-MKI', 4193, { aliases: 'Mondulkiri' }],
        [1976, 'Phnum Penh', '11', 'KHM-PPE', 4193],
        [1977, 'Pouthisat', '12', 'KHM-PTH', 4193, { aliases: 'Pursat' }],
        [1978, 'Preah Vihear', '13', 'KHM-PVI', 4193],
        [1979, 'Prey Veng', '14', 'KHM-PVE', 4193],
        [1980, 'Rotanokiri', '15', 'KHM-RTN', 4193, { aliases: 'Ratanakiri Kiri' }],
        [
            1981,
            'Siemreab-Otdar Meanchey',
            '16',
            'KHM-SOM',
            4193,
            { aliases: 'Siem Reap' },
        ],
        [1982, 'Stoeng Treng', '17', 'KHM-STR', 4193, { aliases: 'Stung Treng' }],
        [1983, 'Svay Rieng', '18', 'KHM-SRI', 4193],
        [1984, 'Takev', '19', 'KHM-TAK', 4193, { aliases: 'Takeo' }],
        [1985, 'Batdambang', '29', 'KHM-BTD', 4193],
        [1986, 'Pailin', '30', 'KHM-BTD', 4193],
        [1987, 'Gilbert Islands', '01', , 4194],
        [1988, 'Line Islands', '02', , 4194],
        [1989, 'Phoenix Islands', '03', , 4194],
        [1990, 'Anjouan', '01', , 4195],
        [1991, 'Grande Comore', '02', , 4195],
        [1992, 'Moheli', '03', , 4195],
        [1993, 'Christ Church Nichola Town', '01', , 4196],
        [1994, 'Saint Anne Sandy Point', '02', , 4196],
        [1995, 'Saint George Basseterre', '03', , 4196],
        [1996, 'Saint George Gingerland', '04', , 4196],
        [1997, 'Saint James Windward', '05', , 4196],
        [1998, 'Saint John Capisterre', '06', , 4196],
        [1999, 'Saint John Figtree', '07', , 4196],
        [2000, 'Saint Mary Cayon', '08', , 4196],
        [2001, 'Saint Paul Capisterre', '09', , 4196],
        [2002, 'Saint Paul Charlestown', '10', , 4196],
        [2003, 'Saint Peter Basseterre', '11', , 4196],
        [2004, 'Saint Thomas Lowland', '12', , 4196],
        [2005, 'Saint Thomas Middle Island', '13', , 4196],
        [2006, 'Trinity Palmetto Point', '15', , 4196],
        [2007, 'Chagang-do', '01', 'PRK-CDO', 4197],
        [2008, 'Hamgyong-namdo', '03', 'PRK-HNA', 4197],
        [2009, 'Hwanghae-namdo', '06', 'PRK-HNM', 4197],
        [2010, 'Hwanghae-bukto', '07', 'PRK-HBK', 4197],
        [2011, 'Kaesong-si', '08', 'PRK-KSI', 4197],
        [2012, 'Kangwon-do', '09', 'PRK-KDO', 4197],
        [2013, "P'yongan-bukto", '11', 'PRK-PBU', 4197],
        [2014, "P'yongyang-si", '12', 'PRK-PSI', 4197],
        [2015, 'Yanggang-do', '13', 'PRK-YDO', 4197],
        [2016, "Namp'o-si", '14', 'PRK-PNA', 4197],
        [2017, "P'yongan-namdo", '15', 'PRK-PNA', 4197],
        [2018, 'Hamgyong-bukto', '17', 'PRK-HBU', 4197],
        [2019, 'Najin Sonbong-si', '18', 'PRK-HBU', 4197],
        [2020, 'Cheju-do', '01', 'KOR-CDO', 4198],
        [2021, 'Cholla-bukto', '03', 'KOR-CBU', 4198],
        [2022, "Ch'ungch'ong-bukto", '05', 'KOR-CBK', 4198],
        [2023, 'Kangwon-do', '06', 'KOR-KDO', 4198],
        [2024, 'Pusan-jikhalsi', '10', 'KOR-PSN', 4198],
        [2025, "Seoul-t'ukpyolsi", '11', 'KOR-SJK', 4198],
        [2026, "Inch'on-jikhalsi", '12', 'KOR-ICH', 4198],
        [2027, 'Kyonggi-do', '13', 'KOR-KGD', 4198],
        [2028, 'Kyongsang-bukto', '14', 'KOR-KBU', 4198],
        [2029, 'Taegu-jikhalsi', '15', 'KOR-TGN', 4198],
        [2030, 'Cholla-namdo', '16', 'KOR-CNA', 4198],
        [2031, "Ch'ungch'ong-namdo", '17', 'KOR-CNM', 4198],
        [2032, 'Kwangju-jikhalsi', '18', 'KOR-KWN', 4198],
        [2033, 'Taejon-jikhalsi', '19', 'KOR-CNM', 4198],
        [2034, 'Kyongsang-namdo', '20', 'KOR-KNA', 4198],
        [2035, 'Ulsan-gwangyoksi', '21', 'KOR-KNA', 4198],
        [2036, 'Al Ahmadi', '01', 'KWT-AHM', 4199],
        [2037, 'Al Kuwayt', '02', 'KWT-AKW', 4199],
        [2038, 'Al Jahra', '05', 'KWT-AJH', 4199],
        [2039, 'Al Farwaniyah', '07', 'KWT-AFR', 4199],
        [2040, 'Hawalli', '08', 'KWT-HWL', 4199],
        [2041, 'Mubarak al Kabir', '09', 'KWT-AJH', 4199],
        [2042, 'Creek', '01', , 4200],
        [2043, 'Midland', '03', , 4200],
        [2044, 'South Town', '04', , 4200],
        [2045, 'Spot Bay', '05', , 4200],
        [2046, 'Stake Bay', '06', , 4200],
        [2047, 'West End', '07', , 4200],
        [2048, 'Almaty', '01', 'KAZ-ALA', 4201, { aliases: 'Alma-Ata' }],
        [2049, 'Almaty City', '02', 'KAZ-ALA', 4201, { aliases: 'Alma-Ata' }],
        [2050, 'Aqmola', '03', 'KAZ-KKC', 4201, { aliases: 'Kokchetav' }],
        [2051, 'Aqtobe', '04', 'KAZ-AKB', 4201, { aliases: 'Aktyubinsk' }],
        [2052, 'Astana', '05', 'KAZ-TSL', 4201],
        [2053, 'Atyrau', '06', 'KAZ-GRV', 4201],
        [2054, 'West Kazakhstan', '07', 'KAZ-URL', 4201],
        [2055, 'Bayqonyr', '08', 'KAZ-KZO', 4201],
        [2056, 'Mangghystau', '09', 'KAZ-MGY', 4201],
        [2057, 'South Kazakhstan', '10', 'KAZ-CHM', 4201],
        [2058, 'Pavlodar', '11', 'KAZ-PVL', 4201],
        [2059, 'Qaraghandy', '12', 'KAZ-DZK', 4201],
        [2060, 'Qostanay', '13', 'KAZ-KST', 4201],
        [2061, 'Qyzylorda', '14', 'KAZ-KZO', 4201],
        [2062, 'East Kazakhstan', '15', 'KAZ-VKZ', 4201],
        [2063, 'North Kazakhstan', '16', 'KAZ-NKZ', 4201],
        [2064, 'Zhambyl', '17', 'KAZ-TKG', 4201],
        [2065, 'Attapu', '01', 'LAO-ATT', 4202],
        [2066, 'Champasak', '02', 'LAO-CHM', 4202],
        [2067, 'Houaphan', '03', 'LAO-HPH', 4202],
        [2068, 'Khammouan', '04', 'LAO-KHM', 4202],
        [2069, 'Louang Namtha', '05', 'LAO-LNM', 4202],
        [2070, 'Oudomxai', '07', 'LAO-ODM', 4202],
        [2071, 'Phongsali', '08', 'LAO-PHN', 4202],
        [2072, 'Saravan', '09', 'LAO-SRV', 4202],
        [2073, 'Savannakhet', '10', 'LAO-SVN', 4202],
        [2074, 'Vientiane', '11', 'LAO-VNT', 4202],
        [2075, 'Xaignabouri', '13', 'LAO-XGN', 4202],
        [2076, 'Xiangkhoang', '14', 'LAO-XNG', 4202],
        [2077, 'Louangphrabang', '17', 'LAO-LNG', 4202],
        [2078, 'Beqaa', '01', 'LBN-ABI', 4203, { aliases: "Al Biqa'" }],
        [2079, 'Liban-Nord', '03', 'LBN-ASH', 4203, { aliases: 'Ash Shamal' }],
        [2080, 'Beyrouth', '04', 'LBN-BYR', 4203, { aliases: 'Beirut' }],
        [
            2081,
            'Mont-Liban',
            '05',
            'LBN-JLB',
            4203,
            { aliases: 'Mount Lebanon;Jabal Lubnan' },
        ],
        [2082, 'Liban-Sud', '06', 'LBN-SAY', 4203, { aliases: 'Sayda' }],
        [2083, 'Nabatiye', '07', 'LBN-ANB', 4203, { aliases: 'An Nabatiyah' }],
        [2084, 'Beqaa', '08', 'LBN-ABI', 4203, { aliases: "Al Biqa'" }],
        [2085, 'Liban-Nord', '09', 'LBN-ASH', 4203, { aliases: 'Ash Shamal' }],
        [2086, 'Aakk,r', '10', 'LBN-ASH', 4203],
        [2087, 'Baalbek-Hermel', '11', 'LBN-ABI', 4203],
        [2088, 'Anse-la-Raye', '01', , 4204],
        [2089, 'Dauphin', '02', , 4204],
        [2090, 'Castries', '03', , 4204],
        [2091, 'Choiseul', '12', , 4269],
        [2092, 'Dennery', '05', , 4204],
        [2093, 'Gros-Islet', '06', , 4204],
        [2094, 'Laborie', '07', , 4204],
        [2095, 'Micoud', '08', , 4204],
        [2096, 'Soufriere', '09', , 4204],
        [2097, 'Vieux-Fort', '10', , 4204],
        [2098, 'Praslin', '05', 'SWZ-MNZ', 4286],
        [2099, 'Balzers', '01', , 4205],
        [2100, 'Eschen', '02', , 4205],
        [2101, 'Gamprin', '03', , 4205],
        [2102, 'Mauren', '04', , 4205],
        [2103, 'Planken', '05', , 4205],
        [2104, 'Ruggell', '06', , 4205],
        [2105, 'Schaan', '07', , 4205],
        [2106, 'Schellenberg', '08', , 4205],
        [2107, 'Triesen', '09', , 4205],
        [2108, 'Triesenberg', '10', , 4205],
        [2109, 'Vaduz', '11', , 4205],
        [2110, 'Gbarpolu', '21', , 4205],
        [2111, 'River Gee', '22', , 4205],
        [2112, 'Amparai', '01', , 4206],
        [2113, 'Anuradhapura', '02', , 4206],
        [2114, 'Badulla', '03', , 4206],
        [2115, 'Batticaloa', '04', , 4206],
        [2116, 'Galle', '06', , 4206],
        [2117, 'Hambantota', '07', , 4206],
        [2118, 'Kalutara', '09', , 4206],
        [2119, 'Kandy', '10', , 4206],
        [2120, 'Kegalla', '11', , 4206],
        [2121, 'Kurunegala', '12', , 4206],
        [2122, 'Matale', '14', , 4206],
        [2123, 'Matara', '15', , 4206],
        [2124, 'Moneragala', '16', , 4206],
        [2125, 'Nuwara Eliya', '17', , 4206],
        [2126, 'Polonnaruwa', '18', , 4206],
        [2127, 'Puttalam', '19', , 4206],
        [2128, 'Ratnapura', '20', , 4206],
        [2129, 'Trincomalee', '21', , 4206],
        [2130, 'Colombo', '23', , 4206],
        [2131, 'Gampaha', '24', , 4206],
        [2132, 'Jaffna', '25', , 4206],
        [2133, 'Mannar', '26', , 4206],
        [2134, 'Mullaittivu', '27', , 4206],
        [2135, 'Vavuniya', '28', , 4206],
        [2136, 'North Central', '30', , 4206],
        [2137, 'North Western', '32', , 4206],
        [2138, 'Sabaragamuwa', '33', , 4206],
        [2139, 'Uva', '35', , 4206],
        [2140, 'Bong', '01', 'LBR-BON', 4207],
        [2141, 'Grand Cape Mount', '04', 'LBR-GCM', 4207],
        [2142, 'Monrovia', '07', 'LBR-MNT', 4207],
        [2143, 'Nimba', '09', 'LBR-NIM', 4207],
        [2144, 'Sino', '10', 'LBR-SIN', 4207],
        [2145, 'Grand Bassa', '11', 'LBR-GBA', 4207],
        [2146, 'Montserrado', '14', 'LBR-MNT', 4207],
        [2147, 'Grand Gedeh', '19', 'LBR-GJI', 4207],
        [2148, 'Lofa', '20', 'LBR-LOF', 4207],
        [2149, 'Berea', '10', 'LSO-BER', 4208],
        [2150, 'Butha-Buthe', '11', 'LSO-BBU', 4208],
        [2151, 'Leribe', '12', 'LSO-LER', 4208],
        [2152, 'Mafeteng', '13', 'LSO-MFT', 4208],
        [2153, 'Maseru', '14', 'LSO-MAS', 4208],
        [2154, 'Mohales Hoek', '15', 'LSO-MHO', 4208],
        [2155, 'Mokhotlong', '16', 'LSO-MKH', 4208],
        [2156, 'Qachas Nek', '17', 'LSO-QNE', 4208],
        [2157, 'Quthing', '18', 'LSO-QTH', 4208],
        [2158, 'Thaba-Tseka', '19', 'LSO-THT', 4208],
        [2159, 'Alytaus Apskritis', '56', , 4209],
        [2160, 'Kauno Apskritis', '57', , 4209],
        [2161, 'Klaipedos Apskritis', '58', , 4209],
        [2162, 'Marijampoles Apskritis', '59', , 4209],
        [2163, 'Panevezio Apskritis', '60', , 4209],
        [2164, 'Siauliu Apskritis', '61', , 4209],
        [2165, 'Taurages Apskritis', '62', , 4209],
        [2166, 'Telsiu Apskritis', '63', , 4209],
        [2167, 'Utenos Apskritis', '64', , 4209],
        [2168, 'Vilniaus Apskritis', '65', , 4209],
        [2169, 'Diekirch', '01', , 4210],
        [2170, 'Grevenmacher', '02', , 4210],
        [2171, 'Aizkraukles', '01', , 4211],
        [2172, 'Aluksnes', '02', , 4211],
        [2173, 'Balvu', '03', , 4211],
        [2174, 'Bauskas', '04', , 4211],
        [2175, 'Cesu', '05', , 4211],
        [2176, 'Daugavpils', '06', , 4211],
        [2177, 'Daugavpils', '07', , 4211],
        [2178, 'Dobeles', '08', , 4211],
        [2179, 'Gulbenes', '09', , 4211],
        [2180, 'Jekabpils', '10', , 4211],
        [2181, 'Jelgava', '11', , 4211],
        [2182, 'Jelgavas', '12', , 4211],
        [2183, 'Jurmala', '13', , 4211],
        [2184, 'Kraslavas', '14', , 4211],
        [2185, 'Kuldigas', '15', , 4211],
        [2186, 'Liepaja', '16', , 4211],
        [2187, 'Liepajas', '17', , 4211],
        [2188, 'Limbazu', '18', , 4211],
        [2189, 'Ludzas', '19', , 4211],
        [2190, 'Madonas', '20', , 4211],
        [2191, 'Ogres', '21', , 4211],
        [2192, 'Preilu', '22', , 4211],
        [2193, 'Rezekne', '23', , 4211],
        [2194, 'Rezeknes', '24', , 4211],
        [2195, 'Riga', '25', , 4211],
        [2196, 'Rigas', '26', , 4211],
        [2197, 'Saldus', '27', , 4211],
        [2198, 'Talsu', '28', , 4211],
        [2199, 'Tukuma', '29', , 4211],
        [2200, 'Valkas', '30', , 4211],
        [2201, 'Valmieras', '31', , 4211],
        [2202, 'Ventspils', '32', , 4211],
        [2203, 'Ventspils', '33', , 4211],
        [2204, 'Al Aziziyah', '03', , 4212],
        [2205, 'Al Jufrah', '05', , 4212],
        [2206, 'Al Kufrah', '08', , 4212],
        [2207, "Ash Shati'", '13', , 4212],
        [2208, 'Murzuq', '30', , 4212],
        [2209, 'Sabha', '34', , 4212],
        [2210, 'Tarhunah', '41', , 4212],
        [2211, 'Tubruq', '42', , 4212],
        [2212, 'Zlitan', '45', , 4212],
        [2213, 'Ajdabiya', '47', , 4212],
        [2214, 'Al Fatih', '48', , 4212],
        [2215, 'Al Jabal al Akhdar', '49', , 4212],
        [2216, 'Al Khums', '50', , 4212],
        [2217, 'An Nuqat al Khams', '51', , 4212],
        [2218, 'Awbari', '52', , 4212],
        [2219, 'Az Zawiyah', '53', , 4212],
        [2220, 'Banghazi', '54', , 4212],
        [2221, 'Darnah', '55', , 4212],
        [2222, 'Ghadamis', '56', , 4212],
        [2223, 'Gharyan', '57', , 4212],
        [2224, 'Misratah', '58', , 4212],
        [2225, 'Sawfajjin', '59', , 4212],
        [2226, 'Surt', '60', , 4212],
        [2227, 'Tarabulus', '61', , 4212],
        [2228, 'Yafran', '62', , 4212],
        [2229, 'Agadir', '01', 'MAR-SOU', 4213],
        [2230, 'Al Hoceima', '02', 'MAR-CNR', 4213],
        [2231, 'Azilal', '03', 'MAR-CEN', 4213],
        [2232, 'Ben Slimane', '04', 'MAR-CEN', 4213],
        [2233, 'Beni Mellal', '05', 'MAR-CEN', 4213],
        [2234, 'Boulemane', '06', 'MAR-CNR', 4213],
        [2235, 'Casablanca', '07', 'MAR-CEN', 4213],
        [2236, 'Chaouen', '08', 'MAR-NWT', 4213],
        [2237, 'El Jadida', '09', 'MAR-CEN', 4213],
        [2238, 'El Kelaa des Srarhna', '10', 'MAR-TNF', 4213],
        [2239, 'Er Rachidia', '11', 'MAR-CSO', 4213],
        [2240, 'Essaouira', '12', 'MAR-TNF', 4213],
        [2241, 'Fes', '13', 'MAR-CNR', 4213],
        [2242, 'Figuig', '14', 'MAR-EST', 4213],
        [2243, 'Kenitra', '15', 'MAR-NWT', 4213],
        [2244, 'Khemisset', '16', 'MAR-CEN', 4213],
        [2245, 'Khenifra', '17', 'MAR-CSO', 4213],
        [2246, 'Khouribga', '18', 'MAR-CEN', 4213],
        [2247, 'Marrakech', '19', 'MAR-TNF', 4213],
        [2248, 'Meknes', '20', 'MAR-CSO', 4213],
        [2249, 'Nador', '21', 'MAR-EST', 4213],
        [2250, 'Ouarzazate', '22', 'MAR-SOU', 4213],
        [2251, 'Oujda', '23', 'MAR-EST', 4213],
        [2252, 'Rabat-Sale', '24', 'MAR-NWT', 4213],
        [2253, 'Safi', '25', 'MAR-TNF', 4213],
        [2254, 'Settat', '26', 'MAR-CEN', 4213],
        [2255, 'Tanger', '27', 'MAR-NWT', 4213],
        [2256, 'Tata', '29', 'MAR-SOU', 4213],
        [2257, 'Taza', '30', 'MAR-CNR', 4213],
        [2258, 'Tiznit', '32', 'MAR-SOU', 4213],
        [2259, 'Guelmim', '33', 'MAR-SOU', 4213],
        [2260, 'Ifrane', '34', 'MAR-CNR', 4213],
        [2261, 'Laayoune', '35', , 4141],
        [2262, 'Tan-Tan', '36', 'MAR-SOU', 4213],
        [2263, 'Taounate', '37', 'MAR-CNR', 4213],
        [2264, 'Sidi Kacem', '38', 'MAR-CSO', 4213],
        [2265, 'Taroudannt', '39', 'MAR-SOU', 4213],
        [2266, 'Tetouan', '40', 'MAR-NWT', 4213],
        [2267, 'Larache', '41', 'MAR-NWT', 4213],
        [2268, 'Grand Casablanca', '45', 'MAR-CEN', 4213],
        [2269, 'Fes-Boulemane', '46', 'MAR-CSO', 4213],
        [2270, 'Marrakech-Tensift-Al Haouz', '47', 'MAR-TNF', 4213],
        [2271, 'Meknes-Tafilalet', '48', 'MAR-CSO', 4213],
        [2272, 'Rabat-Sale-Zemmour-Zaer', '49', 'MAR-NWT', 4213],
        [2273, 'Chaouia-Ouardigha', '50', 'MAR-CEN', 4213],
        [2274, 'Doukkala-Abda', '51', 'MAR-TNF', 4213],
        [2275, 'Gharb-Chrarda-Beni Hssen', '52', 'MAR-NWT', 4213],
        [2276, 'Guelmim-Es Smara', '53', 'MAR-SOU', 4213],
        [2277, 'Oriental', '54', 'MAR-EST', 4213],
        [2278, 'Souss-Massa-Dr,a', '55', 'MAR-SOU', 4213],
        [2279, 'Tadla-Azilal', '56', 'MAR-CEN', 4213],
        [2280, 'Tanger-Tetouan', '57', 'MAR-NWT', 4213],
        [2281, 'Taza-Al Hoceima-Taounate', '58', 'MAR-NWT', 4213],
        [2282, 'La,youne-Boujdour-Sakia El Hamra', '59', 'MAR-TNF', 4213],
        [2283, 'La Condamine', '01', , 4214],
        [2284, 'Monaco', '02', , 4214],
        [2285, 'Monte-Carlo', '03', , 4214],
        [2286, 'Balti', '46', , 4215],
        [2287, 'Cahul', '47', , 4215],
        [2288, 'Chisinau', '48', , 4215],
        [2289, 'Stinga Nistrului', '49', , 4215],
        [2290, 'Edinet', '50', , 4215],
        [2291, 'Gagauzia', '51', , 4215],
        [2292, 'Lapusna', '52', , 4215],
        [2293, 'Orhei', '53', , 4215],
        [2294, 'Soroca', '54', , 4215],
        [2295, 'Tighina', '55', , 4215],
        [2296, 'Ungheni', '56', , 4215],
        [2297, 'Stinga Nistrului', '58', , 4215],
        [2298, 'Anenii Noi', '59', , 4215],
        [2299, 'Balti', '60', , 4215],
        [2300, 'Basarabeasca', '61', , 4215],
        [2301, 'Bender', '62', , 4215],
        [2302, 'Briceni', '63', , 4215],
        [2303, 'Cahul', '64', , 4215],
        [2304, 'Cantemir', '65', , 4215],
        [2305, 'Calarasi', '41', 'ROM-CLR', 4264],
        [2306, 'Causeni', '67', , 4215],
        [2307, 'Cimislia', '68', , 4215],
        [2308, 'Criuleni', '69', , 4215],
        [2309, 'Donduseni', '70', , 4215],
        [2310, 'Drochia', '71', , 4215],
        [2311, 'Dubasari', '72', , 4215],
        [2312, 'Edinet', '73', , 4215],
        [2313, 'Falesti', '74', , 4215],
        [2314, 'Floresti', '75', , 4215],
        [2315, 'Glodeni', '76', , 4215],
        [2316, 'Hincesti', '77', , 4215],
        [2317, 'Ialoveni', '78', , 4215],
        [2318, 'Leova', '79', , 4215],
        [2319, 'Nisporeni', '80', , 4215],
        [2320, 'Ocnita', '81', , 4215],
        [2321, 'Rezina', '83', , 4215],
        [2322, 'Riscani', '84', , 4215],
        [2323, 'Singerei', '85', , 4215],
        [2324, 'Soldanesti', '86', , 4215],
        [2325, 'Soroca', '87', , 4215],
        [2326, 'Stefan-Voda', '88', , 4215],
        [2327, 'Straseni', '89', , 4215],
        [2328, 'Taraclia', '90', , 4215],
        [2329, 'Telenesti', '91', , 4215],
        [2330, 'Ungheni', '92', , 4215],
        [2331, 'Antsiranana', '01', 'MDG-ANT', 4217],
        [2332, 'Fianarantsoa', '02', 'MDG-FNR', 4217],
        [2333, 'Mahajanga', '03', 'MDG-MHJ', 4217],
        [2334, 'Toamasina', '04', 'MDG-TMS', 4217],
        [2335, 'Antananarivo', '05', 'MDG-ANN', 4217],
        [2336, 'Toliara', '06', 'MDG-TOL', 4217],
        [2337, 'Aracinovo', '01', , 4219],
        [2338, 'Bac', '02', , 4219],
        [2339, 'Belcista', '03', , 4219],
        [2340, 'Berovo', '04', , 4219],
        [2341, 'Bistrica', '05', , 4219],
        [2342, 'Bitola', '06', , 4219],
        [2343, 'Blatec', '07', , 4219],
        [2344, 'Bogdanci', '08', , 4219],
        [2345, 'Bogomila', '09', , 4219],
        [2346, 'Bogovinje', '10', , 4219],
        [2347, 'Bosilovo', '11', , 4219],
        [2348, 'Brvenica', '12', , 4219],
        [2349, 'Cair', '13', , 4219],
        [2350, 'Capari', '14', , 4219],
        [2351, 'Caska', '15', , 4219],
        [2352, 'Cegrane', '16', , 4219],
        [2353, 'Centar', '17', , 4219],
        [2354, 'Centar Zupa', '18', , 4219],
        [2355, 'Cesinovo', '19', , 4219],
        [2356, 'Cucer-Sandevo', '20', , 4219],
        [2357, 'Debar', '21', , 4219],
        [2358, 'Delcevo', '22', , 4219],
        [2359, 'Delogozdi', '23', , 4219],
        [2360, 'Demir Hisar', '24', , 4219],
        [2361, 'Demir Kapija', '25', , 4219],
        [2362, 'Dobrusevo', '26', , 4219],
        [2363, 'Dolna Banjica', '27', , 4219],
        [2364, 'Dolneni', '28', , 4219],
        [2365, 'Dorce Petrov', '29', , 4219],
        [2366, 'Drugovo', '30', , 4219],
        [2367, 'Dzepciste', '31', , 4219],
        [2368, 'Gazi Baba', '32', , 4219],
        [2369, 'Gevgelija', '33', , 4219],
        [2370, 'Gostivar', '34', , 4219],
        [2371, 'Gradsko', '35', , 4219],
        [2372, 'Ilinden', '36', , 4219],
        [2373, 'Izvor', '37', , 4219],
        [2374, 'Jegunovce', '38', , 4219],
        [2375, 'Kamenjane', '39', , 4219],
        [2376, 'Karbinci', '40', , 4219],
        [2377, 'Karpos', '41', , 4219],
        [2378, 'Kavadarci', '42', , 4219],
        [2379, 'Kicevo', '43', , 4219],
        [2380, 'Kisela Voda', '44', , 4219],
        [2381, 'Klecevce', '45', , 4219],
        [2382, 'Kocani', '46', , 4219],
        [2383, 'Konce', '47', , 4219],
        [2384, 'Kondovo', '48', , 4219],
        [2385, 'Konopiste', '49', , 4219],
        [2386, 'Kosel', '50', , 4219],
        [2387, 'Kratovo', '51', , 4219],
        [2388, 'Kriva Palanka', '52', , 4219],
        [2389, 'Krivogastani', '53', , 4219],
        [2390, 'Krusevo', '54', , 4219],
        [2391, 'Kuklis', '55', , 4219],
        [2392, 'Kukurecani', '56', , 4219],
        [2393, 'Kumanovo', '57', , 4219],
        [2394, 'Labunista', '58', , 4219],
        [2395, 'Lipkovo', '59', , 4219],
        [2396, 'Lozovo', '60', , 4219],
        [2397, 'Lukovo', '61', , 4219],
        [2398, 'Makedonska Kamenica', '62', , 4219],
        [2399, 'Makedonski Brod', '63', , 4219],
        [2400, 'Mavrovi Anovi', '64', , 4219],
        [2401, 'Meseista', '65', , 4219],
        [2402, 'Miravci', '66', , 4219],
        [2403, 'Mogila', '67', , 4219],
        [2404, 'Murtino', '68', , 4219],
        [2405, 'Negotino', '69', , 4219],
        [2406, 'Negotino-Polosko', '70', , 4219],
        [2407, 'Novaci', '71', , 4219],
        [2408, 'Novo Selo', '72', , 4219],
        [2409, 'Oblesevo', '73', , 4219],
        [2410, 'Ohrid', '74', , 4219],
        [2411, 'Orasac', '75', , 4219],
        [2412, 'Orizari', '76', , 4219],
        [2413, 'Oslomej', '77', , 4219],
        [2414, 'Pehcevo', '78', , 4219],
        [2415, 'Petrovec', '79', , 4219],
        [2416, 'Plasnica', '80', , 4219],
        [2417, 'Podares', '81', , 4219],
        [2418, 'Prilep', '82', , 4219],
        [2419, 'Probistip', '83', , 4219],
        [2420, 'Radovis', '84', , 4219],
        [2421, 'Rankovce', '85', , 4219],
        [2422, 'Resen', '86', , 4219],
        [2423, 'Rosoman', '87', , 4219],
        [2424, 'Rostusa', '88', , 4219],
        [2425, 'Samokov', '89', , 4219],
        [2426, 'Saraj', '90', , 4219],
        [2427, 'Sipkovica', '91', , 4219],
        [2428, 'Sopiste', '92', , 4219],
        [2429, 'Sopotnica', '93', , 4219],
        [2430, 'Srbinovo', '94', , 4219],
        [2431, 'Staravina', '95', , 4219],
        [2432, 'Star Dojran', '96', , 4219],
        [2433, 'Staro Nagoricane', '97', , 4219],
        [2434, 'Stip', '98', , 4219],
        [2435, 'Struga', '99', , 4219],
        [2436, 'Strumica', 'A1', , 4219],
        [2437, 'Studenicani', 'A2', , 4219],
        [2438, 'Suto Orizari', 'A3', , 4219],
        [2439, 'Sveti Nikole', 'A4', , 4219],
        [2440, 'Tearce', 'A5', , 4219],
        [2441, 'Tetovo', 'A6', , 4219],
        [2442, 'Topolcani', 'A7', , 4219],
        [2443, 'Valandovo', 'A8', , 4219],
        [2444, 'Vasilevo', 'A9', , 4219],
        [2445, 'Veles', 'B1', , 4219],
        [2446, 'Velesta', 'B2', , 4219],
        [2447, 'Vevcani', 'B3', , 4219],
        [2448, 'Vinica', 'B4', , 4219],
        [2449, 'Vitoliste', 'B5', , 4219],
        [2450, 'Vranestica', 'B6', , 4219],
        [2451, 'Vrapciste', 'B7', , 4219],
        [2452, 'Vratnica', 'B8', , 4219],
        [2453, 'Vrutok', 'B9', , 4219],
        [2454, 'Zajas', 'C1', , 4219],
        [2455, 'Zelenikovo', 'C2', , 4219],
        [2456, 'Zelino', 'C3', , 4219],
        [2457, 'Zitose', 'C4', , 4219],
        [2458, 'Zletovo', 'C5', , 4219],
        [2459, 'Zrnovci', 'C6', , 4219],
        [2460, 'Bamako', '01', 'MLI-BAM', 4220],
        [2461, 'Kayes', '03', 'MLI-KAY', 4220],
        [2462, 'Mopti', '04', 'MLI-MOP', 4220],
        [2463, 'Segou', '05', 'MLI-SEG', 4220],
        [2464, 'Sikasso', '06', 'MLI-SKS', 4220],
        [2465, 'Koulikoro', '07', 'MLI-BAM', 4220],
        [2466, 'Tombouctou', '08', 'MLI-GAO', 4220],
        [2467, 'Gao', '09', 'MLI-GAO', 4220],
        [2468, 'Kidal', '10', 'MLI-GAO', 4220],
        [2469, 'Rakhine State', '01', 'MMR-AST', 4221, { aliases: 'Arakan State' }],
        [2470, 'Chin State', '02', 'MMR-CST', 4221],
        [2471, 'Irrawaddy', '03', 'MMR-IRR', 4221],
        [2472, 'Kachin State', '04', 'MMR-KST', 4221],
        [2473, 'Karan State', '05', 'MMR-KSA', 4221],
        [2474, 'Kayah State', '06', 'MMR-KSE', 4221],
        [2475, 'Magwe', '07', 'MMR-MAG', 4221],
        [2476, 'Mandalay', '08', 'MMR-MND', 4221],
        [2477, 'Pegu', '09', 'MMR-PEG', 4221],
        [2478, 'Sagaing', '10', 'MMR-SGN', 4221],
        [2479, 'Shan State', '11', 'MMR-SST', 4221],
        [2480, 'Tenasserim', '12', 'MMR-TNS', 4221],
        [2481, 'Mon State', '13', 'MMR-MST', 4221],
        [2482, 'Rangoon', '14', 'MMR-RNG', 4221],
        [2483, 'Yangon', '17', 'MMR-RNG', 4221],
        [2484, 'Arhangay', '01', 'MNG-ARH', 4222],
        [2485, 'Bayanhongor', '02', 'MNG-BYN', 4222],
        [2486, 'Bayan-Olgiy', '03', 'MNG-BOL', 4222],
        [2487, 'Darhan', '05', 'MNG-DRH', 4222],
        [2488, 'Dornod', '06', 'MNG-DRN', 4222],
        [2489, 'Dornogovi', '07', 'MNG-DRG', 4222],
        [2490, 'Dundgovi', '08', 'MNG-DND', 4222],
        [2491, 'Dzavhan', '09', 'MNG-DZV', 4222],
        [2492, 'Govi-Altay', '10', 'MNG-GAL', 4222],
        [2493, 'Hentiy', '11', 'MNG-HNT', 4222],
        [2494, 'Hovd', '12', 'MNG-HOV', 4222],
        [2495, 'Hovsgol', '13', 'MNG-HVS', 4222],
        [2496, 'Omnogovi', '14', 'MNG-OMN', 4222],
        [2497, 'Ovorhangay', '15', 'MNG-OVR', 4222],
        [2498, 'Selenge', '16', 'MNG-SLN', 4222],
        [2499, 'Suhbaatar', '17', 'MNG-SHB', 4222],
        [2500, 'Tov', '18', 'MNG-TOV', 4222],
        [2501, 'Uvs', '19', 'MNG-UVS', 4222],
        [2502, 'Ulaanbaatar', '20', 'MNG-ULN', 4222],
        [2503, 'Bulgan', '21', 'MNG-BLG', 4222],
        [2504, 'Erdenet', '22', 'MNG-BLG', 4222],
        [2505, 'Darhan-Uul', '23', 'MNG-DRH', 4222],
        [2506, 'Govisumber', '24', 'MNG-DRG', 4222],
        [2507, 'Orhon', '25', 'MNG-BLG', 4222],
        [2508, 'Ilhas', '01', , 4223],
        [2509, 'Macau', '02', , 4223],
        [2510, 'Hodh Ech Chargui', '01', 'MRT-HEC', 4226],
        [2511, 'Hodh El Gharbi', '02', 'MRT-HEG', 4226],
        [2512, 'Assaba', '03', 'MRT-ASB', 4226],
        [2513, 'Gorgol', '04', 'MRT-GRG', 4226],
        [2514, 'Brakna', '05', 'MRT-BRK', 4226],
        [2515, 'Trarza', '06', 'MRT-TRR', 4226],
        [2516, 'Dakhlet Nouadhibou', '08', 'MRT-DND', 4226],
        [2517, 'Tagant', '09', 'MRT-TGN', 4226],
        [2518, 'Guidimaka', '10', 'MRT-GDM', 4226],
        [2519, 'Tiris Zemmour', '11', 'MRT-TZM', 4226],
        [2520, 'Inchiri', '12', 'MRT-INC', 4226],
        [2521, 'Saint Anthony', '01', , 4227],
        [2522, 'Saint Georges', '02', , 4227],
        [2523, 'Black River', '12', , 4229],
        [2524, 'Flacq', '13', , 4229],
        [2525, 'Grand Port', '14', , 4229],
        [2526, 'Moka', '15', , 4229],
        [2527, 'Pamplemousses', '16', , 4229],
        [2528, 'Plaines Wilhems', '17', , 4229],
        [2529, 'Port Louis', '18', , 4229],
        [2530, 'Riviere du Rempart', '19', , 4229],
        [2531, 'Savanne', '20', , 4229],
        [2532, 'Agalega Islands', '21', , 4229],
        [2533, 'Cargados Carajos', '22', , 4229],
        [2534, 'Rodrigues', '23', , 4229],
        [2535, 'Seenu', '01', 'MDV-SEE', 4230],
        [2536, 'Aliff', '02', 'MDV-ALF', 4230],
        [2537, 'Laviyani', '03', 'MDV-LVY', 4230],
        [2538, 'Waavu', '04', 'MDV-WAA', 4230],
        [2539, 'Laamu', '05', 'MDV-LAA', 4230],
        [2540, 'Haa Aliff', '07', 'MDV-HAL', 4230],
        [2541, 'Thaa', '08', 'MDV-THA', 4230],
        [2542, 'Meemu', '12', 'MDV-MEE', 4230],
        [2543, 'Raa', '13', 'MDV-RAA', 4230],
        [2544, 'Faafu', '14', 'MDV-FAA', 4230],
        [2545, 'Daalu', '17', 'MDV-DAA', 4230],
        [2546, 'Baa', '20', 'MDV-BAA', 4230],
        [2547, 'Haa Daalu', '23', 'MDV-HDA', 4230],
        [2548, 'Shaviyani', '24', 'MDV-SHV', 4230],
        [2549, 'Noonu', '25', 'MDV-NOO', 4230],
        [2550, 'Kaafu', '26', 'MDV-KAA', 4230],
        [2551, 'Gaafu Aliff', '27', 'MDV-GAL', 4230],
        [2552, 'Gaafu Daalu', '28', 'MDV-GDA', 4230],
        [2553, 'Naviyani', '29', 'MDV-NVY', 4230],
        [2554, 'Male', '40', 'MDV-KAA', 4230],
        [2555, 'Chikwawa', '02', 'MWI-STH', 4231],
        [2556, 'Chiradzulu', '03', 'MWI-STH', 4231],
        [2557, 'Chitipa', '04', 'MWI-NRT', 4231],
        [2558, 'Thyolo', '05', 'MWI-STH', 4231],
        [2559, 'Dedza', '06', 'MWI-CNT', 4231],
        [2560, 'Dowa', '07', 'MWI-CNT', 4231],
        [2561, 'Karonga', '08', 'MWI-NRT', 4231],
        [2562, 'Kasungu', '09', 'MWI-CNT', 4231],
        [2563, 'Lilongwe', '11', 'MWI-CNT', 4231],
        [2564, 'Mangochi', '12', 'MWI-STH', 4231],
        [2565, 'Mchinji', '13', 'MWI-CNT', 4231],
        [2566, 'Mzimba', '15', 'MWI-NRT', 4231],
        [2567, 'Ntcheu', '16', 'MWI-CNT', 4231],
        [2568, 'Nkhata Bay', '17', 'MWI-STH', 4231],
        [2569, 'Nkhotakota', '18', 'MWI-CNT', 4231],
        [2570, 'Nsanje', '19', 'MWI-STH', 4231],
        [2571, 'Ntchisi', '20', 'MWI-CNT', 4231],
        [2572, 'Rumphi', '21', 'MWI-NRT', 4231],
        [2573, 'Salima', '22', 'MWI-CNT', 4231],
        [2574, 'Zomba', '23', 'MWI-STH', 4231],
        [2575, 'Blantyre', '24', 'MWI-STH', 4231],
        [2576, 'Mwanza', '12', 'TZA-MWN', 4302],
        [2577, 'Balaka', '26', 'MWI-STH', 4231],
        [2578, 'Likoma', '27', 'MWI-NRT', 4231],
        [2579, 'Machinga', '28', 'MWI-STH', 4231],
        [2580, 'Mulanje', '29', 'MWI-STH', 4231],
        [2581, 'Phalombe', '30', 'MWI-STH', 4231],
        [2582, 'Aguascalientes', '01', 'MEX-AGS', 4232],
        [2583, 'Baja California', '02', 'MEX-BCN', 4232],
        [2584, 'Baja California Sur', '03', 'MEX-BCS', 4232],
        [2585, 'Campeche', '04', 'MEX-CMP', 4232],
        [2586, 'Chiapas', '05', 'MEX-CHP', 4232],
        [2587, 'Chihuahua', '06', 'MEX-CHH', 4232],
        [2588, 'Coahuila de Zaragoza', '07', 'MEX-CDZ', 4232],
        [2589, 'Colima', '08', 'MEX-COL', 4232],
        [2590, 'Durango', '10', 'MEX-DRN', 4232],
        [2591, 'Guanajuato', '11', 'MEX-GNJ', 4232],
        [2592, 'Guerrero', '12', 'MEX-GRR', 4232],
        [2593, 'Hidalgo', '13', 'MEX-HDL', 4232],
        [2594, 'Jalisco', '14', 'MEX-JLS', 4232],
        [2595, 'Mexico State', '15', 'MEX-MEX', 4232, { aliases: 'Mexico' }],
        [2596, 'Michoacan de Ocampo', '16', 'MEX-MDO', 4232],
        [2597, 'Morelos', '17', 'MEX-MRL', 4232],
        [2598, 'Nayarit', '18', 'MEX-NYR', 4232],
        [2599, 'Nuevo Leon', '19', 'MEX-NLE', 4232],
        [2600, 'Oaxaca', '20', 'MEX-OAX', 4232],
        [2601, 'Puebla', '21', 'MEX-PUE', 4232],
        [2602, 'Queretaro de Arteaga', '22', 'MEX-QDA', 4232],
        [2603, 'Quintana Roo', '23', 'MEX-QRO', 4232],
        [2604, 'San Luis Potosi', '24', 'MEX-SLP', 4232],
        [2605, 'Sinaloa', '25', 'MEX-SIN', 4232],
        [2606, 'Sonora', '26', 'MEX-SON', 4232],
        [2607, 'Tabasco', '27', 'MEX-TBS', 4232],
        [2608, 'Tamaulipas', '28', 'MEX-TML', 4232],
        [2609, 'Tlaxcala', '29', 'MEX-TLX', 4232],
        [2610, 'Veracruz-Llave', '30', 'MEX-VLL', 4232],
        [2611, 'Yucatan', '31', 'MEX-YCT', 4232],
        [2612, 'Zacatecas', '32', 'MEX-ZCT', 4232],
        [2613, 'Johor', '01', 'MYS-JOH', 4233],
        [2614, 'Kedah', '02', 'MYS-KED', 4233],
        [2615, 'Kelantan', '03', 'MYS-KLN', 4233],
        [2616, 'Melaka', '04', 'MYS-MEL', 4233],
        [2617, 'Negeri Sembilan', '05', 'MYS-NSM', 4233],
        [2618, 'Pahang', '06', 'MYS-PHN', 4233],
        [2619, 'Perak', '07', 'MYS-PER', 4233],
        [2620, 'Perlis', '08', 'MYS-PRL', 4233],
        [2621, 'Pulau Pinang', '09', 'MYS-PPN', 4233],
        [2622, 'Sarawak', '11', 'MYS-SRW', 4233],
        [2623, 'Selangor', '12', 'MYS-SLN', 4233],
        [2624, 'Terengganu', '13', 'MYS-TRN', 4233],
        [2625, 'Kuala Lumpur', '14', 'MYS-KLL', 4233],
        [2626, 'Labuan', '15', 'MYS-SAB', 4233],
        [2627, 'Sabah', '16', 'MYS-SAB', 4233],
        [2628, 'Putrajaya', '17', 'MYS-SLN', 4233],
        [2629, 'Cabo Delgado', '01', 'MOZ-CDL', 4234],
        [2630, 'Gaza', 'GZ', , 4258],
        [2631, 'Inhambane', '03', 'MOZ-INH', 4234],
        [2632, 'Maputo', '04', 'MOZ-MAP', 4234],
        [2633, 'Sofala', '05', 'MOZ-SOF', 4234],
        [2634, 'Nampula', '06', 'MOZ-NMP', 4234],
        [2635, 'Niassa', '07', 'MOZ-NIA', 4234],
        [2636, 'Tete', '08', 'MOZ-TET', 4234],
        [2637, 'Zambezia', '09', 'MOZ-ZMB', 4234],
        [2638, 'Manica', '10', 'MOZ-MAN', 4234],
        [2639, 'Maputo', '11', 'MOZ-MAP', 4234],
        [2640, 'Bethanien', '01', 'NAM-BTH', 4235],
        [2641, 'Caprivi Oos', '02', 'NAM-COO', 4235],
        [2642, 'Boesmanland', '03', 'NAM-BSM', 4235],
        [2643, 'Gobabis', '04', 'NAM-GBB', 4235],
        [2644, 'Grootfontein', '05', 'NAM-GRT', 4235],
        [2645, 'Kaokoland', '06', 'NAM-KKL', 4235],
        [2646, 'Karibib', '07', 'NAM-KRB', 4235],
        [2647, 'Keetmanshoop', '08', 'NAM-KTM', 4235],
        [2648, 'Luderitz', '09', 'NAM-LDR', 4235],
        [2649, 'Maltahohe', '10', 'NAM-MLT', 4235],
        [2650, 'Okahandja', '11', 'NAM-OKH', 4235],
        [2651, 'Omaruru', '12', 'NAM-OMR', 4235],
        [2652, 'Otjiwarongo', '13', 'NAM-OTJ', 4235],
        [2653, 'Outjo', '14', 'NAM-OUT', 4235],
        [2654, 'Owambo', '15', 'NAM-OWM', 4235],
        [2655, 'Rehoboth', '16', 'NAM-RHB', 4235],
        [2656, 'Swakopmund', '17', 'NAM-SWK', 4235],
        [2657, 'Tsumeb', '18', 'NAM-TSM', 4235],
        [2658, 'Karasburg', '20', 'NAM-KRS', 4235],
        [2659, 'Windhoek', '21', 'NAM-WND', 4235],
        [2660, 'Damaraland', '22', 'NAM-DMR', 4235],
        [2661, 'Hereroland Oos', '23', 'NAM-HOO', 4235],
        [2662, 'Hereroland Wes', '24', 'NAM-HWE', 4235],
        [2663, 'Kavango', '25', 'NAM-KVN', 4235],
        [2664, 'Mariental', '26', 'NAM-NML', 4235],
        [2665, 'Namaland', '27', 'NAM-NML', 4235],
        [2666, 'Caprivi', '28', 'NAM-COO', 4235],
        [2667, 'Erongo', '29', 'NAM-SWK', 4235],
        [2668, 'Hardap', '30', 'NAM-NML', 4235],
        [2669, 'Karas', '31', 'NAM-KRS', 4235],
        [2670, 'Kunene', '32', 'NAM-KKL', 4235],
        [2671, 'Ohangwena', '33', 'NAM-OWM', 4235],
        [2672, 'Okavango', '34', 'NAM-KVN', 4235],
        [2673, 'Omaheke', '35', 'NAM-WND', 4235],
        [2674, 'Omusati', '36', 'NAM-OWM', 4235],
        [2675, 'Oshana', '37', 'NAM-OWM', 4235],
        [2676, 'Oshikoto', '38', 'NAM-TSM', 4235],
        [2677, 'Otjozondjupa', '39', 'NAM-GRT', 4235],
        [2678, 'Agadez', '01', 'NER-AGD', 4237],
        [2679, 'Diffa', '02', 'NER-DIF', 4237],
        [2680, 'Dosso', '03', 'NER-DOS', 4237],
        [2681, 'Maradi', '04', 'NER-MAR', 4237],
        [2682, 'Niamey', '05', 'NER-NIA', 4237],
        [2683, 'Tahoua', '06', 'NER-TAH', 4237],
        [2684, 'Zinder', '07', 'NER-ZND', 4237],
        [2685, 'Niamey', '08', 'NER-NIA', 4237],
        [2686, 'Lagos', '05', 'NI05', 4239],
        [2687, 'Rivers', '10', 'NI10', 4239],
        [
            2688,
            'Federal Capital Territory',
            '11',
            'NI11',
            4239,
            { aliases: 'Abjuba;Abuja Capital Territory' },
        ],
        [2689, 'Ogun', '16', 'NI16', 4239],
        [2690, 'Ondo', '17', 'NI17', 4239],
        [2691, 'Akwa Ibom', '21', 'NI21', 4239],
        [2692, 'Cross River', '22', 'NI22', 4239],
        [2693, 'Kaduna', '23', 'NI23', 4239],
        [2694, 'Katsina', '24', 'NI24', 4239],
        [2695, 'Anambra', '25', 'NI25', 4239],
        [2696, 'Benue', '26', 'NI26', 4239],
        [2697, 'Borno', '27', 'NI27', 4239],
        [2698, 'Imo', '28', 'NI28', 4239],
        [2699, 'Kano', '29', 'NI29', 4239],
        [2700, 'Kwara', '30', 'NI30', 4239],
        [2701, 'Niger', '31', 'NI31', 4239],
        [2702, 'Oyo', '32', 'NI32', 4239],
        [2703, 'Adamawa', '35', 'NI35', 4239],
        [2704, 'Delta', '36', 'NI36', 4239],
        [2705, 'Edo', '37', 'NI37', 4239],
        [2706, 'Jigawa', '39', 'NI39', 4239],
        [2707, 'Kebbi', '40', 'NI40', 4239],
        [2708, 'Kogi', '41', 'NI41', 4239],
        [2709, 'Osun', '42', 'NI42', 4239],
        [2710, 'Taraba', '43', 'NI43', 4239],
        [2711, 'Yobe', '44', 'NI44', 4239],
        [2712, 'Abia', '45', 'NI34', 4239],
        [2713, 'Bauchi', '46', 'NI06', 4239],
        [2714, 'Enugu', '47', 'NI38', 4239],
        [2715, 'Ondo', '48', 'NI17', 4239],
        [2716, 'Rivers', '50', 'NI10', 4239],
        [2717, 'Sokoto', '51', 'NI33', 4239],
        [2718, 'Bayelsa', '52', 'NI10', 4239],
        [2719, 'Ebonyi', '53', 'NI38', 4239],
        [2720, 'Ekiti', '54', 'NI17', 4239],
        [2721, 'Gombe', '55', 'NI06', 4239],
        [2722, 'Nassarawa', '56', 'NI19', 4239],
        [2723, 'Zamfara', '57', 'NI33', 4239],
        [2724, 'Boaco', '01', 'NIC-BOA', 4240],
        [2725, 'Carazo', '02', 'NIC-CAR', 4240],
        [2726, 'Chinandega', '03', 'NIC-CHN', 4240],
        [2727, 'Chontales', '04', 'NIC-CHT', 4240],
        [2728, 'Esteli', '05', 'NIC-EST', 4240],
        [2729, 'Granada', '06', 'NIC-GRN', 4240],
        [2730, 'Jinotega', '07', 'NIC-JNT', 4240],
        [2731, 'Leon', '08', 'NIC-LEO', 4240],
        [2732, 'Madriz', '09', 'NIC-MDR', 4240],
        [2733, 'Managua', '10', 'NIC-MAN', 4240],
        [2734, 'Masaya', '11', 'NIC-MAS', 4240],
        [2735, 'Matagalpa', '12', 'NIC-MTG', 4240],
        [2736, 'Nueva Segovia', '13', 'NIC-NSE', 4240],
        [2737, 'Rio San Juan', '14', 'NIC-RSJ', 4240],
        [2738, 'Rivas', '15', 'NIC-RIV', 4240],
        [2739, 'Zelaya', '16', 'NIC-ZEL', 4240],
        [2740, 'Drenthe', '01', 'NLD-DRN', 4241],
        [2741, 'Friesland', '02', 'NLD-FRS', 4241],
        [2742, 'Gelderland', '03', 'NLD-GLD', 4241],
        [2743, 'Groningen', '04', 'NLD-GRN', 4241],
        [2744, 'Noord-Brabant', '06', 'NLD-NBR', 4241],
        [2745, 'Noord-Holland', '07', 'NLD-NHL', 4241],
        [2746, 'Overijssel', '15', 'NLD-OVR', 4241],
        [2747, 'Utrecht', '09', 'NLD-UTR', 4241],
        [2748, 'Zeeland', '10', 'NLD-ZLN', 4241],
        [2749, 'Zuid-Holland', '11', 'NLD-ZHL', 4241],
        [2750, 'Dronten', '12', 'NLD-FLV', 4241, { notes: 'City, Provice 2753' }],
        [2751, 'Zuidelijke IJsselmeerpolders', '13', 'NLD-FLV', 4241],
        [2752, 'Lelystad', '14', 'NLD-FLV', 4241, { notes: 'City, Provice 2753' }],
        [2753, 'Flevoland', '16', 'NLD-FLV', 4241],
        [2754, 'Akershus', '01', 'NOR-AKR', 4242],
        [2755, 'Aust-Agder', '02', 'NOR-AAG', 4242],
        [2756, 'Buskerud', '04', 'NOR-BSK', 4242],
        [2757, 'Finnmark', '05', 'NOR-FNN', 4242],
        [2758, 'Hedmark', '06', 'NOR-HDM', 4242],
        [2759, 'Hordaland', '07', 'NOR-HRD', 4242],
        [2760, 'More og Romsdal', '08', 'NOR-MOR', 4242],
        [2761, 'Nordland', '09', 'NOR-NRD', 4242],
        [2762, 'Nord-Trondelag', '10', 'NOR-NTR', 4242],
        [2763, 'Oppland', '11', 'NOR-OPP', 4242],
        [2764, 'Oslo', '12', 'NOR-OSL', 4242],
        [2765, 'Ostfold', '13', 'NOR-OST', 4242],
        [2766, 'Rogaland', '14', 'NOR-RGL', 4242],
        [2767, 'Sogn og Fjordane', '15', 'NOR-SOF', 4242],
        [2768, 'Sor-Trondelag', '16', 'NOR-STR', 4242],
        [2769, 'Telemark', '17', 'NOR-TLM', 4242],
        [2770, 'Troms', '18', 'NOR-TRM', 4242],
        [2771, 'Vest-Agder', '19', 'NOR-VAG', 4242],
        [2772, 'Vestfold', '20', 'NOR-VST', 4242],
        [2773, 'Bagmati', '01', 'NPL-BGM', 4243],
        [2774, 'Bheri', '02', 'NPL-BHE', 4243],
        [2775, 'Dhawalagiri', '03', 'NPL-DHW', 4243],
        [2776, 'Gandaki', '04', 'NPL-GND', 4243],
        [2777, 'Janakpur', '05', 'NPL-JNK', 4243],
        [2778, 'Karnali', '06', 'NPL-KRN', 4243],
        [2779, 'Kosi', '07', 'NPL-KOS', 4243],
        [2780, 'Lumbini', '08', 'NPL-LMB', 4243],
        [2781, 'Mahakali', '09', 'NPL-MHK', 4243],
        [2782, 'Mechi', '10', 'NPL-MEC', 4243],
        [2783, 'Narayani', '11', 'NPL-NRY', 4243],
        [2784, 'Rapti', '12', 'NPL-RAP', 4243],
        [2785, 'Sagarmatha', '13', 'NPL-SGR', 4243],
        [2786, 'Seti', '14', 'NPL-SET', 4243],
        [2787, 'Aiwo', '01', , 4244],
        [2788, 'Anabar', '02', , 4244],
        [2789, 'Anetan', '03', , 4244],
        [2790, 'Anibare', '04', , 4244],
        [2791, 'Baiti', '05', , 4244],
        [2792, 'Boe', '06', , 4244],
        [2793, 'Buada', '07', , 4244],
        [2794, 'Denigomodu', '08', , 4244],
        [2795, 'Ewa', '09', , 4244],
        [2796, 'Ijuw', '10', , 4244],
        [2797, 'Meneng', '11', , 4244],
        [2798, 'Nibok', '12', , 4244],
        [2799, 'Uaboe', '13', , 4244],
        [2800, 'Yaren', '14', , 4244],
        [2801, 'Chatham Islands', '10', '', 4246],
        [2802, 'Auckland', 'E7', 'NZ', 4246],
        [2803, 'Bay of Plenty', 'E8', 'NZ', 4246],
        [2804, 'Canterbury', 'E9', 'NZ', 4246],
        [2805, 'Gisborne', 'F1', 'NZ', 4246],
        [2806, "Hawke's Bay", 'F2', 'NZ31', 4246],
        [2807, 'Manawatu-Wanganui', 'F3', 'NZ', 4246],
        [2808, 'Marlborough', 'F4', 'NZ', 4246],
        [2809, 'Nelson', 'F5', 'NZ', 4246],
        [2810, 'Northland', 'F6', 'NZ', 4246],
        [2811, 'Otago', 'F7', 'NZ', 4246],
        [2812, 'Southland', 'F8', 'NZ72', 4246],
        [2813, 'Taranaki', 'F9', 'NZ76', 4246],
        [2814, 'Waikato', 'G1', 'NZ85', 4246],
        [2815, 'Wellington', 'G2', 'NZ', 4246],
        [2816, 'West Coast', 'G3', 'NZB3', 4246],
        [2817, 'Ad Dakhiliyah', '01', , 4247],
        [2818, 'Al Batinah', '02', , 4247],
        [2819, 'Az Zahirah', '05', , 4247],
        [2820, 'Masqat', '06', , 4247],
        [2821, 'Musandam', '07', , 4247],
        [2822, 'Zufar', '08', , 4247],
        [2823, 'Bocas del Toro', '01', 'PAN-BDT', 4248],
        [2824, 'Chiriqui', '02', 'PAN-CHR', 4248],
        [2825, 'Cocle', '03', 'PAN-COC', 4248],
        [2826, 'Darien', '05', 'PAN-DAR', 4248],
        [2827, 'Herrera', '06', 'PAN-HRR', 4248],
        [2828, 'Los Santos', '07', 'PAN-LSN', 4248],
        [2829, 'Panama Province', '08', 'PAN-PAN', 4248, { aliases: 'Panama' }],
        [2830, 'San Blas', '09', 'PAN-SBL', 4248],
        [2831, 'Veraguas', '10', 'PAN-VRG', 4248],
        [2832, 'Ancash', '02', 'PER-ANC', 4249],
        [2833, 'Apurimac', '03', 'PER-APR', 4249],
        [2834, 'Arequipa', '04', 'PER-ARQ', 4249],
        [2835, 'Ayacucho', '05', 'PER-AYC', 4249],
        [2836, 'Cajamarca', '06', 'PER-CJM', 4249],
        [2837, 'Callao', '07', 'PER-CAL', 4249],
        [2838, 'Cusco', '08', 'PER-CUS', 4249],
        [2839, 'Huancavelica', '09', 'PER-HNC', 4249],
        [2840, 'Huanuco', '10', 'PER-HUA', 4249],
        [2841, 'Ica', '11', 'PER-ICA', 4249],
        [2842, 'Junin', '12', 'PER-JUN', 4249],
        [2843, 'La Libertad', '05', 'SLV-LLB', 4284],
        [2844, 'Lambayeque', '14', 'PER-LMB', 4249],
        [2845, 'Lima', '15', 'PER-LIM', 4249],
        [2846, 'Loreto', '16', 'PER-LOR', 4249],
        [2847, 'Madre de Dios', '17', 'PER-MDD', 4249],
        [2848, 'Moquegua', '18', 'PER-MOQ', 4249],
        [2849, 'Pasco', '19', 'PER-PAS', 4249],
        [2850, 'Piura', '20', 'PER-PIU', 4249],
        [2851, 'Puno', '21', 'PER-PUN', 4249],
        [2852, 'San Martin', '22', 'PER-SMR', 4249],
        [2853, 'Tacna', '23', 'PER-TAC', 4249],
        [2854, 'Tumbes', '24', 'PER-TMB', 4249],
        [2855, 'Ucayali', '25', 'PER-UCY', 4249],
        [2856, 'Gulf', '02', 'PP02', 4251],
        [2857, 'Milne Bay', '03', 'PP03', 4251],
        [2858, 'Southern Highlands', '05', 'PP05', 4251],
        [2859, 'North Solomons', '07', 'PP07', 4251],
        [2860, 'Chimbu', '08', 'PP08', 4251],
        [2861, 'Eastern Highlands', '09', 'PP09', 4251],
        [2862, 'East New Britain', '10', 'PP10', 4251],
        [2863, 'East Sepik', '11', 'PP11', 4251],
        [2864, 'Madang', '12', 'PP12', 4251],
        [2865, 'Manus', '13', 'PP13', 4251],
        [2866, 'Morobe', '14', 'PP14', 4251],
        [2867, 'New Ireland', '15', 'PP15', 4251],
        [2868, 'Western Highlands', '16', 'PP16', 4251],
        [2869, 'West New Britain', '17', 'PP17', 4251],
        [2870, 'Sandaun', '18', 'PP', 4251, { aliases: 'West Sepik' }],
        [2871, 'Enga', '19', 'PP19', 4251],
        [2872, 'National Capital', '20', 'PP20', 4251],
        [2873, 'Abra', '01', 'PHL-CAR', 4252],
        [2874, 'Agusan del Norte', '02', 'PHL-NMN', 4252],
        [2875, 'Agusan del Sur', '03', 'PHL-NMN', 4252],
        [2876, 'Aklan', '04', 'PHL-WVS', 4252],
        [2877, 'Albay', '05', 'PHL-BCL', 4252],
        [2878, 'Antique', '06', 'PHL-WVS', 4252],
        [2879, 'Bataan', '07', 'PHL-CLZ', 4252],
        [2880, 'Batanes', '08', 'PHL-CGV', 4252],
        [2881, 'Batangas', '09', 'PHL-STG', 4252],
        [2882, 'Benguet', '10', 'PHL-CAR', 4252],
        [2883, 'Bohol', '11', 'PHL-CVS', 4252],
        [2884, 'Bukidnon', '12', 'PHL-NMN', 4252],
        [2885, 'Bulacan', '13', 'PHL-CLZ', 4252],
        [2886, 'Cagayan', '14', 'PHL-CGV', 4252],
        [2887, 'Camarines Norte', '15', 'PHL-BCL', 4252],
        [2888, 'Camarines Sur', '16', 'PHL-BCL', 4252],
        [2889, 'Camiguin', '17', 'PHL-NMN', 4252],
        [2890, 'Capiz', '18', 'PHL-WVS', 4252],
        [2891, 'Catanduanes', '19', 'PHL-BCL', 4252],
        [2892, 'Cavite', '20', 'PHL-STG', 4252],
        [2893, 'Cebu', '21', 'PHL-CVS', 4252],
        [2894, 'Basilan', '22', 'PHL-WMN', 4252],
        [2895, 'Eastern Samar', '23', 'PHL-EVS', 4252],
        [2896, 'Davao', '24', 'PHL-SMN', 4252],
        [2897, 'Davao del Sur', '25', 'PHL-SMN', 4252],
        [2898, 'Davao Oriental', '26', 'PHL-SMN', 4252],
        [2899, 'Ifugao', '27', 'PHL-CAR', 4252],
        [2900, 'Ilocos Norte', '28', 'PHL-ILC', 4252],
        [2901, 'Ilocos Sur', '29', 'PHL-ILC', 4252],
        [2902, 'Iloilo', '30', 'PHL-WVS', 4252],
        [2903, 'Isabela', '31', 'PHL-CGV', 4252],
        [2904, 'Kalinga-Apayao', '32', 'PHL-CAR', 4252],
        [2905, 'Laguna', '33', 'PHL-STG', 4252],
        [2906, 'Lanao del Norte', '34', 'PHL-NMN', 4252],
        [2907, 'Lanao del Sur', '35', 'PHL-CMN', 4252],
        [2908, 'La Union', '07', 'SLV-LUN', 4284],
        [2909, 'Leyte', '37', 'PHL-EVS', 4252],
        [2910, 'Marinduque', '38', 'PHL-STG', 4252],
        [2911, 'Masbate', '39', 'PHL-BCL', 4252],
        [2912, 'Mindoro Occidental', '40', 'PHL-STG', 4252],
        [2913, 'Mindoro Oriental', '41', 'PHL-STG', 4252],
        [2914, 'Misamis Occidental', '42', 'PHL-NMN', 4252],
        [2915, 'Misamis Oriental', '43', 'PHL-NMN', 4252],
        [2916, 'Mountain', '44', 'PHL-CAR', 4252],
        [2917, 'Negros Oriental', '46', 'PHL-CVS', 4252],
        [2918, 'Nueva Ecija', '47', 'PHL-CLZ', 4252],
        [2919, 'Nueva Vizcaya', '48', 'PHL-CGV', 4252],
        [2920, 'Palawan', '49', 'PHL-STG', 4252],
        [2921, 'Pampanga', '50', 'PHL-CLZ', 4252],
        [2922, 'Pangasinan', '51', 'PHL-ILC', 4252],
        [2923, 'Rizal', '53', 'PHL-STG', 4252],
        [2924, 'Romblon', '54', 'PHL-STG', 4252],
        [2925, 'Samar', '55', 'PHL-EVS', 4252],
        [2926, 'Maguindanao', '56', 'PHL-CMN', 4252],
        [2927, 'North Cotabato', '57', 'PHL-CMN', 4252],
        [2928, 'Sorsogon', '58', 'PHL-BCL', 4252],
        [2929, 'Southern Leyte', '59', 'PHL-EVS', 4252],
        [2930, 'Sulu', '60', 'PHL-WMN', 4252],
        [2931, 'Surigao del Norte', '61', 'PHL-NMN', 4252],
        [2932, 'Surigao del Sur', '62', 'PHL-SMN', 4252],
        [2933, 'Tarlac', '63', 'PHL-CLZ', 4252],
        [2934, 'Zambales', '64', 'PHL-CLZ', 4252],
        [2935, 'Zamboanga del Norte', '65', 'PHL-WMN', 4252],
        [2936, 'Zamboanga del Sur', '66', 'PHL-WMN', 4252],
        [2937, 'Northern Samar', '67', 'PHL-EVS', 4252],
        [2938, 'Quirino', '68', 'PHL-CGV', 4252],
        [2939, 'Siquijor', '69', 'PHL-CVS', 4252],
        [2940, 'South Cotabato', '70', 'PHL-SMN', 4252],
        [2941, 'Sultan Kudarat', '71', 'PHL-SMN', 4252],
        [2942, 'Tawitawi', '72', 'PHL-WMN', 4252],
        [2943, 'Angeles', 'A1', 'PHL-CLZ', 4252],
        [2944, 'Bacolod', 'A2', 'PHL-WVS', 4252],
        [2945, 'Bago', 'A3', 'PHL-WVS', 4252],
        [2946, 'Baguio', 'A4', 'PHL-CAR', 4252],
        [2947, 'Bais', 'A5', 'PHL-CVS', 4252],
        [2948, 'Basilan City', 'A6', 'PHL-WMN', 4252],
        [2949, 'Batangas City', 'A7', 'PHL-STG', 4252],
        [2950, 'Butuan', 'A8', 'PHL-NMN', 4252],
        [2951, 'Cabanatuan', 'A9', 'PHL-CLZ', 4252],
        [2952, 'Cadiz', 'B1', 'PHL-WVS', 4252],
        [2953, 'Cagayan de Oro', 'B2', 'PHL-NMN', 4252],
        [2954, 'Calbayog', 'B3', 'PHL-EVS', 4252],
        [2955, 'Caloocan', 'B4', 'PHL-NCR', 4252],
        [2956, 'Canlaon', 'B5', 'PHL-CVS', 4252],
        [2957, 'Cavite City', 'B6', 'PHL-STG', 4252],
        [2958, 'Cebu City', 'B7', 'PHL-CVS', 4252],
        [2959, 'Cotabato', 'B8', 'PHL-CMN', 4252],
        [2960, 'Dagupan', 'B9', 'PHL-ILC', 4252],
        [2961, 'Danao', 'C1', 'PHL-CVS', 4252],
        [2962, 'Dapitan', 'C2', 'PHL-WMN', 4252],
        [2963, 'Davao City', 'C3', 'PHL-SMN', 4252],
        [2964, 'Dipolog', 'C4', 'PHL-WMN', 4252],
        [2965, 'Dumaguete', 'C5', 'PHL-CVS', 4252],
        [2966, 'General Santos', 'C6', 'PHL-SMN', 4252],
        [2967, 'Gingoog', 'C7', 'PHL-NMN', 4252],
        [2968, 'Iligan', 'C8', 'PHL-NMN', 4252],
        [2969, 'Iloilo City', 'C9', 'PHL-WVS', 4252],
        [2970, 'Iriga', 'D1', 'PHL-BCL', 4252],
        [2971, 'La Carlota', 'D2', 'PHL-WVS', 4252],
        [2972, 'Laoag', 'D3', 'PHL-ILC', 4252],
        [2973, 'Lapu-Lapu', 'D4', 'PHL-CVS', 4252],
        [2974, 'Legaspi', 'D5', 'PHL-BCL', 4252],
        [2975, 'Lipa', 'D6', 'PHL-STG', 4252],
        [2976, 'Lucena', 'D7', 'PHL-STG', 4252],
        [2977, 'Mandaue', 'D8', 'PHL-CVS', 4252],
        [2978, 'Manila', 'D9', 'PHL-NCR', 4252],
        [2979, 'Marawi', 'E1', 'PHL-CMN', 4252],
        [2980, 'Naga', 'E2', 'PHL-BCL', 4252],
        [2981, 'Olongapo', 'E3', 'PHL-CLZ', 4252],
        [2982, 'Ormoc', 'E4', 'PHL-EVS', 4252],
        [2983, 'Oroquieta', 'E5', 'PHL-NMN', 4252],
        [2984, 'Ozamis', 'E6', 'PHL-NMN', 4252],
        [2985, 'Pagadian', 'E7', 'PHL-WMN', 4252],
        [2986, 'Palayan', 'E8', 'PHL-CLZ', 4252],
        [2987, 'Pasay', 'E9', 'PHL-NCR', 4252],
        [2988, 'Puerto Princesa', 'F1', 'PHL-STG', 4252],
        [2989, 'Quezon City', 'F2', 'PHL-STG', 4252],
        [2990, 'Roxas', 'F3', 'PHL-WVS', 4252],
        [2991, 'San Carlos', 'F4', 'PHL-WVS', 4252, { notes: 'Negros Occidental' }],
        [2992, 'San Carlos', 'F5', 'PHL-ILC', 4252, { notes: 'Pangasinan' }],
        [2993, 'San Pablo', 'F7', 'PHL-STG', 4252],
        [2994, 'Silay', 'F8', 'PHL-WVS', 4252],
        [2995, 'Surigao', 'F9', 'PHL-NMN', 4252],
        [2996, 'Tacloban', 'G1', 'PHL-EVS', 4252],
        [2997, 'Tagaytay', 'G2', 'PHL-STG', 4252],
        [2998, 'Tagbilaran', 'G3', 'PHL-CVS', 4252],
        [2999, 'Tangub', 'G4', 'PHL-NMN', 4252],
        [3000, 'Trece Martires', 'G6', 'PHL-STG', 4252],
        [3001, 'Zamboanga', 'G7', 'PHL-WMN', 4252],
        [3002, 'Aurora', 'G8', 'PHL-CLZ', 4252],
        [3003, 'Quezon', 'H2', 'PHL-STG', 4252],
        [3004, 'Negros Occidental', 'H3', 'PHL-WVS', 4252],
        [3005, 'Federally Administered Tribal Areas', '01', 'PAK-FAT', 4253],
        [3006, 'Balochistan', '02', 'PAK-BLC', 4253],
        [3007, 'North-West Frontier', '03', 'PAK-NWF', 4253],
        [3008, 'Sindh', '05', 'PAK-SIN', 4253],
        [3009, 'Azad Kashmir', '06', 'PAK-AKP', 4253],
        [3010, 'Northern Areas', '07', 'PAK-NOR', 4253],
        [3011, 'Islamabad', '08', 'PAK-PNJ', 4253],
        [
            3012,
            'Biala Podlaska',
            '23',
            'POL-BPD',
            4254,
            { notes: 'City, Province 3064' },
        ],
        [
            3013,
            'Bialystok',
            '24',
            'POL-BLY',
            4254,
            { notes: 'City, Province 3070' },
        ],
        [3014, 'Bielsko', '25', 'POL-BLS', 4254, { notes: 'City, Province 3072' }],
        [
            3015,
            'Bydgoszcz',
            '26',
            'POL-BYD',
            4254,
            { notes: 'City, Province 3061' },
        ],
        [3016, 'Chelm', '27', 'POL-CHL', 4254, { notes: 'City, Province 3064' }],
        [
            3017,
            'Ciechanow',
            '28',
            'POL-CCH',
            4254,
            { notes: 'City, Province 3067' },
        ],
        [
            3018,
            'Czestochowa',
            '29',
            'POL-CZS',
            4254,
            { notes: 'City, Province 3072' },
        ],
        [3019, 'Elblag', '30', 'POL-ELB', 4254, { notes: 'City, Province 3074' }],
        [3020, 'Gdansk', '31', 'POL-GDN', 4254, { notes: 'City, Province 3071' }],
        [3021, 'Gorzow', '32', 'POL-GRZ', 4254, { notes: 'City, Province 3065' }],
        [
            3022,
            'Jelenia Gora',
            '33',
            'POL-JGO',
            4254,
            { notes: 'City, Province 3061' },
        ],
        [3023, 'Kalisz', '34', 'POL-KLS', 4254, { notes: 'City, Province 3075' }],
        [3024, 'Katowice', '35', 'POL-KTW', 4254, { notes: 'City, Province 3072' }],
        [3025, 'Kielce', '36', 'POL-KIE', 4254, { notes: 'City, Province 3073' }],
        [3026, 'Konin', '37', 'POL-KON', 4254, { notes: 'City, Province 3075' }],
        [3027, 'Koszalin', '38', 'POL-KSZ', 4254, { notes: 'City, Province 3076' }],
        [3028, 'Krakow', '39', 'POL-KRK', 4254, { notes: 'City, Province 3066' }],
        [3029, 'Krosno', '40', 'POL-KRS', 4254, { notes: 'City, Province 3069' }],
        [3030, 'Legnica', '41', 'POL-LGN', 4254, { notes: 'City, Province 3061' }],
        [3031, 'Leszno', '42', 'POL-LSZ', 4254, { notes: 'City, Province 3075' }],
        [3032, 'Lodz', '43', 'POL-LOD', 4254, { notes: 'City, Province 3063' }],
        [3033, 'Lomza', '44', 'POL-LOM', 4254, { notes: 'City, Province 3070' }],
        [3034, 'Lublin', '45', 'POL-LBL', 4254, { notes: 'City, Province 3064' }],
        [
            3035,
            'Nowy Sacz',
            '46',
            'POL-NSA',
            4254,
            { notes: 'City, Province 3066' },
        ],
        [3036, 'Olsztyn', '47', 'POL-OLS', 4254, { notes: 'City, Province 3074' }],
        [3037, 'Opole', '48', 'POL-OPO', 4254, { notes: 'City, Province 3068' }],
        [
            3038,
            'Ostroleka',
            '49',
            'POL-OST',
            4254,
            { notes: 'City, Province 3067' },
        ],
        [3039, 'Pila', '50', 'POL-PIL', 4254, { notes: 'City, Province 3075' }],
        [3040, 'Piotrkow', '51', 'POL-PTR', 4254, { notes: 'City, Province 3063' }],
        [3041, 'Plock', '52', 'POL-PLC', 4254, { notes: 'City, Province 3067' }],
        [3042, 'Poznan', '53', 'POL-PZN', 4254, { notes: 'City, Province 3075' }],
        [3043, 'Przemysl', '54', 'POL-PRZ', 4254, { notes: 'City, Province 3069' }],
        [3044, 'Radom', '55', 'POL-RAD', 4254, { notes: 'City, Province 3067' }],
        [3045, 'Rzeszow', '56', 'POL-RZS', 4254, { notes: 'City, Province 3069' }],
        [3046, 'Siedlce', '57', 'POL-SDL', 4254, { notes: 'City, Province 3067' }],
        [3047, 'Sieradz', '58', 'POL-SRD', 4254, { notes: 'City, Province 3063' }],
        [
            3048,
            'Skierniewice',
            '59',
            'POL-SKR',
            4254,
            { notes: 'City, Province 3063' },
        ],
        [3049, 'Slupsk', '60', 'POL-SLP', 4254, { notes: 'City, Province 3071' }],
        [3050, 'Suwalki', '61', 'POL-SWL', 4254, { notes: 'City, Province 3070' }],
        [3051, 'Szczecin', '62', 'POL-SZC', 4254, { notes: 'City, Province 3076' }],
        [
            3052,
            'Tarnobrzeg',
            '63',
            'POL-TRN',
            4254,
            { notes: 'City, Province 3069' },
        ],
        [3053, 'Tarnow', '64', 'POL-TRW', 4254, { notes: 'City, Province 3066' }],
        [3054, 'Torun', '65', 'POL-TOR', 4254, { notes: 'City, Province 3062' }],
        [
            3055,
            'Walbrzych',
            '66',
            'POL-WLB',
            4254,
            { notes: 'City, Province 3061' },
        ],
        [
            3056,
            'Warszawa',
            '67',
            'POL-WRS',
            4254,
            { aliases: 'Warsaw', notes: 'City, Province 3067' },
        ],
        [
            3057,
            'Wloclawek',
            '68',
            'POL-WLC',
            4254,
            { notes: 'City, Province 3062' },
        ],
        [3058, 'Wroclaw', '69', 'POL-WRC', 4254, { notes: 'City, Province 3061' }],
        [3059, 'Zamosc', '70', 'POL-ZMS', 4254, { notes: 'City, Province 3064' }],
        [
            3060,
            'Zielona Gora',
            '71',
            'POL-ZGO',
            4254,
            { notes: 'City, Province 3065' },
        ],
        [
            3061,
            'Dolnoslaskie',
            '72',
            'POL-WRC',
            4254,
            { aliases: 'Lower Silesian Voivodeship' },
        ],
        [
            3062,
            'Kujawsko-Pomorskie',
            '73',
            'POL-TOR',
            4254,
            { aliases: 'Kuyavian-Pomeranian Voivodeship' },
        ],
        [3063, 'Lodzkie', '74', 'POL-LOD', 4254, { aliases: 'Lodz Voivodeship' }],
        [
            3064,
            'Lubelskie',
            '75',
            'POL-LBL',
            4254,
            { aliases: 'Lublin Voivodeship' },
        ],
        [
            3065,
            'Lubuskie',
            '76',
            'POL-GRZ',
            4254,
            { aliases: 'Lubusz Voivodeship' },
        ],
        [
            3066,
            'Malopolskie',
            '77',
            'POL-KRK',
            4254,
            { aliases: 'Lesser Poland Voivodeship' },
        ],
        [
            3067,
            'Mazowieckie',
            '78',
            'POL-WRS',
            4254,
            { aliases: 'Masovian Voivodeship' },
        ],
        [3068, 'Opolskie', '79', 'POL-OPO', 4254, { aliases: 'Opole Voivodeship' }],
        [
            3069,
            'Podkarpackie',
            '80',
            'POL-RZS',
            4254,
            { aliases: 'Subcarpathian Voivodeship' },
        ],
        [
            3070,
            'Podlaskie',
            '81',
            'POL-BLY',
            4254,
            { aliases: 'Podlaskie Voivodeship' },
        ],
        [
            3071,
            'Pomorskie',
            '82',
            'POL-GDN',
            4254,
            { aliases: 'Pomeranian Voivodeship' },
        ],
        [
            3072,
            'Slaskie',
            '83',
            'POL-KTW',
            4254,
            { aliases: 'Silesian Voivodeship' },
        ],
        [
            3073,
            'Swietokrzyskie',
            '84',
            'POL-KIE',
            4254,
            { aliases: 'Swietokrzyskie Voivodeship' },
        ],
        [
            3074,
            'Warminsko-Mazurskie',
            '85',
            'POL-OLS',
            4254,
            { aliases: 'Warmian-Masurian Voivodeship' },
        ],
        [
            3075,
            'Wielkopolskie',
            '86',
            'POL-PZN',
            4254,
            { aliases: 'Greater Poland Voivodeship' },
        ],
        [
            3076,
            'Zachodniopomorskie',
            '87',
            'POL-SZC',
            4254,
            { aliases: 'West Pomeranian Voivodeship' },
        ],
        [3077, 'West Bank', 'WE', , 4258],
        [3078, 'Aveiro', '02', 'PRT-AVE', 4259],
        [3079, 'Beja', '03', 'PRT-BEJ', 4259],
        [3080, 'Braga', '04', 'PRT-BRA', 4259],
        [3081, 'Braganca', '05', 'PRT-BRG', 4259],
        [3082, 'Castelo Branco', '06', 'PRT-CBR', 4259],
        [3083, 'Coimbra', '07', 'PRT-CMB', 4259],
        [3084, 'Evora', '08', 'PRT-EVO', 4259],
        [3085, 'Faro', '09', 'PRT-FAR', 4259],
        [
            3086,
            'Madeira',
            '10',
            'PRT-LIS',
            4259,
            {
                notes: 'This is not really part of Lisboa, it is an archipeligo off the mainland.',
            },
        ],
        [3087, 'Guarda', '11', 'PRT-GUA', 4259],
        [3088, 'Leiria', '13', 'PRT-LEI', 4259],
        [3089, 'Lisboa', '14', 'PRT-LIS', 4259],
        [3090, 'Portalegre', '16', 'PRT-PRT', 4259],
        [3091, 'Porto', '17', 'PRT-POR', 4259],
        [3092, 'Santarem', '18', 'PRT-SNT', 4259],
        [3093, 'Setubal', '19', 'PRT-STB', 4259],
        [3094, 'Viana do Castelo', '20', 'PRT-VDC', 4259],
        [3095, 'Vila Real', '21', 'PRT-VRE', 4259],
        [3096, 'Viseu', '22', 'PRT-VIS', 4259],
        [3097, 'Azores', '23', 'PRT-AZR', 4259],
        [3098, 'Alto Parana', '01', 'PRY-APA', 4261],
        [3099, 'Amambay', '02', 'PRY-AMM', 4261],
        [3100, 'Boqueron', '03', 'PRY-BQR', 4261],
        [3101, 'Caaguazu', '04', 'PRY-CAA', 4261],
        [3102, 'Caazapa', '05', 'PRY-CAZ', 4261],
        [3103, 'Concepcion', '07', 'PRY-CNC', 4261],
        [3104, 'Cordillera', '08', 'PRY-CRD', 4261],
        [3105, 'Guaira', '10', 'PRY-GUA', 4261],
        [3106, 'Itapua', '11', 'PRY-ITA', 4261],
        [3107, 'Neembucu', '13', 'PRY-NMB', 4261],
        [3108, 'Paraguari', '15', 'PRY-PRG', 4261],
        [3109, 'Presidente Hayes', '16', 'PRY-PHA', 4261],
        [3110, 'San Pedro', '17', 'PRY-SPE', 4261],
        [3111, 'Canindeyu', '19', 'PRY-CNN', 4261],
        [3112, 'Nueva Asuncion', '21', 'PRY-NAS', 4261],
        [3113, 'Alto Paraguay', '23', 'PRY-APR', 4261],
        [3114, 'Ad Dawhah', '01', , 4262],
        [3115, 'Al Ghuwariyah', '02', , 4262],
        [3116, 'Al Jumaliyah', '03', , 4262],
        [3117, 'Al Khawr', '04', , 4262],
        [3118, 'Al Wakrah Municipality', '05', , 4262],
        [3119, 'Ar Rayyan', '06', , 4262],
        [3120, 'Madinat ach Shamal', '08', , 4262],
        [3121, 'Umm Salal', '09', , 4262],
        [3122, 'Al Wakrah', '10', , 4262],
        [3123, 'Jariyan al Batnah', '11', , 4262],
        [3124, "Umm Sa'id", '12', , 4262],
        [3125, 'Alba', '01', 'ROM-ALB', 4264],
        [3126, 'Arad', '02', 'ROM-ARA', 4264],
        [3127, 'Arges', '03', 'ROM-ARG', 4264],
        [3128, 'Bacau', '04', 'ROM-BAC', 4264],
        [3129, 'Bihor', '05', 'ROM-BIH', 4264],
        [3130, 'Bistrita-Nasaud', '06', 'ROM-BNA', 4264],
        [3131, 'Botosani', '07', 'ROM-BTS', 4264],
        [3132, 'Braila', '08', 'ROM-BRA', 4264],
        [3133, 'Brasov', '09', 'ROM-BRS', 4264],
        [3134, 'Bucuresti', '10', 'ROM-BUC', 4264],
        [3135, 'Buzau', '11', 'ROM-BUZ', 4264],
        [3136, 'Caras-Severin', '12', 'ROM-CSV', 4264],
        [3137, 'Cluj', '13', 'ROM-CLU', 4264],
        [3138, 'Constanta', '14', 'ROM-CNS', 4264],
        [3139, 'Covasna', '15', 'ROM-CVS', 4264],
        [3140, 'Dambovita', '16', 'ROM-DMB', 4264],
        [3141, 'Dolj', '17', 'ROM-DOL', 4264],
        [3142, 'Galati', '18', 'ROM-GAL', 4264],
        [3143, 'Gorj', '19', 'ROM-GOR', 4264],
        [3144, 'Harghita', '20', 'ROM-HRG', 4264],
        [3145, 'Hunedoara', '21', 'ROM-HND', 4264],
        [3146, 'Ialomita', '22', 'ROM-ILM', 4264],
        [3147, 'Iasi', '23', 'ROM-IAS', 4264],
        [3148, 'Maramures', '25', 'ROM-MRM', 4264],
        [3149, 'Mehedinti', '26', 'ROM-MHD', 4264],
        [3150, 'Mures', '27', 'ROM-MUR', 4264],
        [3151, 'Neamt', '28', 'ROM-NEA', 4264],
        [3152, 'Olt', '29', 'ROM-OLT', 4264],
        [3153, 'Prahova', '30', 'ROM-PRH', 4264],
        [3154, 'Salaj', '31', 'ROM-SAL', 4264],
        [3155, 'Satu Mare', '32', 'ROM-SMA', 4264],
        [3156, 'Sibiu', '33', 'ROM-SIB', 4264],
        [3157, 'Suceava', '34', 'ROM-SUC', 4264],
        [3158, 'Teleorman', '35', 'ROM-TLR', 4264],
        [3159, 'Timis', '36', 'ROM-TIM', 4264],
        [3160, 'Tulcea', '37', 'ROM-TUL', 4264],
        [3161, 'Vaslui', '38', 'ROM-VAS', 4264],
        [3162, 'Valcea', '39', 'ROM-VIL', 4264],
        [3163, 'Vrancea', '40', 'ROM-VRN', 4264],
        [3164, 'Giurgiu', '42', 'ROM-GIU', 4264],
        [3165, 'Ilfov', '43', 'ROM-BUC', 4264],
        [3166, 'Kosovo', '01', , 4265],
        [3167, 'Vojvodina', '02', , 4265],
        [3168, 'Adygeya, Republic of', '01', , 4266],
        [3169, 'Aginsky Buryatsky AO', '02', , 4266],
        [3170, 'Gorno-Altay', '03', , 4266],
        [3171, 'Altaisky krai', '04', , 4266],
        [3172, 'Amur', '05', , 4266],
        [3173, "Arkhangel'sk", '06', , 4266],
        [3174, "Astrakhan'", '07', , 4266],
        [3175, 'Bashkortostan', '08', , 4266],
        [3176, 'Belgorod', '09', , 4266],
        [3177, 'Bryansk', '10', , 4266],
        [3178, 'Buryat', '11', , 4266],
        [3179, 'Chechnya', '12', , 4266],
        [3180, 'Chelyabinsk', '13', , 4266],
        [3181, 'Chita', '14', , 4266],
        [3182, 'Chukot', '15', , 4266],
        [3183, 'Chuvashia', '16', , 4266],
        [3184, 'Dagestan', '17', , 4266],
        [3185, 'Evenk', '18', , 4266],
        [3186, 'Ingush', '19', , 4266],
        [3187, 'Irkutsk', '20', , 4266],
        [3188, 'Ivanovo', '21', , 4266],
        [3189, 'Kabardin-Balkar', '22', , 4266],
        [3190, 'Kaliningrad', '23', , 4266],
        [3191, 'Kalmyk', '24', , 4266],
        [3192, 'Kaluga', '25', , 4266],
        [3193, 'Kamchatka', '26', , 4266],
        [3194, 'Karachay-Cherkess', '27', , 4266],
        [3195, 'Karelia', '28', , 4266],
        [3196, 'Kemerovo', '29', , 4266],
        [3197, 'Khabarovsk', '30', , 4266],
        [3198, 'Khakass', '31', , 4266],
        [3199, 'Khanty-Mansiy', '32', , 4266],
        [3200, 'Kirov', '33', , 4266],
        [3201, 'Komi', '34', , 4266],
        [3202, 'Komi-Permyak', '35', , 4266],
        [3203, 'Koryak', '36', , 4266],
        [3204, 'Kostroma', '37', , 4266],
        [3205, 'Krasnodar', '38', , 4266],
        [3206, 'Krasnoyarsk', '39', , 4266],
        [3207, 'Kurgan', '40', , 4266],
        [3208, 'Kursk', '41', , 4266],
        [3209, 'Leningrad', '42', , 4266],
        [3210, 'Lipetsk', '43', , 4266],
        [3211, 'Magadan', '44', , 4266],
        [3212, 'Mariy-El', '45', , 4266],
        [3213, 'Mordovia', '46', , 4266],
        [3214, 'Moskva', '47', , 4266],
        [3215, 'Moscow City', '48', , 4266],
        [3216, 'Murmansk', '49', , 4266],
        [3217, 'Nenets', '50', , 4266],
        [3218, 'Nizhegorod', '51', , 4266],
        [3219, 'Novgorod', '52', , 4266],
        [3220, 'Novosibirsk', '53', , 4266],
        [3221, 'Omsk', '54', , 4266],
        [3222, 'Orenburg', '55', , 4266],
        [3223, 'Orel', '56', , 4266],
        [3224, 'Penza', '57', , 4266],
        [3225, "Perm'", '58', , 4266],
        [3226, "Primor'ye", '59', , 4266],
        [3227, 'Pskov', '60', , 4266],
        [3228, 'Rostov', '61', , 4266],
        [3229, "Ryazan'", '62', , 4266],
        [3230, 'Sakha', '63', , 4266],
        [3231, 'Sakhalin', '64', , 4266],
        [3232, 'Samara', '65', , 4266],
        [3233, 'Saint Petersburg City', '66', , 4266],
        [3234, 'Saratov', '67', , 4266],
        [3235, 'North Ossetia', '68', , 4266],
        [3236, 'Smolensk', '69', , 4266],
        [3237, "Stavropol'", '70', , 4266],
        [3238, 'Sverdlovsk', '71', , 4266],
        [3239, 'Tambovskaya oblast', '72', , 4266],
        [3240, 'Tatarstan', '73', , 4266],
        [3241, 'Taymyr', '74', , 4266],
        [3242, 'Tomsk', '75', , 4266],
        [3243, 'Tula', '76', , 4266],
        [3244, "Tver'", '77', , 4266],
        [3245, "Tyumen'", '78', , 4266],
        [3246, 'Tuva', '79', , 4266],
        [3247, 'Udmurt', '80', , 4266],
        [3248, "Ul'yanovsk", '81', , 4266],
        [3249, 'Ust-Orda Buryat', '82', , 4266],
        [3250, 'Vladimir', '83', , 4266],
        [3251, 'Volgograd', '84', , 4266],
        [3252, 'Vologda', '85', , 4266],
        [3253, 'Voronezh', '86', , 4266],
        [3254, 'Yamal-Nenets', '87', , 4266],
        [3255, "Yaroslavl'", '88', , 4266],
        [3256, 'Yevrey', '89', , 4266],
        [3257, 'Permskiy Kray', '90', , 4266],
        [3258, 'Krasnoyarskiy Kray', '91', , 4266],
        [3259, 'Butare', '01', 'RWA-BUT', 4267],
        [3260, 'Gitarama', '06', 'RWA-GTR', 4267],
        [3261, 'Kigali', '09', 'RWA-KIG', 4267],
        [3262, 'Kigali', '12', 'RWA-KIG', 4267],
        [3263, 'Al Bahah', '02', 'SAU-ABA', 4268],
        [3264, 'Al Jawf', '21', 'YEM-AJW', 4318],
        [3265, 'Al Madinah', '05', 'SAU-AMD', 4268],
        [3266, 'Al Qasim', '08', 'SAU-AQA', 4268],
        [3267, 'Al Qurayyat', '09', 'SAU-AQR', 4268],
        [3268, 'Ar Riyad', '10', 'SAU-ARI', 4268],
        [3269, "Ha'il", '13', 'SAU-HAI', 4268],
        [3270, 'Makkah', '14', 'SAU-MKK', 4268],
        [3271, 'Al Hudud ash Shamaliyah', '15', 'SAU-AHA', 4268],
        [3272, 'Najran', '16', 'SAU-NJR', 4268],
        [3273, 'Jizan', '17', 'SAU-JIZ', 4268],
        [3274, 'Tabuk', '19', 'SAU-TAB', 4268],
        [3275, 'Al Jawf', '20', 'SAU-AJA', 4268],
        [3276, 'Malaita', '03', , 4269],
        [3277, 'Guadalcanal', '06', , 4269],
        [3278, 'Isabel', '07', , 4269],
        [3279, 'Makira', '08', , 4269],
        [3280, 'Temotu', '09', , 4269],
        [3281, 'Rennell and Bellona', '13', , 4269],
        [3282, 'Anse aux Pins', '01', , 4270],
        [3283, 'Anse Boileau', '02', , 4270],
        [3284, 'Anse Etoile', '03', , 4270],
        [3285, 'Anse Louis', '04', , 4270],
        [3286, 'Anse Royale', '05', , 4270],
        [3287, 'Baie Lazare', '06', , 4270],
        [3288, 'Baie Sainte Anne', '07', , 4270],
        [3289, 'Beau Vallon', '08', , 4270],
        [3290, 'Bel Air', '09', , 4270],
        [3291, 'Bel Ombre', '10', , 4270],
        [3292, 'Cascade', '11', , 4270],
        [3293, 'Glacis', '12', , 4270],
        [3294, 'La Digue', '15', , 4270],
        [3295, 'La Riviere Anglaise', '16', , 4270],
        [3296, 'Mont Buxton', '17', , 4270],
        [3297, 'Mont Fleuri', '18', , 4270],
        [3298, 'Plaisance', '19', , 4270],
        [3299, 'Pointe La Rue', '20', , 4270],
        [3300, 'Port Glaud', '21', , 4270],
        [3301, 'Saint Louis', '22', , 4270],
        [3302, 'Takamaka', '23', , 4270],
        [3303, "Al Istiwa'iyah", '28', 'SDN-AIS', 4271],
        [3304, 'Al Khartum', '29', 'SDN-AKH', 4271],
        [3305, 'Bahr al Ghazal', '32', 'SDN-BAG', 4271],
        [3306, 'Darfur', '33', 'SDN-DRF', 4271],
        [3307, 'Kurdufan', '34', 'SDN-KRD', 4271],
        [3308, 'Upper Nile', '35', 'SDN-AAN', 4271, { aliases: "A'ali an Nil" }],
        [3309, 'Alvsborgs Lan', '01', 'SWE-ALV', 4272],
        [3310, 'Blekinge Lan', '02', 'SWE-BLK', 4272],
        [3311, 'Gavleborgs Lan', '03', 'SWE-GVL', 4272],
        [3312, 'Goteborgs och Bohus Lan', '04', 'SWE-GOB', 4272],
        [3313, 'Gotlands Lan', '05', 'SWE-GTL', 4272],
        [3314, 'Hallands Lan', '06', 'SWE-HLL', 4272],
        [3315, 'Jamtlands Lan', '07', 'SWE-JMT', 4272],
        [3316, 'Jonkopings Lan', '08', 'SWE-JNK', 4272],
        [3317, 'Kalmar Lan', '09', 'SWE-KLM', 4272],
        [3318, 'Dalarnas Lan', '10', 'SWE-KPP', 4272, { aliases: 'Kopparbergs' }],
        [3319, 'Kristianstads Lan', '11', 'SWE-KRS', 4272],
        [3320, 'Kronobergs Lan', '12', 'SWE-KRN', 4272],
        [3321, 'Malmohus Lan', '13', 'SWE-MLM', 4272],
        [3322, 'Norrbottens Lan', '14', 'SWE-NRR', 4272],
        [3323, 'Orebro Lan', '15', 'SWE-OBR', 4272],
        [3324, 'Ostergotlands Lan', '16', 'SWE-OST', 4272],
        [3325, 'Skaraborgs Lan', '17', 'SWE-SKR', 4272],
        [3326, 'Sodermanlands Lan', '18', 'SWE-SDR', 4272],
        [3327, 'Uppsala Lan', '21', 'SWE-UPP', 4272],
        [3328, 'Varmlands Lan', '22', 'SWE-VRM', 4272],
        [3329, 'Vasterbottens Lan', '23', 'SWE-VST', 4272],
        [3330, 'Vasternorrlands Lan', '24', 'SWE-VSR', 4272],
        [3331, 'Vastmanlands Lan', '25', 'SWE-VSM', 4272],
        [3332, 'Stockholms Lan', '26', 'SWE-STC', 4272],
        [
            3333,
            'Skane Lan',
            '27',
            'SWE-MLM',
            4272,
            { notes: 'County consisting of Kristianstads & Malmohus.' },
        ],
        [
            3334,
            'Vastra Gotaland',
            '28',
            'SWE-GOB',
            4272,
            {
                notes: 'New county formed in 1988 consisting of Alvsborgs, Goteborgs och Bohus Lan, and Skaraborgs.',
            },
        ],
        [3335, 'Ascension', '01', , 4274],
        [3336, 'Saint Helena', '02', , 4274],
        [3337, 'Tristan da Cunha', '03', , 4274],
        [3338, 'Ajdovscina', '01', , 4275],
        [3339, 'Beltinci', '02', , 4275],
        [3340, 'Bled', '03', , 4275],
        [3341, 'Bohinj', '04', , 4275],
        [3342, 'Borovnica', '05', , 4275],
        [3343, 'Bovec', '06', , 4275],
        [3344, 'Brda', '07', , 4275],
        [3345, 'Brezice', '08', , 4275],
        [3346, 'Brezovica', '09', , 4275],
        [3347, 'Celje', '11', , 4275],
        [3348, 'Cerklje na Gorenjskem', '12', , 4275],
        [3349, 'Cerknica', '13', , 4275],
        [3350, 'Cerkno', '14', , 4275],
        [3351, 'Crensovci', '15', , 4275],
        [3352, 'Crna na Koroskem', '16', , 4275],
        [3353, 'Crnomelj', '17', , 4275],
        [3354, 'Divaca', '19', , 4275],
        [3355, 'Dobrepolje', '20', , 4275],
        [3356, 'Dol pri Ljubljani', '22', , 4275],
        [3357, 'Dornava', '24', , 4275],
        [3358, 'Dravograd', '25', , 4275],
        [3359, 'Duplek', '26', , 4275],
        [3360, 'Gorenja Vas-Poljane', '27', , 4275],
        [3361, 'Gorisnica', '28', , 4275],
        [3362, 'Gornja Radgona', '29', , 4275],
        [3363, 'Gornji Grad', '30', , 4275],
        [3364, 'Gornji Petrovci', '31', , 4275],
        [3365, 'Grosuplje', '32', , 4275],
        [3366, 'Hrastnik', '34', , 4275],
        [3367, 'Hrpelje-Kozina', '35', , 4275],
        [3368, 'Idrija', '36', , 4275],
        [3369, 'Ig', '37', , 4275],
        [3370, 'Ilirska Bistrica', '38', , 4275],
        [3371, 'Ivancna Gorica', '39', , 4275],
        [3372, 'Izola-Isola', '40', , 4275],
        [3373, 'Jursinci', '42', , 4275],
        [3374, 'Kanal', '44', , 4275],
        [3375, 'Kidricevo', '45', , 4275],
        [3376, 'Kobarid', '46', , 4275],
        [3377, 'Kobilje', '47', , 4275],
        [3378, 'Komen', '49', , 4275],
        [3379, 'Koper-Capodistria', '50', , 4275],
        [3380, 'Kozje', '51', , 4275],
        [3381, 'Kranj', '52', , 4275],
        [3382, 'Kranjska Gora', '53', , 4275],
        [3383, 'Krsko', '54', , 4275],
        [3384, 'Kungota', '55', , 4275],
        [3385, 'Lasko', '57', , 4275],
        [3386, 'Ljubljana', '61', , 4275],
        [3387, 'Ljubno', '62', , 4275],
        [3388, 'Logatec', '64', , 4275],
        [3389, 'Loski Potok', '66', , 4275],
        [3390, 'Lukovica', '68', , 4275],
        [3391, 'Medvode', '71', , 4275],
        [3392, 'Menges', '72', , 4275],
        [3393, 'Metlika', '73', , 4275],
        [3394, 'Mezica', '74', , 4275],
        [3395, 'Mislinja', '76', , 4275],
        [3396, 'Moravce', '77', , 4275],
        [3397, 'Moravske Toplice', '78', , 4275],
        [3398, 'Mozirje', '79', , 4275],
        [3399, 'Murska Sobota', '80', , 4275],
        [3400, 'Muta', '81', , 4275],
        [3401, 'Naklo', '82', , 4275],
        [3402, 'Nazarje', '83', , 4275],
        [3403, 'Nova Gorica', '84', , 4275],
        [3404, 'Odranci', '86', , 4275],
        [3405, 'Ormoz', '87', , 4275],
        [3406, 'Osilnica', '88', , 4275],
        [3407, 'Pesnica', '89', , 4275],
        [3408, 'Pivka', '91', , 4275],
        [3409, 'Podcetrtek', '92', , 4275],
        [3410, 'Postojna', '94', , 4275],
        [3411, 'Puconci', '97', , 4275],
        [3412, 'Racam', '98', , 4275],
        [3413, 'Radece', '99', , 4275],
        [3414, 'Radenci', 'A1', , 4275],
        [3415, 'Radlje ob Dravi', 'A2', , 4275],
        [3416, 'Radovljica', 'A3', , 4275],
        [3417, 'Rogasovci', 'A6', , 4275],
        [3418, 'Rogaska Slatina', 'A7', , 4275],
        [3419, 'Rogatec', 'A8', , 4275],
        [3420, 'Semic', 'B1', , 4275],
        [3421, 'Sencur', 'B2', , 4275],
        [3422, 'Sentilj', 'B3', , 4275],
        [3423, 'Sentjernej', 'B4', , 4275],
        [3424, 'Sevnica', 'B6', , 4275],
        [3425, 'Sezana', 'B7', , 4275],
        [3426, 'Skocjan', 'B8', , 4275],
        [3427, 'Skofja Loka', 'B9', , 4275],
        [3428, 'Skofljica', 'C1', , 4275],
        [3429, 'Slovenj Gradec', 'C2', , 4275],
        [3430, 'Slovenske Konjice', 'C4', , 4275],
        [3431, 'Smarje pri Jelsah', 'C5', , 4275],
        [3432, 'Smartno ob Paki', 'C6', , 4275],
        [3433, 'Sostanj', 'C7', , 4275],
        [3434, 'Starse', 'C8', , 4275],
        [3435, 'Store', 'C9', , 4275],
        [3436, 'Sveti Jurij', 'D1', , 4275],
        [3437, 'Tolmin', 'D2', , 4275],
        [3438, 'Trbovlje', 'D3', , 4275],
        [3439, 'Trebnje', 'D4', , 4275],
        [3440, 'Trzic', 'D5', , 4275],
        [3441, 'Turnisce', 'D6', , 4275],
        [3442, 'Velenje', 'D7', , 4275],
        [3443, 'Velike Lasce', 'D8', , 4275],
        [3444, 'Vipava', 'E1', , 4275],
        [3445, 'Vitanje', 'E2', , 4275],
        [3446, 'Vodice', 'E3', , 4275],
        [3447, 'Vrhnika', 'E5', , 4275],
        [3448, 'Vuzenica', 'E6', , 4275],
        [3449, 'Zagorje ob Savi', 'E7', , 4275],
        [3450, 'Zavrc', 'E9', , 4275],
        [3451, 'Zelezniki', 'F1', , 4275],
        [3452, 'Ziri', 'F2', , 4275],
        [3453, 'Zrece', 'F3', , 4275],
        [3454, 'Dobrova-Horjul-Polhov Gradec', 'G4', , 4275],
        [3455, 'Domzale', 'G7', , 4275],
        [3456, 'Jesenice', 'H4', , 4275],
        [3457, 'Kamnik', 'H6', , 4275],
        [3458, 'Kocevje', 'H7', , 4275],
        [3459, 'Kuzma', 'I2', , 4275],
        [3460, 'Lenart', 'I3', , 4275],
        [3461, 'Litija', 'I5', , 4275],
        [3462, 'Ljutomer', 'I6', , 4275],
        [3463, 'Loska Dolina', 'I7', , 4275],
        [3464, 'Luce', 'I9', , 4275],
        [3465, 'Majsperk', 'J1', , 4275],
        [3466, 'Maribor', 'J2', , 4275],
        [3467, 'Miren-Kostanjevica', 'J5', , 4275],
        [3468, 'Novo Mesto', 'J7', , 4275],
        [3469, 'Piran', 'J9', , 4275],
        [3470, 'Preddvor', 'K5', , 4275],
        [3471, 'Ptuj', 'K7', , 4275],
        [3472, 'Ribnica', 'L1', , 4275],
        [3473, 'Sentjur pri Celju', 'L7', , 4275],
        [3474, 'Slovenska Bistrica', 'L8', , 4275],
        [3475, 'Videm', 'N2', , 4275],
        [3476, 'Vojnik', 'N3', , 4275],
        [3477, 'Zalec', 'N5', , 4275],
        [3478, 'Banska Bystrica', '01', , 4277],
        [3479, 'Bratislava', '02', , 4277],
        [3480, 'Kosice', '03', , 4277],
        [3481, 'Nitra', '04', , 4277],
        [3482, 'Presov', '05', , 4277],
        [3483, 'Trencin', '06', , 4277],
        [3484, 'Trnava', '07', , 4277],
        [3485, 'Zilina', '08', , 4277],
        [3486, 'Western Area', '04', 'SLE-WAR', 4278],
        [3487, 'Acquaviva', '01', , 4279],
        [3488, 'Chiesanuova', '02', , 4279],
        [3489, 'Domagnano', '03', , 4279],
        [3490, 'Faetano', '04', , 4279],
        [3491, 'Fiorentino', '05', , 4279],
        [3492, 'Borgo Maggiore', '06', , 4279],
        [3493, 'San Marino', '07', , 4279],
        [3494, 'Monte Giardino', '08', , 4279],
        [3495, 'Serravalle', '09', , 4279],
        [3496, 'Dakar', '01', 'SEN-DKR', 4280],
        [3497, 'Diourbel', '03', 'SEN-DRB', 4280],
        [3498, 'Saint-Louis', '04', 'SEN-STL', 4280],
        [3499, 'Tambacounda', '05', 'SEN-TMB', 4280],
        [3500, 'Thies', '07', 'SEN-THI', 4280],
        [3501, 'Fatick', '09', 'SEN-FTC', 4280],
        [3502, 'Kaolack', '10', 'SEN-KLC', 4280],
        [3503, 'Kolda', '11', 'SEN-KLD', 4280],
        [3504, 'Ziguinchor', '12', 'SEN-ZGN', 4280],
        [3505, 'Louga', '13', 'SEN-LOU', 4280],
        [3506, 'Saint-Louis', '14', 'SEN-STL', 4280],
        [3507, 'Matam', '15', 'SEN-STL', 4280],
        [3508, 'Bakool', '01', 'SOM-BAK', 4281],
        [3509, 'Banaadir', '02', 'SOM-BND', 4281],
        [3510, 'Bari', '03', 'SOM-BAR', 4281],
        [3511, 'Bay', '04', 'SOM-BAY', 4281],
        [3512, 'Galguduud', '05', 'SOM-GLG', 4281],
        [3513, 'Gedo', '06', 'SOM-GED', 4281],
        [3514, 'Hiiraan', '07', 'SOM-HII', 4281],
        [3515, 'Jubbada Dhexe', '08', 'SOM-JDH', 4281],
        [3516, 'Jubbada Hoose', '09', 'SOM-JHO', 4281],
        [3517, 'Mudug', '10', 'SOM-MUD', 4281],
        [3518, 'Nugaal', '11', 'SOM-NUG', 4281],
        [3519, 'Sanaag', '12', 'SOM-SAN', 4281],
        [3520, 'Shabeellaha Dhexe', '13', 'SOM-SDH', 4281],
        [3521, 'Shabeellaha Hoose', '14', 'SOM-SHO', 4281],
        [3522, 'Woqooyi Galbeed', '16', 'SOM-WGL', 4281],
        [3523, 'Nugaal', '18', 'SOM-NUG', 4281],
        [3524, 'Togdheer', '19', 'SOM-TGD', 4281],
        [3525, 'Woqooyi Galbeed', '20', 'SOM-WGL', 4281],
        [3526, 'Awdal', '21', 'SOM-AWD', 4281],
        [3527, 'Sool', '22', 'SOM-SOL', 4281],
        [3528, 'Brokopondo', '10', 'SUR-BRK', 4282],
        [3529, 'Commewijne', '11', 'SUR-CMM', 4282],
        [3530, 'Coronie', '12', 'SUR-COR', 4282],
        [3531, 'Marowijne', '13', 'SUR-MRW', 4282],
        [3532, 'Nickerie', '14', 'SUR-NCK', 4282],
        [3533, 'Paramaribo', '16', 'SUR-PRM', 4282],
        [3534, 'Saramacca', '17', 'SUR-SRM', 4282],
        [3535, 'Sipaliwini', '18', 'SUR-SPL', 4282],
        [3536, 'Wanica', '19', 'SUR-WNC', 4282],
        [3537, 'Principe', '01', , 4283],
        [3538, 'Sao Tome', '02', , 4283],
        [3539, 'Ahuachapan', '01', 'SLV-AHC', 4284],
        [3540, 'Cabanas', '02', 'SLV-CBN', 4284],
        [3541, 'Chalatenango', '03', 'SLV-CHL', 4284],
        [3542, 'Cuscatlan', '04', 'SLV-CSC', 4284],
        [3543, 'Morazan', '08', 'SLV-MRZ', 4284],
        [3544, 'San Miguel', '09', 'SLV-SMI', 4284],
        [3545, 'San Salvador', '10', 'SLV-SSL', 4284],
        [3546, 'Santa Ana', '11', 'SLV-SAN', 4284],
        [3547, 'San Vicente', '12', 'SLV-SVC', 4284],
        [3548, 'Sonsonate', '13', 'SLV-SNS', 4284],
        [3549, 'Usulutan', '14', 'SLV-USL', 4284],
        [3550, 'Al Hasakah', '01', 'SYR-AHS', 4285],
        [3551, 'Al Ladhiqiyah', '02', 'SYR-ALD', 4285],
        [3552, 'Al Qunaytirah', '03', 'SYR-AQN', 4285],
        [3553, 'Ar Raqqah', '04', 'SYR-ARQ', 4285],
        [3554, "As Suwayda'", '05', 'SYR-ASW', 4285],
        [3555, 'Dar', '06', 'SYR-DAR', 4285],
        [3556, 'Dayr az Zawr', '07', 'SYR-DAZ', 4285],
        [3557, 'Rif Dimashq', '08', 'SYR-DMS', 4285],
        [3558, 'Halab', '09', 'SYR-HAL', 4285],
        [3559, 'Hamah', '10', 'SYR-HAM', 4285],
        [3560, 'Hims', '11', 'SYR-HIM', 4285],
        [3561, 'Idlib', '12', 'SYR-IDL', 4285],
        [3562, 'Dimashq', '13', 'SYR-DMS', 4285],
        [3563, 'Tartus', '14', 'SYR-TRT', 4285],
        [3564, 'Hhohho', '01', 'SWZ-HHH', 4286],
        [3565, 'Lubombo', '02', 'SWZ-LBM', 4286],
        [3566, 'Manzini', '03', 'SWZ-MNZ', 4286],
        [3567, 'Shiselweni', '04', 'SWZ-SHS', 4286],
        [3568, 'Batha', '01', 'TCD-BAT', 4288],
        [3569, 'Biltine', '02', 'TCD-BLT', 4288],
        [3570, 'Borkou-Ennedi-Tibesti', '03', 'TCD-BET', 4288],
        [3571, 'Chari-Baguirmi', '04', 'TCD-CBG', 4288],
        [3572, 'Guera', '05', 'TCD-GUE', 4288],
        [3573, 'Kanem', '06', 'TCD-KAN', 4288],
        [3574, 'Lac', '07', 'TCD-LAC', 4288],
        [3575, 'Logone Occidental', '08', 'TCD-LOC', 4288],
        [3576, 'Logone Oriental', '09', 'TCD-LOR', 4288],
        [3577, 'Mayo-Kebbi', '10', 'TCD-MKE', 4288],
        [3578, 'Moyen-Chari', '11', 'TCD-MCH', 4288],
        [3579, 'Ouaddai', '12', 'TCD-OUA', 4288],
        [3580, 'Salamat', '13', 'TCD-SLM', 4288],
        [3581, 'Tandjile', '14', 'TCD-TND', 4288],
        [3582, 'Lama-Kara', '09', 'TOG-BSS', 4290],
        [3583, 'Tsevie', '18', 'TOG-ZIO', 4290],
        [3584, 'Centrale', '22', 'TOG-SOT', 4290],
        [3585, 'Kara', '23', 'TOG-BSS', 4290],
        [3586, 'Maritime', '24', 'TOG-GLF', 4290],
        [3587, 'Mae Hong Son', '01', 'THA-MHS', 4291],
        [3588, 'Chiang Mai', '02', 'THA-CMA', 4291],
        [3589, 'Chiang Rai', '03', 'THA-CRA', 4291],
        [3590, 'Nan', '04', 'THA-NAN', 4291],
        [3591, 'Lamphun', '05', 'THA-LMP', 4291],
        [3592, 'Lampang', '06', 'THA-LMN', 4291],
        [3593, 'Phrae', '07', 'THA-PHR', 4291],
        [3594, 'Tak', '08', 'THA-TAK', 4291],
        [3595, 'Sukhothai', '09', 'THA-SKH', 4291],
        [3596, 'Uttaradit', '10', 'THA-UTT', 4291],
        [3597, 'Kamphaeng Phet', '11', 'THA-KPH', 4291],
        [3598, 'Phitsanulok', '12', 'THA-PHT', 4291],
        [3599, 'Phichit', '13', 'THA-PHC', 4291],
        [3600, 'Phetchabun', '14', 'THA-PHH', 4291],
        [3601, 'Uthai Thani', '15', 'THA-UTH', 4291],
        [3602, 'Nakhon Sawan', '16', 'THA-NSA', 4291],
        [3603, 'Nong Khai', '17', 'THA-NKH', 4291],
        [3604, 'Loei', '18', 'THA-LOE', 4291],
        [3605, 'Sakon Nakhon', '20', 'THA-SNK', 4291],
        [3606, 'Nakhon Phanom', '21', 'THA-NPH', 4291],
        [3607, 'Khon Kaen', '22', 'THA-KKA', 4291],
        [3608, 'Kalasin', '23', 'THA-KLS', 4291],
        [3609, 'Maha Sarakham', '24', 'THA-MSR', 4291],
        [3610, 'Roi Et', '25', 'THA-RET', 4291],
        [
            3611,
            'Chaiyaphum',
            '26',
            'THA-CHY',
            4291,
            { aliases: 'Chaiyaphum Province' },
        ],
        [3612, 'Nakhon Ratchasima', '27', 'THA-NRT', 4291],
        [3613, 'Buriram', '28', 'THA-BRR', 4291],
        [3614, 'Surin', '29', 'THA-SUR', 4291],
        [3615, 'Sisaket', '30', 'THA-SSK', 4291],
        [3616, 'Narathiwat', '31', 'THA-NRH', 4291],
        [3617, 'Chai Nat', '32', 'THA-CNA', 4291],
        [3618, 'Sing Buri', '33', 'THA-SBU', 4291],
        [3619, 'Lop Buri', '34', 'THA-LBU', 4291],
        [3620, 'Ang Thong', '35', 'THA-ATH', 4291],
        [3621, 'Phra Nakhon Si Ayutthaya', '36', 'THA-PNS', 4291],
        [3622, 'Saraburi', '37', 'THA-SRB', 4291],
        [3623, 'Nonthaburi', '38', 'THA-NNT', 4291],
        [3624, 'Pathum Thani', '39', 'THA-PTH', 4291],
        [3625, 'Krung Thep', '40', 'THA-KTH', 4291],
        [3626, 'Phayao', '41', 'THA-PHA', 4291],
        [3627, 'Samut Prakan', '42', 'THA-SPR', 4291],
        [3628, 'Nakhon Nayok', '43', 'THA-NNA', 4291],
        [3629, 'Chachoengsao', '44', 'THA-CHC', 4291],
        [3630, 'Prachin Buri', '45', 'THA-PBU', 4291],
        [3631, 'Chon Buri', '46', 'THA-CBU', 4291],
        [3632, 'Rayong', '47', 'THA-RYN', 4291],
        [3633, 'Chanthaburi', '48', 'THA-CHN', 4291],
        [3634, 'Trat', '49', 'THA-TRA', 4291],
        [3635, 'Kanchanaburi', '50', 'THA-KNC', 4291],
        [3636, 'Suphan Buri', '51', 'THA-SBR', 4291],
        [3637, 'Ratchaburi', '52', 'THA-RTC', 4291],
        [3638, 'Nakhon Pathom', '53', 'THA-NPT', 4291],
        [3639, 'Samut Songkhram', '54', 'THA-SSN', 4291],
        [3640, 'Samut Sakhon', '55', 'THA-SSH', 4291],
        [3641, 'Phetchaburi', '56', 'THA-PHB', 4291],
        [3642, 'Prachuap Khiri Khan', '57', 'THA-PKK', 4291],
        [3643, 'Chumphon', '58', 'THA-CHM', 4291],
        [3644, 'Ranong', '59', 'THA-RNN', 4291],
        [3645, 'Surat Thani', '60', 'THA-STH', 4291],
        [3646, 'Phangnga', '61', 'THA-PHN', 4291],
        [3647, 'Phuket', '62', 'THA-PHK', 4291],
        [3648, 'Krabi', '63', 'THA-KRA', 4291],
        [3649, 'Nakhon Si Thammarat', '64', 'THA-NST', 4291],
        [3650, 'Trang', '65', 'THA-TRN', 4291],
        [3651, 'Phatthalung', '66', 'THA-PHL', 4291],
        [3652, 'Satun', '67', 'THA-SAT', 4291],
        [3653, 'Songkhla', '68', 'THA-SNG', 4291],
        [3654, 'Pattani', '69', 'THA-PTT', 4291],
        [3655, 'Yala', '70', 'THA-YAL', 4291],
        [3656, 'Ubon Ratchathani', '71', 'THA-URT', 4291],
        [3657, 'Yasothon', '72', 'THA-YST', 4291],
        [3658, 'Ubon Ratchathani', '75', 'THA-URT', 4291],
        [3659, 'Udon Thani', '76', 'THA-UTA', 4291],
        [3660, 'Mukdahan', '78', 'THA-NPH', 4291],
        [3661, 'Kuhistoni Badakhshon', '01', , 4292],
        [3662, 'Khatlon', '02', , 4292],
        [3663, 'Sughd', '03', , 4292],
        [3664, 'Ahal', '01', , 4295],
        [3665, 'Balkan', '02', , 4295],
        [3666, 'Dashoguz', '03', , 4295],
        [3667, 'Lebap', '04', , 4295],
        [3668, 'Mary', '05', , 4295],
        [3669, 'Al Qasrayn', '02', 'TUN-AQS', 4296],
        [3670, 'Al Qayrawan', '03', 'TUN-AQY', 4296],
        [3671, 'Jundubah', '06', 'TUN-JND', 4296],
        [3672, 'Qafsah', '10', 'TUN-QFS', 4296],
        [3673, 'Kef', '14', 'TUN-AKA', 4296, { aliases: 'Al Kaf' }],
        [3674, 'Al Mahdiyah', '15', 'TUN-AMH', 4296],
        [3675, 'Al Munastir', '16', 'TUN-AMN', 4296],
        [3676, 'Bajah', '17', 'TUN-BAJ', 4296],
        [3677, 'Banzart', '18', 'TUN-BNZ', 4296],
        [3678, 'Nabul', '19', 'TUN-NAB', 4296],
        [3679, 'Silyanah', '22', 'TUN-SLY', 4296],
        [3680, 'Susah', '23', 'TUN-SUS', 4296],
        [3681, 'Bin', '27', 'TUN-BNR', 4296, { aliases: "Bin 'Arus" }],
        [3682, 'Madanin', '28', 'TUN-MDN', 4296, { aliases: 'Madaniyin' }],
        [3683, 'Qabis', '29', 'TUN-QAB', 4296],
        [3684, 'Qaf?ah', '30', 'TUN-QFS', 4296, { aliases: 'Qafsah' }],
        [3685, 'Qibili', '31', 'TUN-QBL', 4296],
        [3686, 'Safaqis', '32', 'TUN-SFQ', 4296],
        [3687, 'Sidi Bu Zayd', '33', 'TUN-SBZ', 4296],
        [3688, 'Tatawin', '34', 'TUN-TTW', 4296],
        [3689, 'Tawzar', '35', 'TUN-TWZ', 4296],
        [3690, 'Tunis', '36', 'TUN-TUN', 4296],
        [3691, 'Zaghwan', '37', 'TUN-ZGH', 4296],
        [3692, 'Ariana', '38', 'TUN-ARY', 4296, { aliases: 'Aryanah' }],
        [3693, 'Manouba', '39', 'TUN-ARY', 4296],
        [3694, 'Tongatapu', '02', , 4297],
        [3695, 'Vava', '03', , 4297],
        [3696, 'Adiyaman', '02', 'TUR-ADY', 4298],
        [3697, 'Afyonkarahisar', '03', 'TUR-AFY', 4298],
        [3698, 'Agri', '04', 'TUR-AGR', 4298],
        [3699, 'Amasya', '05', 'TUR-AMS', 4298],
        [3700, 'Antalya', '07', 'TUR-ANT', 4298],
        [3701, 'Artvin', '08', 'TUR-ART', 4298],
        [3702, 'Aydin', '09', 'TUR-AYD', 4298],
        [3703, 'Balikesir', '10', 'TUR-BLK', 4298],
        [3704, 'Bilecik', '11', 'TUR-BLC', 4298],
        [3705, 'Bingol', '12', 'TUR-BNG', 4298],
        [3706, 'Bitlis', '13', 'TUR-BTL', 4298],
        [3707, 'Bolu', '14', 'TUR-BOL', 4298],
        [3708, 'Burdur', '15', 'TUR-BRD', 4298],
        [3709, 'Bursa', '16', 'TUR-BUR', 4298],
        [3710, 'Canakkale', '17', 'TUR-CNK', 4298],
        [3711, 'Corum', '19', 'TUR-COR', 4298],
        [3712, 'Denizli', '20', 'TUR-DNZ', 4298],
        [3713, 'Diyarbakir', '21', 'TUR-DYR', 4298],
        [3714, 'Edirne', '22', 'TUR-EDR', 4298],
        [3715, 'Elazig', '23', 'TUR-ELZ', 4298],
        [3716, 'Erzincan', '24', 'TUR-ERZ', 4298],
        [3717, 'Erzurum', '25', 'TUR-ERR', 4298],
        [3718, 'Eskisehir', '26', 'TUR-ESK', 4298],
        [3719, 'Giresun', '28', 'TUR-GRS', 4298],
        [3720, 'Hatay', '31', 'TUR-HAT', 4298],
        [3721, 'Icel', '32', 'TUR-ICE', 4298],
        [3722, 'Isparta', '33', 'TUR-ISP', 4298],
        [3723, 'Istanbul', '34', 'TUR-IST', 4298],
        [3724, 'Izmir', '35', 'TUR-IZM', 4298],
        [3725, 'Kastamonu', '37', 'TUR-KST', 4298],
        [3726, 'Kayseri', '38', 'TUR-KYS', 4298],
        [3727, 'Kirklareli', '39', 'TUR-KRK', 4298],
        [3728, 'Kirsehir', '40', 'TUR-KRS', 4298],
        [3729, 'Kocaeli', '41', 'TUR-KOC', 4298],
        [3730, 'Kutahya', '43', 'TUR-KTH', 4298],
        [3731, 'Malatya', '44', 'TUR-MLT', 4298],
        [3732, 'Manisa', '45', 'TUR-MAN', 4298],
        [3733, 'Kahramanmaras', '46', 'TUR-KMA', 4298],
        [3734, 'Mugla', '48', 'TUR-MUG', 4298],
        [3735, 'Mus', '49', 'TUR-MUS', 4298],
        [3736, 'Nevsehir', '50', 'TUR-NVS', 4298],
        [3737, 'Ordu', '52', 'TUR-ORD', 4298],
        [3738, 'Rize', '53', 'TUR-RIZ', 4298],
        [3739, 'Sakarya', '54', 'TUR-SKR', 4298],
        [3740, 'Samsun', '55', 'TUR-SMS', 4298],
        [3741, 'Sinop', '57', 'TUR-SIN', 4298],
        [3742, 'Sivas', '58', 'TUR-SIV', 4298],
        [3743, 'Tekirdag', '59', 'TUR-TKR', 4298],
        [3744, 'Tokat', '60', 'TUR-TOK', 4298],
        [3745, 'Trabzon', '61', 'TUR-TRB', 4298],
        [3746, 'Tunceli', '62', 'TUR-TNC', 4298],
        [3747, 'Sanliurfa', '63', 'TUR-URF', 4298],
        [3748, 'Usak', '64', 'TUR-USA', 4298],
        [3749, 'Van', '65', 'TUR-VAN', 4298],
        [3750, 'Yozgat', '66', 'TUR-YZG', 4298],
        [3751, 'Ankara', '68', 'TUR-ANK', 4298],
        [3752, 'Gumushane', '69', 'TUR-GMS', 4298],
        [3753, 'Hakkari', '70', 'TUR-HKK', 4298],
        [3754, 'Konya', '71', 'TUR-KON', 4298],
        [3755, 'Mardin', '72', 'TUR-MRD', 4298],
        [3756, 'Nigde', '73', 'TUR-NIG', 4298],
        [3757, 'Siirt', '74', 'TUR-SII', 4298],
        [3758, 'Aksaray', '75', 'TUR-NVS', 4298],
        [3759, 'Batman', '76', 'TUR-SII', 4298],
        [3760, 'Bayburt', '77', 'TUR-GMS', 4298],
        [3761, 'Karaman', '78', 'TUR-KON', 4298],
        [3762, 'Kirikkale', '79', 'TUR-ANK', 4298],
        [3763, 'Sirnak', '80', 'TUR-SII', 4298],
        [3764, 'Adana', '81', 'TUR-ADA', 4298],
        [3765, 'Cankiri', '82', 'TUR-CNR', 4298],
        [3766, 'Gaziantep', '83', 'TUR-GZN', 4298],
        [3767, 'Kars', '84', 'TUR-KAR', 4298],
        [3768, 'Zonguldak', '85', 'TUR-ZNG', 4298],
        [3769, 'Ardahan', '86', 'TUR-KAR', 4298],
        [3770, 'Bartin', '87', 'TUR-ZNG', 4298],
        [3771, 'Igdir', '88', 'TUR-KAR', 4298],
        [3772, 'Karabuk', '89', 'TUR-ZNG', 4298],
        [3773, 'Kilis', '90', 'TUR-GZN', 4298],
        [3774, 'Osmaniye', '91', 'TUR-ADA', 4298],
        [3775, 'Yalova', '92', 'TUR-IST', 4298],
        [3776, 'Duzce', '93', 'TUR-BOL', 4298],
        [3777, 'Arima', '01', , 4299],
        [3778, 'Caroni', '02', , 4299],
        [3779, 'Mayaro', '03', , 4299],
        [3780, 'Nariva', '04', , 4299],
        [3781, 'Port-of-Spain', '05', , 4299],
        [3782, 'San Fernando', '10', , 4299],
        [3783, 'Tobago', '11', , 4299],
        [3784, 'Fu-chien', '01', 'TWN-PNG', 4301],
        [3785, 'Kao-hsiung', '02', 'TWN-KAS', 4301],
        [3786, "T'ai-pei", '03', 'TWN-TPS', 4301],
        [3787, "T'ai-wan", '04', 'TWN-NNT', 4301],
        [3788, 'Pwani', '02', 'TZA-PWA', 4302],
        [3789, 'Dodoma', '03', 'TZA-DOD', 4302],
        [3790, 'Iringa', '04', 'TZA-IRN', 4302],
        [3791, 'Kigoma', '05', 'TZA-KIG', 4302],
        [3792, 'Kilimanjaro', '06', 'TZA-KLM', 4302],
        [3793, 'Lindi', '07', 'TZA-LIN', 4302],
        [3794, 'Mara', '08', 'TZA-MAR', 4302],
        [3795, 'Mbeya', '09', 'TZA-MBE', 4302],
        [3796, 'Morogoro', '10', 'TZA-MRG', 4302],
        [3797, 'Mtwara', '11', 'TZA-MTW', 4302],
        [3798, 'Pemba North', '13', 'TZA-PNR', 4302],
        [3799, 'Ruvuma', '14', 'TZA-RUV', 4302],
        [3800, 'Shinyanga', '15', 'TZA-SHN', 4302],
        [3801, 'Singida', '16', 'TZA-SNG', 4302],
        [3802, 'Tabora', '17', 'TZA-TAB', 4302],
        [3803, 'Tanga', '18', 'TZA-TAN', 4302],
        [3804, 'Kagera', '19', 'TZA-ZMG', 4302, { aliases: 'Ziwa Magharibi' }],
        [3805, 'Pemba South', '20', 'TZA-PSO', 4302],
        [3806, 'Zanzibar Central', '21', 'TZA-ZCN', 4302],
        [3807, 'Zanzibar North', '22', 'TZA-ZNR', 4302],
        [3808, 'Dar es Salaam', '23', 'TZA-DES', 4302],
        [3809, 'Rukwa', '24', 'TZA-RUK', 4302],
        [3810, 'Zanzibar Urban', '25', 'TZA-ZUR', 4302],
        [3811, 'Arusha', '26', 'TZA-ARS', 4302],
        [3812, 'Manyara', '27', 'TZA-ARS', 4302],
        [3813, "Cherkas'ka Oblast'", '01', 'UKR-CHK', 4303],
        [
            3814,
            "Chernihivs'ka Oblast'",
            '02',
            'UKR-CHH',
            4303,
            { aliases: 'Chernigov' },
        ],
        [
            3815,
            "Chernivets'ka Oblast'",
            '03',
            'UKR-CHV',
            4303,
            { aliases: 'Chernovtsy' },
        ],
        [
            3816,
            "Dnipropetrovs'ka Oblast'",
            '04',
            'UKR-DNP',
            4303,
            { aliases: 'Dnepropetrovsk' },
        ],
        [3817, "Donets'ka Oblast'", '05', 'UKR-DNT', 4303, { aliases: 'Donetsk' }],
        [3818, "Ivano-Frankivs'ka Oblast'", '06', 'UKR-IVF', 4303],
        [3819, "Kharkivs'ka Oblast'", '07', 'UKR-KKV', 4303],
        [3820, "Khersons'ka Oblast'", '08', 'UKR-KRN', 4303],
        [3821, "Khmel'nyts'ka Oblast'", '09', 'UKR-KML', 4303],
        [
            3822,
            "Kirovohrads'ka Oblast'",
            '10',
            'UKR-KVG',
            4303,
            { aliases: 'Kirovograd' },
        ],
        [3823, 'Krym', '11', 'UKR-KRM', 4303, { aliases: 'Crimea' }],
        [3824, 'Kyyiv', '12', 'UKR-KYV', 4303, { aliases: 'Kiev City' }],
        [
            3825,
            "Kyyivs'ka Oblast'",
            '13',
            'UKR-KYV',
            4303,
            { aliases: 'Kiev Province' },
        ],
        [3826, "Luhans'ka Oblast'", '14', 'UKR-LHN', 4303, { aliases: 'Luhansk' }],
        [3827, "L'vivs'ka Oblast'", '15', 'UKR-LVV', 4303, { aliases: "L'viv" }],
        [
            3828,
            "Mykolayivs'ka Oblast'",
            '16',
            'UKR-MKL',
            4303,
            { aliases: 'Mykolaiv;Nikolayev' },
        ],
        [3829, "Odes'ka Oblast'", '17', 'UKR-ODS', 4303, { aliases: 'Odessa' }],
        [3830, "Poltavs'ka Oblast'", '18', 'UKR-PLT', 4303, { aliases: 'Poltava' }],
        [
            3831,
            "Rivnens'ka Oblast'",
            '19',
            'UKR-RVN',
            4303,
            { aliases: 'Rivne;Rovno' },
        ],
        [
            3832,
            "Sevastopol'",
            '20',
            'UKR-KRM',
            4303,
            { notes: 'City, Province 3823' },
        ],
        [3833, "Sums'ka Oblast'", '21', 'UKR-SUM', 4303, { aliases: 'Sumy' }],
        [
            3834,
            "Ternopil's'ka Oblast'",
            '22',
            'UKR-TRN',
            4303,
            { aliases: 'Ternopil' },
        ],
        [
            3835,
            "Vinnyts'ka Oblast'",
            '23',
            'UKR-VNT',
            4303,
            { aliases: 'Vinnytsia;Vinnitsa' },
        ],
        [3836, "Volyns'ka Oblast'", '24', 'UKR-VLY', 4303, { aliases: 'Volyn' }],
        [3837, "Zakarpats'ka Oblast'", '25', 'UKR-ZKP', 4303],
        [3838, "Zaporiz'ka Oblast'", '26', 'UKR-ZPR', 4303],
        [
            3839,
            "Zhytomyrs'ka Oblast'",
            '27',
            'UKR-ZTM',
            4303,
            { aliases: 'Zhytomyr;Zhitomir' },
        ],
        [3840, 'Busoga', '05', 'UGA-BUS', 4304],
        [3841, 'Karamoja', '08', 'UGA-KRM', 4304],
        [3842, 'South Buganda', '12', 'UGA-STB', 4304],
        [3843, 'Nile', '21', 'UGA-NIL', 4304],
        [3844, 'North Buganda', '22', 'UGA-NRB', 4304],
        [3845, 'Kampala', '37', 'UGA-CNT', 4304],
        [3846, 'Mubende', '56', 'UGA-NRB', 4304],
        [3847, 'Adjumani', '65', 'UGA-NIL', 4304],
        [3848, 'Bugiri', '66', 'UGA-BUS', 4304],
        [3849, 'Busia', '67', 'UGA-EST', 4304],
        [3850, 'Katakwi', '69', 'UGA-EST', 4304],
        [3851, 'Nakasongola', '73', 'UGA-NRB', 4304],
        [3852, 'Sembabule', '74', 'UGA-NRB', 4304],
        [3853, 'Arua', '77', 'UGA-NIL', 4304],
        [3854, 'Iganga', '78', 'UGA-BUS', 4304],
        [3855, 'Kabarole', '79', 'UGA-WST', 4304],
        [3856, 'Kaberamaido', '80', 'UGA-EST', 4304],
        [3857, 'Kamwenge', '81', 'UGA-WST', 4304],
        [3858, 'Kanungu', '82', 'UGA-STH', 4304],
        [3859, 'Kayunga', '83', 'UGA-NRB', 4304],
        [3860, 'Kitgum', '84', 'UGA-NRT', 4304],
        [3861, 'Kyenjojo', '85', 'UGA-WST', 4304],
        [3862, 'Mayuge', '86', 'UGA-NRB', 4304],
        [3863, 'Mbale', '87', 'UGA-EST', 4304],
        [3864, 'Moroto', '88', 'UGA-KRM', 4304],
        [3865, 'Mpigi', '89', 'UGA-CNT', 4304],
        [3866, 'Mukono', '90', 'UGA-NRB', 4304],
        [3867, 'Nakapiripirit', '91', 'UGA-KRM', 4304],
        [3868, 'Pader', '92', 'UGA-NRT', 4304],
        [3869, 'Rukungiri', '93', 'UGA-STH', 4304],
        [3870, 'Sironko', '94', 'UGA-EST', 4304],
        [3871, 'Soroti', '95', 'UGA-EST', 4304],
        [3872, 'Wakiso', '96', 'UGA-CNT', 4304],
        [3873, 'Yumbe', '97', 'UGA-NIL', 4304],
        [
            3874,
            'District of Columbia',
            '11',
            ,
            4306,
            { code: 'DC', region: 4342, aliases: 'Washington, D.C.' },
        ],
        [3875, 'Artigas', '01', 'URY-ART', 4307],
        [3876, 'Canelones', '02', 'URY-CNL', 4307],
        [3877, 'Cerro Largo', '03', 'URY-CLA', 4307],
        [3878, 'Colonia', '04', 'URY-COL', 4307],
        [3879, 'Durazno', '05', 'URY-DRZ', 4307],
        [3880, 'Flores', '06', 'URY-FLR', 4307],
        [3881, 'Lavalleja', '08', 'URY-LVL', 4307],
        [3882, 'Maldonado', '09', 'URY-MLD', 4307],
        [3883, 'Montevideo', '10', 'URY-MNT', 4307],
        [3884, 'Paysandu', '11', 'URY-PYS', 4307],
        [3885, 'Rivera', '13', 'URY-RIV', 4307],
        [3886, 'Rocha', '14', 'URY-ROC', 4307],
        [3887, 'Salto', '15', 'URY-SAL', 4307],
        [3888, 'Soriano', '17', 'URY-SOR', 4307],
        [3889, 'Tacuarembo', '18', 'URY-TCR', 4307],
        [3890, 'Treinta y Tres', '19', 'URY-TYT', 4307],
        [3891, 'Andijon', '01', , 4308],
        [3892, 'Bukhoro', '02', , 4308],
        [3893, 'Farghona', '03', , 4308],
        [3894, 'Jizzakh', '04', , 4308],
        [3895, 'Khorazm', '05', , 4308],
        [3896, 'Namangan', '06', , 4308],
        [3897, 'Nawoiy', '07', , 4308],
        [3898, 'Qashqadaryo', '08', , 4308],
        [3899, 'Qoraqalpoghiston', '09', , 4308],
        [3900, 'Samarqand', '10', , 4308],
        [3901, 'Sirdaryo', '11', , 4308],
        [3902, 'Surkhondaryo', '12', , 4308],
        [3903, 'Toshkent', '13', , 4308],
        [3904, 'Toshkent', '14', , 4308],
        [3905, 'Charlotte', '01', , 4310],
        [3906, 'Grenadines', '06', , 4310],
        [3907, 'Anzoategui', '02', 'VEN-ANZ', 4311],
        [3908, 'Apure', '03', 'VEN-APU', 4311],
        [3909, 'Aragua', '04', 'VEN-ARA', 4311],
        [3910, 'Barinas', '05', 'VEN-BRN', 4311],
        [3911, 'Carabobo', '07', 'VEN-CRB', 4311],
        [3912, 'Cojedes', '08', 'VEN-CJD', 4311],
        [3913, 'Delta Amacuro', '09', 'VEN-DAM', 4311],
        [3914, 'Falcon', '11', 'VEN-FLC', 4311],
        [3915, 'Guarico', '12', 'VEN-GUA', 4311],
        [3916, 'Lara', '13', 'VEN-LAR', 4311],
        [3917, 'Merida', '14', 'VEN-MER', 4311],
        [3918, 'Miranda', '15', 'VEN-MRN', 4311],
        [3919, 'Monagas', '16', 'VEN-MNG', 4311],
        [3920, 'Nueva Esparta', '17', 'VEN-NES', 4311],
        [3921, 'Portuguesa', '18', 'VEN-PRT', 4311],
        [3922, 'Tachira', '20', 'VEN-TCH', 4311],
        [3923, 'Trujillo', '21', 'VEN-TRJ', 4311],
        [3924, 'Yaracuy', '22', 'VEN-YRC', 4311],
        [3925, 'Zulia', '23', 'VEN-ZUL', 4311],
        [3926, 'Dependencias Federales', '24', 'VEN-DFR', 4311],
        [3927, 'Vargas', '26', 'VEN-DFD', 4311],
        [3928, 'An Giang', '01', 'VNM-AGI', 4314],
        [3929, 'Bac Thai', '02', 'VNM-BTH', 4314],
        [3930, 'Ben Tre', '03', 'VNM-BTR', 4314],
        [
            3931,
            'Binh Tri Thien',
            '04',
            'VNM-QBN',
            4314,
            { notes: 'Region, of Quang Binh, Quang Tri, Thua Thien' },
        ],
        [3932, 'Cao Bang', '05', 'VNM-CBA', 4314],
        [3933, 'Dac Lac', '07', 'VNM-DLA', 4314],
        [3934, 'Dong Thap', '09', 'VNM-DTH', 4314],
        [3935, 'Ha Bac', '11', 'VNM-HBA', 4314],
        [3936, 'Hai Hung', '12', 'VNM-HHU', 4314],
        [3937, 'Hai Phong', '13', 'VNM-HPH', 4314],
        [
            3938,
            'Ha Nam Ninh',
            '14',
            'VNM-NHA',
            4314,
            { notes: 'Region, of Ha Nam, Nam Dinh, and Ninh Binh' },
        ],
        [
            3939,
            'Ha Son Binh',
            '16',
            'VNM-HTY',
            4314,
            { notes: 'Region, of Ha Tay and Hoa Binh' },
        ],
        [3940, 'Ha Tuyen', '17', 'VNM-TQN', 4314],
        [3941, 'Hoang Lien Son', '19', 'VNM-LCA', 4314],
        [3942, 'Ho Chi Minh', '20', 'VNM-HCM', 4314],
        [3943, 'Kien Giang', '21', 'VNM-KGI', 4314],
        [3944, 'Lai Chau', '22', 'VNM-LCH', 4314],
        [3945, 'Lam Dong', '23', 'VNM-LDO', 4314],
        [3946, 'Long An', '24', 'VNM-LAN', 4314],
        [3947, 'Minh Hai', '25', 'VNM-MHA', 4314],
        [
            3948,
            'Nghe Tinh',
            '26',
            'VNM-NAN',
            4314,
            { notes: 'Region, of Nghe An and Ha Tinh' },
        ],
        [3949, 'Nghia Binh', '27', 'VNM-BDN', 4314],
        [3950, 'Phu Khanh', '28', 'VNM-KHO', 4314],
        [3951, 'Quang Nam-Da Nang', '29', 'VNM-QND', 4314],
        [3952, 'Quang Ninh', '30', 'VNM-QNI', 4314],
        [3953, 'Song Be', '31', 'VNM-SBE', 4314],
        [3954, 'Son La', '32', 'VNM-SLA', 4314],
        [3955, 'Tay Ninh', '33', 'VNM-TNI', 4314],
        [3956, 'Thanh Hoa', '34', 'VNM-THO', 4314],
        [3957, 'Thai Binh', '35', 'VNM-TBI', 4314],
        [3958, 'Thuan Hai', '36', 'VNM-NTN', 4314],
        [3959, 'Tien Giang', '37', 'VNM-TGI', 4314],
        [3960, 'Vinh Phu', '38', 'VNM-VPH', 4314],
        [3961, 'Lang Son', '39', 'VNM-LSO', 4314],
        [3962, 'Dong Nai', '40', 'VNM-DNA', 4314],
        [3963, 'An Giang', '43', 'VNM-AGI', 4314],
        [3964, 'Dac Lac', '44', 'VNM-DLA', 4314],
        [3965, 'Dong Nai', '45', 'VNM-DNA', 4314],
        [3966, 'Dong Thap', '46', 'VNM-DTH', 4314],
        [3967, 'Kien Giang', '47', 'VNM-KGI', 4314],
        [3968, 'Minh Hai', '48', 'VNM-MHA', 4314],
        [3969, 'Song Be', '49', 'VNM-SBE', 4314],
        [3970, 'Vinh Phu', '50', 'VNM-VPH', 4314],
        [3971, 'Ha Noi', '51', 'VNM-HNO', 4314],
        [3972, 'Ho Chi Minh', '52', 'VNM-HCM', 4314],
        [3973, 'Ba Ria-Vung Tau', '53', 'VNM-BRV', 4314],
        [3974, 'Binh Dinh', '54', 'VNM-BDN', 4314],
        [3975, 'Binh Thuan', '55', 'VNM-BTN', 4314],
        [3976, 'Can Tho', '56', 'VNM-CTH', 4314],
        [3977, 'Gia Lai', '57', 'VNM-GLA', 4314],
        [3978, 'Ha Giang', '58', 'VNM-HGN', 4314],
        [3979, 'Ha Tay', '59', 'VNM-HTY', 4314],
        [3980, 'Ha Tinh', '60', 'VNM-HTN', 4314],
        [3981, 'Hoa Binh', '61', 'VNM-HBN', 4314],
        [3982, 'Khanh Hoa', '62', 'VNM-KHO', 4314],
        [3983, 'Kon Tum', '63', 'VNM-KTM', 4314],
        [3984, 'Quang Tri', '64', 'VNM-QTR', 4314],
        [3985, 'Nam Ha', '65', 'VNM-NHA', 4314],
        [3986, 'Nghe An', '66', 'VNM-NAN', 4314],
        [3987, 'Ninh Binh', '67', 'VNM-NBN', 4314],
        [3988, 'Ninh Thuan', '68', 'VNM-NTN', 4314],
        [3989, 'Phu Yen', '69', 'VNM-PYN', 4314],
        [3990, 'Quang Binh', '70', 'VNM-QBN', 4314],
        [3991, 'Quang Ngai', '71', 'VNM-QNG', 4314],
        [3992, 'Quang Tri', '72', 'VNM-QTR', 4314],
        [3993, 'Soc Trang', '73', 'VNM-STR', 4314],
        [3994, 'Thua Thien', '74', 'VNM-TTH', 4314],
        [3995, 'Tra Vinh', '75', 'VNM-TVN', 4314],
        [3996, 'Tuyen Quang', '76', 'VNM-TQN', 4314],
        [3997, 'Vinh Long', '77', 'VNM-VLN', 4314],
        [3998, 'Da Nang', '78', 'VNM-QND', 4314],
        [3999, 'Hai Duong', '79', 'VNM-HHU', 4314],
        [4000, 'Ha Nam', '80', 'VNM-NHA', 4314],
        [4001, 'Hung Yen', '81', 'VNM-HTY', 4314],
        [4002, 'Nam Dinh', '82', 'VNM-NHA', 4314],
        [4003, 'Phu Tho', '83', 'VNM-VPH', 4314],
        [4004, 'Quang Nam', '84', 'VNM-QND', 4314],
        [4005, 'Thai Nguyen', '85', 'VNM-BTH', 4314],
        [4006, 'Can Tho', '87', 'VNM-CTH', 4314],
        [4007, 'Dak Lak', '88', 'VNM-DLA', 4314],
        [4008, 'Lai Chau', '89', 'VNM-LCH', 4314],
        [4009, 'Lao Cai', '90', 'VNM-LCA', 4314],
        [4010, 'Dak Nong', '91', 'VNM-DLA', 4314],
        [4011, 'Dien Bien', '92', 'VNM-LCH', 4314],
        [4012, 'Hau Giang', '93', 'VNM-CTH', 4314],
        [4013, 'Ambrym', '05', , 4315],
        [4014, 'Aoba', '06', , 4315],
        [4015, 'Torba', '07', , 4315],
        [4016, 'Efate', '08', , 4315],
        [4017, 'Epi', '09', , 4315],
        [4018, 'Malakula', '10', , 4315],
        [4019, 'Paama', '11', , 4315],
        [4020, 'Pentecote', '12', , 4315],
        [4021, 'Sanma', '13', , 4315],
        [4022, 'Shepherd', '14', , 4315],
        [4023, 'Tafea', '15', , 4315],
        [4024, 'Malampa', '16', , 4315],
        [4025, 'Penama', '17', , 4315],
        [4026, 'Shefa', '18', , 4315],
        [4027, 'Aiga-i-le-Tai', '02', , 4317],
        [4028, 'Atua', '03', , 4317],
        [4029, 'Fa', '04', , 4317],
        [4030, 'Gaga', '05', , 4317],
        [4031, 'Va', '06', , 4317],
        [4032, 'Gagaifomauga', '07', , 4317],
        [4033, 'Palauli', '08', , 4317],
        [4034, 'Satupa', '09', , 4317],
        [4035, 'Tuamasaga', '10', , 4317],
        [4036, 'Vaisigano', '11', , 4317],
        [4037, 'Abyan', '01', 'YEM-ABY', 4318],
        [4038, 'Adan', '02', 'YEM-ADN', 4318],
        [4039, 'Al Mahrah', '03', 'YEM-AMR', 4318],
        [4040, 'Hadramawt', '04', 'YEM-HDH', 4318],
        [4041, 'Shabwah', '05', 'YEM-ATQ', 4318],
        [4042, 'Al Hudaydah', '08', 'YEM-AHD', 4318],
        [4043, 'Al Mahwit', '10', 'YEM-AMW', 4318],
        [4044, 'Dhamar', '11', 'YEM-DHM', 4318],
        [4045, "Ma'rib", '14', 'YEM-MRB', 4318],
        [4046, 'Sa', '15', 'YEM-SDH', 4318],
        [4047, 'San', '16', 'YEM-SAN', 4318],
        [4048, "Al Bayda'", '20', 'YEM-ABD', 4318],
        [4049, 'Hajjah', '22', 'YEM-HJJ', 4318],
        [4050, 'Ibb', '23', 'YEM-IBB', 4318],
        [4051, 'Lahij', '24', 'YEM-LHJ', 4318],
        [4052, 'Ta', '25', 'YEM-TAZ', 4318],
        [4053, 'KwaZulu-Natal', '02', 'ZAF-KNT', 4320],
        [4054, 'Free State', '03', 'ZAF-OFS', 4320],
        [4055, 'Eastern Cape', '05', 'ZAF-ECP', 4320],
        [4056, 'Gauteng', '06', 'ZAF-GAT', 4320],
        [4057, 'Mpumalanga', '07', 'ZAF-MPM', 4320],
        [4058, 'Northern Cape', '08', 'ZAF-NCP', 4320],
        [4059, 'Limpopo', '09', 'ZAF-NRT', 4320],
        [4060, 'Western Cape', '11', 'ZAF-WCP', 4320],
        [4061, 'Luapula', '04', 'ZMB-LUA', 4321],
        [4062, 'North-Western', '06', 'ZMB-NWS', 4321],
        [4063, 'Copperbelt', '08', 'ZMB-CPP', 4321],
        [4064, 'Lusaka', '09', 'ZMB-LUS', 4321],
        [4065, 'Manicaland', '01', , 4322],
        [4066, 'Midlands', '02', , 4322],
        [4067, 'Mashonaland Central', '03', , 4322],
        [4068, 'Mashonaland East', '04', , 4322],
        [4069, 'Mashonaland West', '05', , 4322],
        [4070, 'Matabeleland North', '06', , 4322],
        [4071, 'Matabeleland South', '07', , 4322],
        [4072, 'Masvingo', '08', , 4322],
        [4073, 'Bulawayo', '09', , 4322],
        [4074, 'Harare', '10', , 4322],
        [4171, 'Hong Kong', 'HK', 'CHN-HKG', 4124],
        [4358, 'Guantanamo Bay', 'US', 'USA-PS1', 4306],
        [4359, "Grand' Anse", '14', 'HTI-GAN', 4175],
        [4360, 'Nord', '09', 'HTI-NOR', 4175],
        [4361, 'Ouest', '11', 'HTI-OUE', 4175],
        [4362, 'Sud', '12', 'HTI-SUD', 4175],
        [4363, 'Saint Andrew', '08', 'JAM-SND', 4188],
        [4364, 'Chaco', '03', 'ARG-CHA', 4088],
        [4365, 'Cordoba', '05', 'ARG-CRD', 4088],
        [4366, 'Distrito Federal', '07', 'ARG-DFD', 4088],
        [4367, 'La Rioja', '12', 'ARG-LRI', 4088],
        [4368, 'Misiones', '14', 'ARG-MSN', 4088],
        [4369, 'Rio Negro', '16', 'ARG-RNE', 4088],
        [4370, 'Santa Cruz', '20', 'ARG-SCR', 4088],
        [4371, 'San Juan', '18', 'ARG-SJU', 4088],
        [4372, 'La Paz', '04', 'BOL-LPA', 4106],
        [4373, 'Santa Cruz', '08', 'BOL-SCR', 4106],
        [4374, 'Amazonas', '04', 'BRA-AMZ', 4107],
        [4375, 'Distrito Federal', '07', 'BRA-DFD', 4107],
        [4376, 'Para', '16', 'BRA-PAR', 4107],
        [4377, 'Santa Catarina', '26', 'BRA-SCT', 4107],
        [4378, 'San Jose', '08', 'CRI-SJO', 4126],
        [4379, 'Amazonas', '01', 'COL-AMZ', 4125],
        [4380, 'Sucre', '27', 'COL-SUC', 4125],
        [4381, 'Bolivar', '03', 'ECU-BLV', 4138],
        [4382, 'Central', '06', 'PRY-CNT', 4261],
        [4383, 'Florida', '07', 'URY-FLD', 4307],
        [4384, 'Amazonas', '01', 'PER-AMZ', 4249],
        [4385, 'La Libertad', '13', 'PER-LLB', 4249],
        [4386, 'Limburg', '05', 'BEL-LBM', 4098],
        [4387, 'Luxembourg', '06', 'BEL-LXM', 4098],
        [4388, 'Ruse', '53', 'BGR-RZG', 4100],
        [4389, 'Centre', 'A3', 'FRA-CEN', 4151],
        [4399, 'Ha', '10', 'BTN-HA_', 4109],
        [4400, 'Punjab', '23', 'IND-PNJ', 4181],
        [4401, 'Timor Timur', 'TP', 'IDN-TTI', 4324],
        [4402, 'Toledo', '06', 'BLZ-TOL', 4113],
        [4403, 'La Union', '36', 'PHL-ILC', 4252],
        [4404, 'Al Jawf', '03', 'SAU-AJA', 4268],
        [4405, "'Asir", '11', 'SAU-ASI', 4268],
        [4406, 'Ash Sharqiyah', '06', 'SAU-ASH', 4268],
        [4407, 'Adrar', '34', 'DZA-ADR', 4137],
        [4408, 'Est', '04', 'CMR-EST', 4123],
        [4409, 'Nord-Ouest', '07', 'CMR-NOU', 4123],
        [4410, 'Ouest', '08', 'CMR-OUE', 4123],
        [4411, 'Centre', '11', 'CMR-CNT', 4123],
        [4412, 'Nord', '13', 'CMR-NOR', 4123],
        [4413, 'Sud', '14', 'CMR-SUD', 4123],
        [4414, 'Plateaux', '08', 'COG-PLT', 4118],
        [4415, 'Ash Sharqiyah', '14', 'EGY-ASH', 4140],
        [4416, 'Western', '05', 'GMB-WST', 4161],
        [4417, 'Central', '04', 'GHA-CNT', 4158],
        [4418, 'Eastern', '05', 'GHA-EST', 4158],
        [4419, 'Northern', '06', 'GHA-NRT', 4158],
        [4420, 'Western', '09', 'GHA-WST', 4158],
        [4421, 'Savanes', '87', 'CIV-FRK', 4120],
        [4422, 'Central', '01', 'KEN-CNT', 4191],
        [4423, 'Coast', '02', 'KEN-COA', 4191],
        [4424, 'Western', '09', 'KEN-WST', 4191],
        [4425, 'Maryland', '06', 'LBR-MRY', 4207],
        [4426, 'Eastern', '03', 'KEN-EST', 4191],
        [4427, 'Mwanza', '25', 'MWI-STH', 4231],
        [4428, 'Gaza', '02', 'MOZ-GAZ', 4234],
        [4429, 'Plateau', '17', , 4105],
        [4430, 'Eastern', '01', 'SLE-EST', 4278],
        [4431, 'Northern', '02', 'SLE-NRT', 4278],
        [4432, 'Southern', '03', 'SLE-STH', 4278],
        [4433, 'Central', '01', 'BWA-CNT', 4111],
        [4434, 'Southern', '10', 'BWA-STH', 4111],
        [4435, 'North-West', '11', 'BWA-NGM', 4111],
        [4436, 'North-Western Province', '01', 'ZAF-NWS', 4320],
        [4437, 'Ash Shamaliyah', '18', 'BHR-AMS', 4101],
        [
            4438,
            'Al Wusta',
            '19',
            'BHR-AMW',
            4101,
            { aliases: 'Al Mintaqah al Wusta' },
        ],
        [4439, 'Praslin', '11', , 4204],
        [4440, 'Central', '01', , 4147],
        [4441, 'Eastern', '02', , 4147],
        [4442, 'Northern', '03', , 4147],
        [4443, 'Western', '05', , 4147],
        [4444, 'Victoria', '07', 'AS07', 4091],
        [4445, 'Central', '01', 'PP01', 4251],
        [4446, 'Northern', '04', 'PP04', 4251],
        [4447, 'Western', '06', 'PP06', 4251],
        [
            4448,
            'Manchester',
            'I2',
            'GBR-ENG-GMN',
            4153,
            { region: 4345, notes: 'Unitary District' },
        ],
        [4449, 'Saint George', '03', , 4080],
        [4450, 'Saint John', '04', , 4080],
        [4451, 'Saint Mary', '05', , 4080],
        [4452, 'Saint Paul', '06', , 4080],
        [4453, 'Saint Peter', '07', , 4080],
        [4454, 'Huila', '09', 'AGO-HUI', 4085],
        [4455, 'Namibe', '13', 'AGO-NAM', 4085],
        [4456, 'Saint Andrew', '02', , 4096],
        [4457, 'Saint James', '04', , 4096],
        [4458, 'Saint Joseph', '06', , 4096],
        [4459, 'Saint Thomas', '11', , 4096],
        [
            4460,
            'Barisal',
            '01',
            'BGD-BRS',
            4097,
            { notes: 'County, Province |Barisal|' },
        ],
        [
            4461,
            'Dhaka',
            '32',
            'BGD-DAK',
            4097,
            { notes: 'County, Province |Dhaka|' },
        ],
        [4462, 'Montana', '47', 'BGR-MKH', 4100],
        [4463, "Ar Rifa' wa al Mintaqah al Janubiyah", '07', 'BHR-ARM', 4101],
        [4464, 'Alibori', '07', 'BEN-BOR', 4103],
        [4465, 'Atakora', '08', 'BEN-ATK', 4103],
        [4466, 'Atlanyique', '09', 'BEN-ATL', 4103],
        [4467, 'Borgou', '10', 'BEN-BOR', 4103],
        [4468, 'Collines', '11', 'BEN-ZOU', 4103],
        [4469, 'Kouffo', '12', 'BEN-MON', 4103],
        [4470, 'Donga', '13', 'BEN-ATK', 4103],
        [4471, 'Littoral', '14', 'BEN-ATL', 4103],
        [4472, 'Mono', '15', 'BEN-MON', 4103],
        [4473, 'Oueme', '16', 'BEN-OUE', 4103],
        [4474, 'Plateau', '17', 'BEN-OUE', 4103],
        [4475, 'Zou', '18', 'BEN-ZOU', 4103],
        [4476, 'Saint George', '06', , 4104],
        [4477, 'Southampton', '10', , 4104],
        [4478, 'Bolivar', '35', 'COL-BLV', 4125],
        [4479, 'Arta', '08', 'DJI-DKH', 4133],
        [4480, 'Saint Andrew', '02', , 4135],
        [4481, 'Saint David', '03', , 4135],
        [4482, 'Saint George', '04', , 4135],
        [4483, 'Saint John', '05', , 4135],
        [4484, 'Saint Mark', '08', , 4135],
        [4485, 'Saint Patrick', '09', , 4135],
        [4486, 'Saint Peter', '11', , 4135],
        [4487, 'Southern', '11', 'ETH-GGO', 4144],
        [4488, 'Central', '79', 'GBR-SCT-CNT', 4153, { region: 4347 }],
        [4489, 'Saint Andrew', '01', , 4154],
        [4490, 'Saint David', '02', , 4154],
        [4491, 'Saint George', '03', , 4154],
        [4492, 'Saint Patrick', '06', , 4154],
        [4493, 'Colon', '03', , 4173],
        [4494, 'La Paz', '12', , 4173],
        [4495, 'Banteay Meanchey', '00', 'KHM-BTD', 4193],
        [4496, 'Batdambang', '01', 'KHM-BTD', 4193],
        [4497, 'Eastern', '02', , 4200],
        [4498, 'Western', '08', , 4200],
        [4499, 'Choiseul', '04', , 4204],
        [4500, 'Central', '29', , 4206],
        [4501, 'Northern', '31', , 4206],
        [4502, 'Southern', '34', , 4206],
        [4503, 'Western', '36', , 4206],
        [4504, 'Laayoune', '35', '', 4213, { notes: 'Located in Western Sahara.' }],
        [4505, 'Calarasi', '66', , 4215],
        [4506, 'Distrito Federal', '09', 'MEX-DTD', 4232],
        [4507, 'Autonoma Atlantico Norte', '17', 'NIC-ZEL', 4240],
        [4508, 'Region Autonoma Atlantico Sur', '18', 'NIC-ZEL', 4240],
        [4509, 'Overijssel', '08', 'NLD-OVR', 4241],
        [4510, 'Al Wusta', '03', , 4247],
        [4511, 'Ash Sharqiyah', '04', , 4247],
        [4512, 'San Jose', 'F6', 'PHL-CLZ', 4252],
        [4513, 'Chechnya Republic', 'CI', , 4266],
        [4514, 'Central', '10', , 4269],
        [4515, 'Western', '11', , 4269],
        [4516, "Grand' Anse", '14', , 4270],
        [4517, 'Al Wahadah State', '40', 'SDN-AAN', 4271],
        [4518, 'Central Equatoria State', '44', 'SDN-AIS', 4271],
        [4519, 'Amnat Charoen', '77', 'THA-URT', 4291],
        [4520, 'Nakhon Phanom', '73', 'THA-NPH', 4291],
        [4521, 'Saint Andrew', '06', , 4299],
        [4522, 'Saint David', '07', , 4299],
        [4523, 'Saint George', '08', , 4299],
        [4524, 'Saint Patrick', '09', , 4299],
        [4525, 'Central', '18', 'UGA-CNT', 4304],
        [4526, 'Eastern', '20', 'UGA-EST', 4304],
        [4527, 'Northern', '23', 'UGA-NRT', 4304],
        [4528, 'Southern', '24', 'UGA-STH', 4304],
        [4529, 'Western', '25', 'UGA-WST', 4304],
        [4530, 'Jinja', '33', 'UGA-BUS', 4304],
        [4531, 'Kalangala', '36', 'UGA-STB', 4304],
        [4532, 'Kiboga', '42', 'UGA-NRB', 4304],
        [4533, 'Mbarara', '52', 'UGA-STH', 4304],
        [4534, 'Masaka', '71', 'UGA-STB', 4304],
        [4535, 'Vinh Puc Province', '86', 'VNM-VPH', 4314],
        [4536, 'Boqueron', '24', 'PRY-BQR', 4261],
        [4537, 'Negros Occidental', '45', 'PHL-WVS', 4252],
        [4538, 'Liban-Sud', '02', 'LBN-SAY', 4203],
    ],
});
exports.default = Province;
//# sourceMappingURL=province.model.js.map